/* autolinas_dss_mkl.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dss_inside.h"
#include "autolinas_dss_upperrow_inside.h"



/* for Intel Math Kernel Library (MLK) */
/* Pardiso direct sparse solver */

static int N;
static int IA[System_maxDofs + 1];
static double X[System_maxDofs];

extern int pardiso_
(void *, int *, int *, int *, int *, int *,
 double *, int *, int *, int *, int *, int *,
 int *, double *, double *, int *);
#define PARDISO pardiso_

/* internal solver memory pointer <pt> */
/* 32-bit: int pt[64]; 64-bit: long int pt[64] */
/* or void *pt[64] should be OK on both architectures */
static void *MKL_pt[64];

/* matrix type and solution scheme */
static int MKL_mtype = 2; /* real symmetric positive definite matrix */
static int MKL_nrhs = 1; /* number of right hand sides. */
static int MKL_maxfct = 1; /* maximum number of numerical factorizations. */
static int MKL_mnum = 1; /* which factorization to use. */

/* PARSIDO control parameters */
static int MKL_iparm[64];

#if 0
static int MKL_msglvl = 1; /* print statistical information */
#else
static int MKL_msglvl = 0; /* no statistical information */
#endif



void AutoLinAS_INSIDE_Factorize (void)
{
  /* PARSIDO control parameters */
  int phase;
  int error;

  /* auxiliary variables */
  double ddum; /* double dummy */
  int idum; /* integer dummy. */

  int iDof;
  int i;
  
  N = System_nDofs;

  /* use <System_index> as JA[] */
  /* modify <System_index> */
  /* index must start from 1 */
  for (i = 0; i < System_maxIndexs[System_nDofs - 1]; i++) {
    System_index[i]++;
  }
  
  /* prepare IA[] */
  /* index must start from 1 */
  for (iDof = 0; iDof < System_nDofs; iDof++) {
    IA[iDof] = System_minIndexs[iDof] + 1;
  }
  IA[System_nDofs] = System_maxIndexs[System_nDofs - 1] + 1;

  /* use <SM> as value[] */



  /* call Pardiso symbolic + numerical factorization stage */

  fprintf (stderr, " solve start\n");

  for (i = 0; i < 64; i++) {
    MKL_iparm[i] = 0;
  }

  MKL_iparm[0] = 0; /* solver default */

  /* Numbers of processors, value of OMP_NUM_THREADS */
#if 1
  MKL_iparm[2] = 1;
#else
  MKL_iparm[2] = 4;
#endif

  for (i = 0; i < 64; i++) {
    MKL_pt[i] = 0;
  }
  


  phase = 11;
  error = 0;
  idum = 0;
  ddum = 0.0;
  PARDISO (MKL_pt, &MKL_maxfct, &MKL_mnum, &MKL_mtype, &phase,
	   &N, SM, IA, System_index, &idum, &MKL_nrhs,
	   MKL_iparm, &MKL_msglvl, &ddum, &ddum, &error);
  if (error != 0) {
    fprintf (stderr, " ERROR during symbolic factorization: %d \n", error);
    exit (1);
  }
  fprintf (stderr, " symbolic factorization completed \n");
  


  phase = 22;
  error = 0;
  idum = 0;
  ddum = 0.0;
  PARDISO (MKL_pt, &MKL_maxfct, &MKL_mnum, &MKL_mtype, &phase,
	   &N, SM, IA, System_index, &idum, &MKL_nrhs,
	   MKL_iparm, &MKL_msglvl, &ddum, &ddum, &error);
  if (error != 0) {
    fprintf (stderr, " ERROR during numerical factorization: %d \n", error);
    exit (2);
  }
  fprintf (stderr, " numerical factorization completed \n");
}

void AutoLinAS_INSIDE_Solve (void)
{
  /* PARSIDO control parameters */
  int phase;
  int error;

  /* auxiliary variables */
  int idum; /* integer dummy. */

  int iDof;

  /* use <SV1> as B[] */

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    X[iDof] = 0.0;
  }



  /* call Pardiso solve stage */

  phase = 33;
  error = 0;
  idum = 0;
  PARDISO (MKL_pt, &MKL_maxfct, &MKL_mnum, &MKL_mtype, &phase,
	   &N, SM, IA, System_index, &idum, &MKL_nrhs,
	   MKL_iparm, &MKL_msglvl, SV1, X, &error);
  if (error != 0) {
    fprintf (stderr, " ERROR during solution: %d \n", error);
    exit (3);
  }
  fprintf (stderr, " solve completed \n");
  
  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SV1[iDof] = X[iDof];
  }
}

