/* autolinas_dss_upperrow.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dss_inside.h"
#include "autolinas_dss_upperrow_inside.h"



/* symmetric matrix */
/* direct sparse solver */
/* sparse matrix, non-zero component storage, upper triangle */

/*

  # * - - * - - - - -        #**
  * # * * - - - - - -        #**
  - * # - * - - - - -        #*
  - * - # - * * - - -        #**
  * - * - # * - * * -   --\  #***
  - - - * * # - - - -   --/  #
  - - - * - - # - * *        #**
  - - - - * - - # * *        #**
  - - - - * - * * # -        #
  - - - - - - * * - #        #

  symmetric, lower triangle (non-zero component storage)

*/



/* module variables */

/* node ID */
static int System_nIndexTmps[Node_maxNodes];
static int System_indexTmp[Node_maxNodes][System_maxNodesOfOneRow];
static int System_nMatrixSize;

int System_minIndexs[System_maxDofs];
int System_maxIndexs[System_maxDofs];
int System_index[System_maxDofs * System_nAverageDofsOfOneRow];
double SM[System_maxDofs * System_nAverageDofsOfOneRow];



void AutoLinAS_INSIDE_InitializeMatrix (void)
{
  int iNode;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    System_nIndexTmps[iNode] = 0; 
  }
}

void AutoLinAS_RegisterNodalIndex (int iNode, int jNode)
{
  int k;
  
  if (iNode <= jNode) {
    int found = 0;
    
    for (k = 0; k < System_nIndexTmps[iNode]; k++) {
      if (System_indexTmp[iNode][k] == jNode) { 
	found = 1;  break; 
      }
    }
    if (!found) {
      assert(System_nIndexTmps[iNode] < System_maxNodesOfOneRow);
      
      System_indexTmp[iNode][System_nIndexTmps[iNode]] = jNode;
      System_nIndexTmps[iNode]++;
    }
  }
}

static void SortIndex (int iNode)
{
  static int dest[System_maxNodesOfOneRow];
  int i, j;
  
  for (i = 0; i < System_nIndexTmps[iNode]; i++) {
    int value = System_indexTmp[iNode][i];
    int value_index = i;

    for (j = 0; j < i; j++) {
      if (value < dest[j]) {
	value_index = j;
	break;
      }
    }
    for (j = i - 1; value_index <= j; j--) {
      dest[j + 1] = dest[j];
    }
    dest[value_index] = value;
  }
  for (i = 0; i < System_nIndexTmps[iNode]; i++) {
    System_indexTmp[iNode][i] = dest[i];
  }
}

static void CompressIndex ()
{
  int iNode;
  int k;
  int iDof_Node, jDof_Node;
  int count;
  
  count = 0;
  for (iNode = 0; iNode < System_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;

      System_minIndexs[iDof] = count;

      for (k = 0; k < System_nIndexTmps[iNode]; k++) {
	int jNode = System_indexTmp[iNode][k]; 
	
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  int jDof = jNode * Node_nDofs + jDof_Node;

	  if (iDof <= jDof) {
	    System_index[count] = jDof;
	    count++;
	  }
	}
      }
      
      System_maxIndexs[iDof] = count;
    }
  }
}

void AutoLinAS_AllocateMatrix (void)
{
  int iNode;
  int maxIndex, meanIndex;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    SortIndex (iNode); 
  }
  CompressIndex ();

  System_nMatrixSize = 0;
  maxIndex = 0;
  for (iNode = 0; iNode < System_nNodes; iNode++) {
    if (maxIndex < System_nIndexTmps[iNode]) {
      maxIndex = System_nIndexTmps[iNode];
    }
    System_nMatrixSize += System_nIndexTmps[iNode]; 
  }
  meanIndex = System_nMatrixSize / System_nNodes;
  System_nMatrixSize *= Node_nDofs * Node_nDofs;
  assert(System_nMatrixSize <= System_maxDofs * System_nAverageDofsOfOneRow);

#ifdef MEASURE
  fprintf (stderr, " nDofs %d \n", System_nDofs);
  fprintf (stderr, " maxIndex/node %d \n", maxIndex);
  fprintf (stderr, " meanIndex/node %d \n", meanIndex);
  fprintf (stderr, " NonZeroMatrix size %d \n", System_nMatrixSize);
#endif
}

void AutoLinAS_ClearMatrix (void)
{
  int i;

  for (i = 0; i < System_maxIndexs[System_nDofs - 1]; i++) {
    SM[i] = 0.0;
  }
}

void AutoLinAS_AssembleNodalMatrix 
(int iNode, int jNode,
 double matrix[Node_nDofs][Node_nDofs])
{
  int k;
  int iDof_Node, jDof_Node;
  
  for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
    int iDof = iNode * Node_nDofs + iDof_Node;
    
    for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
      int jDof = jNode * Node_nDofs + jDof_Node;
      
      if (iDof <= jDof) {
	for (k = System_minIndexs[iDof]; 
	     k < System_maxIndexs[iDof]; 
	     k++) {
	  if (System_index[k] == jDof) {
	    SM[k] += matrix[iDof_Node][jDof_Node];
	    break;
	  }
	}
      }
    }
  }      
}

void AutoLinAS_INSIDE_SetDirichletCondition (void)
{
  int iDof;
  int k;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVModified[iDof] = 0.0;
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {

      for (k = System_minIndexs[iDof]; 
	   k < System_maxIndexs[iDof]; 
	   k++) {
	int jDof = System_index[k];

	assert(iDof <= jDof);
	SVModified[jDof] -= SM[k] * SVFixed[iDof];
      }
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {

    for (k = System_minIndexs[iDof]; 
	 k < System_maxIndexs[iDof]; 
	 k++) {
      int jDof = System_index[k];

      if (SVFixedFlag[jDof]) {
	if (iDof < jDof) {
	  SVModified[iDof] -= SM[k] * SVFixed[jDof];
	}
      }
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {

    for (k = System_minIndexs[iDof]; 
	 k < System_maxIndexs[iDof]; 
	 k++) {
      int jDof = System_index[k];

      if (SVFixedFlag[jDof]) {
	SM[k] = 0.0;
      }
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {

      for (k = System_minIndexs[iDof]; 
	   k < System_maxIndexs[iDof]; 
	   k++) {
	SM[k] = 0.0;
      }
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {

      for (k = System_minIndexs[iDof]; 
	   k < System_maxIndexs[iDof]; 
	   k++) {
	int jDof = System_index[k];
	
	if (jDof == iDof) {
	  SM[k] = 1.0;
	}
      }
    }
  }
}



