/* autolinas_dscg_nonzero_vectorized.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dscg_inside.h"



/* symmetric matrix */
/* iterative method, conjugate gradient */
/* diagonal scaling preconditioner */
/* sparse matrix, non-zero component storage */

/* column order, tuned using DOF x DOF loop unrolling */

/*

  # * - - * - - - - -        #**
  * # * * - - - - - -        *#**
  - * # - * - - - - -        *#*
  - * - # - * * - - -        *#**
  * - * - # * - * * -   --\  **#***
  - - - * * # - - - -   --/  **#
  - - - * - - # - * *        *#**
  - - - - * - - # * *        *#**
  - - - - * - * * # -        ***#
  - - - - - - * * - #        **#

  full matrix (non-zero component storage)

*/



/* module variables */

double SVModified[System_maxDofs];

double SVScale[System_maxDofs];

#if (Space_nDimensions == 2)

#if (Element_nNodes == 3) 
#define System_maxNodesOfOneRow 21
#endif

#if (Element_nNodes == 4) 
#define System_maxNodesOfOneRow 21
#endif

#endif

#if (Space_nDimensions == 3)

#if (Element_nNodes == 4) 
#define System_maxNodesOfOneRow 21
#endif

#if (Element_nNodes == 8) 
#define System_maxNodesOfOneRow 21
#endif

#if (Element_nNodes == 10) 
/* NEED TO REDEFINE */
ERROR
#endif

#if (Element_nNodes == 20) 
/* NEED TO REDEFINE */
ERROR
#endif

#endif

static int System_nIndexTmps[Node_maxNodes];
static int System_index[System_maxNodesOfOneRow][Node_maxNodes];
static int System_nMatrixSize;
static int System_maxIndex;

/* ***TUNED*** */
#define System_maxColumns (System_maxNodesOfOneRow * Node_nDofs * Node_nDofs) 
static double SM[System_maxColumns][Node_maxNodes];



void AutoLinAS_INSIDE_InitializeMatrix (void)
{
  int iNode;
  int i;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    System_nIndexTmps[iNode] = 0; 

    for (i = 0; i < System_maxNodesOfOneRow; i++) {
      System_index[i][iNode] = iNode; 
    }
  }
}

void AutoLinAS_RegisterNodalIndex (int iNode, int jNode)
{
  int found = 0;
  int k;

  for (k = 0; k < System_nIndexTmps[iNode]; k++) {
    if (System_index[k][iNode] == jNode) { 
      found = 1;  break; 
    }
  }
  if (!found) {
    assert(System_nIndexTmps[iNode] < System_maxNodesOfOneRow);
    
    System_index[System_nIndexTmps[iNode]][iNode] = jNode;
    System_nIndexTmps[iNode]++;
  }
}

void AutoLinAS_AllocateMatrix (void)
{
  int iNode;
  int meanIndex;

  System_nMatrixSize = 0;
  System_maxIndex = 0;
  for (iNode = 0; iNode < System_nNodes; iNode++) {
    if (System_maxIndex < System_nIndexTmps[iNode]) {
      System_maxIndex = System_nIndexTmps[iNode];
    }
    System_nMatrixSize += System_nIndexTmps[iNode]; 
  }
  meanIndex = System_nMatrixSize / System_nNodes;
  System_nMatrixSize *= Node_nDofs * Node_nDofs;

#ifdef MEASURE
  fprintf (stderr, " nDofs %d \n", System_nDofs);
  fprintf (stderr, " maxIndex/node %d \n", System_maxIndex);
  fprintf (stderr, " meanIndex/node %d \n", meanIndex);
  fprintf (stderr, " NonZeroMatrix size %d \n", System_nMatrixSize);
#endif
}

void AutoLinAS_ClearMatrix (void)
{
  int iNode;
  int i;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (i = 0; i < System_maxColumns; i++) {
      SM[i][iNode] = 0.0;
    }
  }
}

void AutoLinAS_AssembleNodalMatrix 
(int iNode, int jNode,
 double matrix[Node_nDofs][Node_nDofs])
{
  int k;
  int iDof_Node, jDof_Node;
  
  for (k = 0; k < System_nIndexTmps[iNode]; k++) {
    if (System_index[k][iNode] == jNode) {
      
      for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  SM[k * Node_nDofs * Node_nDofs 
	     + iDof_Node * Node_nDofs 
	     + jDof_Node][iNode]
	    += matrix[iDof_Node][jDof_Node];
	}
      }
      break;
    }
  }
}

void AutoLinAS_INSIDE_SetDirichletCondition (void)
{
  int iNode;
  int k;
  int iDof_Node, jDof_Node;

  {
    int iDof;

    for (iDof = 0; iDof < System_nDofs; iDof++) {
      SVModified[iDof] = 0.0;
    }
  }

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;

      if (SVFixedFlag[iDof]) {
	
	for (k = 0; k < System_nIndexTmps[iNode]; k++) {
	  int jNode = System_index[k][iNode];
	  
	  for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	    int jDof = jNode * Node_nDofs + jDof_Node;
	    
	    SVModified[jDof] 
	      -= SM[k * Node_nDofs * Node_nDofs 
		    + iDof_Node * Node_nDofs
		    + jDof_Node][iNode]
	      * SVFixed[iDof];
	  }
	}
      }
    }
  }

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      
      for (k = 0; k < System_nIndexTmps[iNode]; k++) {
	int jNode = System_index[k][iNode];
	
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  int jDof = jNode * Node_nDofs + jDof_Node;
	  
	  if (SVFixedFlag[jDof]) {
	    SM[k * Node_nDofs * Node_nDofs
	       + iDof_Node * Node_nDofs
	       + jDof_Node][iNode] = 0.0;
	  }
	}
      }
    }
  }

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;
      
      if (SVFixedFlag[iDof]) {
	
	for (k = 0; k < System_nIndexTmps[iNode]; k++) {
	  
	  for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	    SM[k * Node_nDofs * Node_nDofs 
	       + iDof_Node * Node_nDofs 
	       + jDof_Node][iNode] = 0.0;
	  }
	}
      }
    }
  }

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;

      if (SVFixedFlag[iDof]) {
	
	for (k = 0; k < System_nIndexTmps[iNode]; k++) {
	  int jNode = System_index[k][iNode];
	  
	  for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	    int jDof = jNode * Node_nDofs + jDof_Node;
	    
	    if (jDof == iDof) {
	      SM[k * Node_nDofs * Node_nDofs
		 + iDof_Node * Node_nDofs
		 + jDof_Node][iNode] = 1.0;
	    }
	  }
	}
      }
    }
  }
}

static double GetValue (int iDof, int jDof)
{
  int iNode, jNode;
  int iDof_Node, jDof_Node;
  int k;

  iNode = iDof / Node_nDofs;
  iDof_Node = iDof % Node_nDofs;

  jNode = jDof / Node_nDofs;
  jDof_Node = jDof % Node_nDofs;

  for (k = 0; k < System_nIndexTmps[iNode]; k++) {
    if (System_index[k][iNode] == jNode) {
      return SM[k * Node_nDofs * Node_nDofs 
		+ iDof_Node * Node_nDofs
		+ jDof_Node][iNode];
    }
  }
  return 0.0;
}

void AutoLinAS_INSIDE_MakeScalingVector (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    double value = GetValue (iDof, iDof);

    SVScale[iDof] = sqrt (value);
  }
}

void AutoLinAS_INSIDE_ScaleMatrix (void)
{
  int iNode;
  int k;
  int iDof_Node, jDof_Node;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;

      for (k = 0; k < System_nIndexTmps[iNode]; k++) {
	int jNode = System_index[k][iNode];
	
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  int jDof = jNode * Node_nDofs + jDof_Node;
	  
	  SM[k * Node_nDofs * Node_nDofs 
	     + iDof_Node * Node_nDofs
	     + jDof_Node][iNode] 
	    /= SVScale[iDof] * SVScale[jDof];
	}
      }
    }
  }
}

static void SolveConjugateGradientWithDiagonalScaling (void)
{
  static double x[System_maxDofs];
  static double b[System_maxDofs];
  static double Kx[System_maxDofs];
  static double r[System_maxDofs];
  static double p[System_maxDofs];
  static double Kp[System_maxDofs];
  double bb;
  double rr_old;
  double rr;
  int iDof;
#if (Node_nDofs == 1)
  static double sv0_j0[Node_maxNodes];
#endif
#if (Node_nDofs == 2)
  static double sv0_j0[Node_maxNodes];
  static double sv0_j1[Node_maxNodes];
#endif
#if (Node_nDofs == 3)
  static double sv0_j0[Node_maxNodes];
  static double sv0_j1[Node_maxNodes];
  static double sv0_j2[Node_maxNodes];
#endif
  int iNode;
  int k;
#ifdef MEASURE
  int iterationCount = 0;
#endif

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    b[iDof] = SV1[iDof];
    x[iDof] = SVInitial[iDof];
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    b[iDof] /= SVScale[iDof];  
    x[iDof] *= SVScale[iDof];
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    Kx[iDof] = 0.0;
  }
  
  for (k = 0; k < System_maxIndex; k++) {
    
    for (iNode = 0; iNode < System_nNodes; iNode++) {
      int jNode = System_index[k][iNode];
      
#if (Node_nDofs == 1)
      sv0_j0[iNode] = x[jNode];
#endif
#if (Node_nDofs == 2)
      sv0_j0[iNode] = x[jNode * 2 + 0];
      sv0_j1[iNode] = x[jNode * 2 + 1];
#endif
#if (Node_nDofs == 3)
      sv0_j0[iNode] = x[jNode * 3 + 0];
      sv0_j1[iNode] = x[jNode * 3 + 1];
      sv0_j2[iNode] = x[jNode * 3 + 2];
#endif
    }
    
    for (iNode = 0; iNode < System_nNodes; iNode++) {
#if (Node_nDofs == 1)
      Kx[iNode] += SM[k][iNode] * sv0_j0[iNode];
#endif
#if (Node_nDofs == 2)
      Kx[iNode * 2 + 0] 
	+= SM[k * 4 + 0][iNode] * sv0_j0[iNode]
	+ SM[k * 4 + 1][iNode] * sv0_j1[iNode];
      
      Kx[iNode * 2 + 1] 
	+= SM[k * 4 + 2][iNode] * sv0_j0[iNode]
	+ SM[k * 4 + 3][iNode] * sv0_j1[iNode];
#endif
#if (Node_nDofs == 3)
      Kx[iNode * 3 + 0] 
	+= SM[k * 9 + 0][iNode] * sv0_j0[iNode]
	+ SM[k * 9 + 1][iNode] * sv0_j1[iNode]
	+ SM[k * 9 + 2][iNode] * sv0_j2[iNode];
      
      Kx[iNode * 3 + 1] 
	+= SM[k * 9 + 3][iNode] * sv0_j0[iNode]
	+ SM[k * 9 + 4][iNode] * sv0_j1[iNode]
	+ SM[k * 9 + 5][iNode] * sv0_j2[iNode];
      
      Kx[iNode * 3 + 2] 
	+= SM[k * 9 + 6][iNode] * sv0_j0[iNode]
	+ SM[k * 9 + 7][iNode] * sv0_j1[iNode]
	+ SM[k * 9 + 8][iNode] * sv0_j2[iNode];
#endif
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    r[iDof] = b[iDof] - Kx[iDof];
    p[iDof] = r[iDof];
  }

  bb = 0.0;
  for (iDof = 0; iDof < System_nDofs; iDof++) {
    bb += b[iDof] * b[iDof];
  }

  rr_old = 0.0;
  for (iDof = 0; iDof < System_nDofs; iDof++) {
    rr_old += r[iDof] * r[iDof];
  }

  rr = 0.0;

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nDofs * 3);
  AutoLinAS_AddOp (M_MUL_OP, System_nDofs * 3);
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 11);
#endif

  do {
    double rp;
    double pKp;
    double alpha;
    double beta;
    
    for (iDof = 0; iDof < System_nDofs; iDof++) {
      Kp[iDof] = 0.0;
    }
    
    for (k = 0; k < System_maxIndex; k++) {
      
      for (iNode = 0; iNode < System_nNodes; iNode++) {
	int jNode = System_index[k][iNode];
	
#if (Node_nDofs == 1)
	sv0_j0[iNode] = p[jNode];
#endif
#if (Node_nDofs == 2)
	sv0_j0[iNode] = p[jNode * 2 + 0];
	sv0_j1[iNode] = p[jNode * 2 + 1];
#endif
#if (Node_nDofs == 3)
	sv0_j0[iNode] = p[jNode * 3 + 0];
	sv0_j1[iNode] = p[jNode * 3 + 1];
	sv0_j2[iNode] = p[jNode * 3 + 2];
#endif
      }

      for (iNode = 0; iNode < System_nNodes; iNode++) {
#if (Node_nDofs == 1)
	Kp[iNode] += SM[k][iNode] * sv0_j0[iNode];
#endif
#if (Node_nDofs == 2)
	Kp[iNode * 2 + 0] 
	  += SM[k * 4 + 0][iNode] * sv0_j0[iNode]
	  + SM[k * 4 + 1][iNode] * sv0_j1[iNode];
	
	Kp[iNode * 2 + 1] 
	  += SM[k * 4 + 2][iNode] * sv0_j0[iNode]
	  + SM[k * 4 + 3][iNode] * sv0_j1[iNode];
#endif
#if (Node_nDofs == 3)
	Kp[iNode * 3 + 0] 
	  += SM[k * 9 + 0][iNode] * sv0_j0[iNode]
	  + SM[k * 9 + 1][iNode] * sv0_j1[iNode]
	  + SM[k * 9 + 2][iNode] * sv0_j2[iNode];
	
	Kp[iNode * 3 + 1] 
	  += SM[k * 9 + 3][iNode] * sv0_j0[iNode]
	  + SM[k * 9 + 4][iNode] * sv0_j1[iNode]
	  + SM[k * 9 + 5][iNode] * sv0_j2[iNode];
	
	Kp[iNode * 3 + 2] 
	  += SM[k * 9 + 6][iNode] * sv0_j0[iNode]
	  + SM[k * 9 + 7][iNode] * sv0_j1[iNode]
	  + SM[k * 9 + 8][iNode] * sv0_j2[iNode];
#endif
      }
    }

    rp = 0.0;
    for (iDof = 0; iDof < System_nDofs; iDof++) {
      rp += r[iDof] * p[iDof];
    }

    pKp = 0.0;
    for (iDof = 0; iDof < System_nDofs; iDof++) {
      pKp += p[iDof] * Kp[iDof];
    }

    alpha = rp / pKp;

    for (iDof = 0; iDof < System_nDofs; iDof++) {
      x[iDof] += p[iDof] * alpha;
      r[iDof] -= Kp[iDof] * alpha;
    }

    rr = 0.0;
    for (iDof = 0; iDof < System_nDofs; iDof++) {
      rr += r[iDof] * r[iDof];
    }

    beta = rr / rr_old;
    rr_old = rr;

    for (iDof = 0; iDof < System_nDofs; iDof++) {
      p[iDof] = r[iDof] + p[iDof] * beta;
    }

#ifdef MEASURE
    AutoLinAS_AddOp 
      (M_ADD_OP, 
       System_nNodes * System_maxIndex * Node_nDofs * Node_nDofs);
    AutoLinAS_AddOp 
      (M_MUL_OP, 
       System_nNodes * System_maxIndex * Node_nDofs * Node_nDofs);
    AutoLinAS_AddOp 
      (M_ADD_OP, 
       System_nNodes * Node_nDofs);
    
    AutoLinAS_AddOp 
      (M_MEMORY, 
       System_nNodes * System_maxIndex / (sizeof (double) / sizeof (int)));
    AutoLinAS_AddIO 
      (M_MEMORY, 
       System_nNodes * System_maxIndex * (Node_nDofs * Node_nDofs + Node_nDofs));
    AutoLinAS_AddIO 
      (M_MEMORY, 
       System_nNodes * Node_nDofs * 2);

    AutoLinAS_AddOp (M_ADD_OP, System_nDofs * 6);
    AutoLinAS_AddOp (M_MUL_OP, System_nDofs * 6);
    AutoLinAS_AddIO (M_MEMORY, System_nDofs * 12);
#endif

#ifdef MEASURE
    iterationCount++;
    if (iterationCount % 1000 == 0) {
      fprintf (stderr, " iteration %d \n", iterationCount);
    }
#endif

  } while (sqrt (bb / System_nDofs) * 0.000001 
	   < sqrt (rr / System_nDofs));

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    x[iDof] /= SVScale[iDof];
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SV1[iDof] = x[iDof];
  }

#ifdef MEASURE
  fprintf (stderr, " total iteration %d \n", iterationCount);
#endif

#ifdef MEASURE
  AutoLinAS_AddOp (M_OTHER_OP, System_nDofs);
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 4);
#endif
}

void AutoLinAS_Initialize 
(int nNodes)
{
  AutoLinAS_INSIDE_InitializeSystem (nNodes);
  AutoLinAS_INSIDE_InitializeMatrix ();
}

void AutoLinAS_Fix (void)
{
  AutoLinAS_INSIDE_SetDirichletCondition ();
  AutoLinAS_INSIDE_MakeScalingVector ();
  AutoLinAS_INSIDE_ScaleMatrix ();
}

void AutoLinAS_Solve (void)
{
  AutoLinAS_ClearInitialResult ();
  AutoLinAS_SolveUsingInitialResult ();
}

void AutoLinAS_SolveUsingInitialResult (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SV1[iDof] = SV0[iDof];
    SV1[iDof] += SVModified[iDof];
    if (SVFixedFlag[iDof]) {
      SV1[iDof] = SVFixed[iDof];
    }
  }

#ifdef MEASURE_TIME
  AutoLinAS_InitializeMeasure ();
#endif

  SolveConjugateGradientWithDiagonalScaling ();

#ifdef MEASURE_TIME
  AutoLinAS_TerminateMeasure ();
#endif
}

