/* autolinas_ldl_skyline.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_ldl_inside.h"
#include "autolinas_ldl_skyline_inside.h"



/*

  *
  * *
  - * *
  - - * *
  - - - * *
  - - - * * *
  - - - * * * *
  - - - - * * * *
  - - - - - - * * *
  - - - - - - * * * *

  symmetric, lower triangle (skyline storage)

*/



/* module variables */

int System_nMatrixSize;
int System_minIndexs[System_maxDofs];
int System_ij[System_maxDofs];

#ifdef DYNAMIC_MEMORY_ALLOCATION
/* using dynamic memory allocation for skyline matrix */
double *SM;
#else
double SM[System_maxMatrixSize];
#endif

/* 
   SM[System_ij[iDof] + jDof];
   if (System_minIndexs[iDof] <= jDof && jDof <= iDof)
*/



void AutoLinAS_INSIDE_InitializeMatrix (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    System_minIndexs[iDof] = iDof; 
  }
}

void AutoLinAS_RegisterNodalIndex (int iNode, int jNode)
{
  int iDof_Node;
  
  for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
    int iDof = iNode * Node_nDofs + iDof_Node;
    
    if (jNode * Node_nDofs < System_minIndexs[iDof]) {
      System_minIndexs[iDof] = jNode * Node_nDofs;
    }
  }
}

void AutoLinAS_AllocateMatrix (void)
{
  int iDof;

  System_ij[0] = 0;
  for (iDof = 1; iDof < System_nDofs; iDof++) {
    System_ij[iDof] 
      = System_ij[iDof - 1] + iDof - System_minIndexs[iDof];
  }
  System_nMatrixSize = System_ij[System_nDofs - 1] + System_nDofs;
  assert(System_nMatrixSize <= System_maxMatrixSize);

#ifdef MEASURE
  fprintf (stderr, " nDofs %d \n", System_nDofs);
  fprintf (stderr, " mean bandwidth %d \n", 
	   System_nMatrixSize / System_nDofs);
  fprintf (stderr, " SkylineMatrix size %d \n", 
	   System_nMatrixSize);
#endif
}

void AutoLinAS_ClearMatrix (void)
{
  int i;

#ifdef DYNAMIC_MEMORY_ALLOCATION
  /* using dynamic memory allocation for skyline matrix */
  if (SM == NULL) {
    SM = (double *)malloc (sizeof (double) * System_maxMatrixSize);
    if (SM == NULL) {
      fprintf 
	(stderr, 
	 " cannot allocate memory for SM (size %d) \n",
	 System_maxMatrixSize);
      exit (1);
    }
  }
#endif

  for (i = 0; i < System_nMatrixSize; i++) {
    SM[i] = 0.0;
  }
}

void AutoLinAS_AssembleNodalMatrix 
(int iNode, int jNode,
 double matrix[Node_nDofs][Node_nDofs])
{
  int iDof_Node, jDof_Node;
  
  for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
    int iDof = iNode * Node_nDofs + iDof_Node;
    
    for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
      int jDof = jNode * Node_nDofs + jDof_Node;
      
      if (jDof <= iDof) {
	SM[System_ij[iDof] + jDof] 
	  += matrix[iDof_Node][jDof_Node];
      }
    }
  }
}

void AutoLinAS_INSIDE_SetDirichletCondition (void)
{
  int iDof;
  
  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVModified[iDof] = 0.0;
  }
  
  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {
      double value = SVFixed[iDof];
      double coefficient = SM[System_ij[iDof] + iDof] * PENALTY_NUMBER;
      
      SM[System_ij[iDof] + iDof] = coefficient;
      SVModified[iDof] = value * coefficient;
    }
  }
}

