/* autolinas_measure.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <sys/time.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"



/* module variables */

static int SecondAtInitialization;

static double InitializedTime;
static double TotalOpCount;
static double TotalIOCount;

static double StartTime;
static double AddOpCount;
static double MulOpCount;
static double OtherOpCount;
static double MemoryIOCount;
static double CacheIOCount;
static double DiskIOCount;

#define MAX_TIME_COUNTER 8
static double TimeCounters[MAX_TIME_COUNTER];



static void InitializeTime (void)
{
  struct timeval tv;
  struct timezone tz;

  gettimeofday (&tv, &tz);
  SecondAtInitialization = tv.tv_sec;
}

static double Now (void)
{
  struct timeval tv;
  struct timezone tz;

  gettimeofday (&tv, &tz);
  return (double)(tv.tv_sec - SecondAtInitialization) 
    + tv.tv_usec * 0.000001;
}

void AutoLinAS_InitializeMeasure (void)
{
  InitializeTime ();

  InitializedTime = Now ();
  TotalOpCount = 0.0;
  TotalIOCount = 0.0;

  fprintf (stderr, " START TIME \n");
}

void AutoLinAS_TerminateMeasure (void)
{
  double totalTime;
  
  totalTime = Now () - InitializedTime;

  fprintf (stderr, " END TIME \n");
  
  fprintf (stderr, " time. = %f \n", 
	   totalTime);
  fprintf (stderr, " number of FP operations = %f \n",
	   TotalOpCount);
  fprintf (stderr, " number of I/O operations = %f \n", 
	   TotalIOCount);
  fprintf (stderr, " FP = %f M Flops \n", 
	   TotalOpCount / totalTime / 1000000);
  fprintf (stderr, " I/O = %f M op./ sec \n", 
	   TotalIOCount / totalTime / 1000000);
}

void AutoLinAS_ClearTime (void)
{
  StartTime = Now ();
}

double AutoLinAS_GetTime (void)
{
  return Now () - StartTime;
}

void AutoLinAS_CountTime (int counterId)
{
  assert(0 <= counterId);
  assert(counterId < MAX_TIME_COUNTER);

  TimeCounters[counterId] += Now () - StartTime;
}

void AutoLinAS_DivideTimeCounters (double value)
{
  int iCounter;

  for (iCounter = 0; iCounter < MAX_TIME_COUNTER; iCounter++) {
    TimeCounters[iCounter] /= value;
  }
}

void AutoLinAS_PrintTimeCounters (void)
{
  int iCounter;

  for (iCounter = 0; iCounter < MAX_TIME_COUNTER; iCounter++) {
    fprintf (stderr, 
	     " time counter %d %f seconds \n", 
	     iCounter, TimeCounters[iCounter]);
  }
}

void AutoLinAS_ClearOp (void)
{
  AddOpCount = 0.0;
  MulOpCount = 0.0;
  OtherOpCount = 0.0;
}

void AutoLinAS_AddOp (int opType, double count)
{
  switch (opType) {
  case M_ADD_OP: AddOpCount += count;    break;
  case M_MUL_OP: MulOpCount += count;    break;
  case M_OTHER_OP: OtherOpCount += count;    break;
  }
  TotalOpCount += count;
}

double AutoLinAS_GetOp (int opType)
{
  switch (opType) {
  case M_ADD_OP: return AddOpCount;
  case M_MUL_OP: return MulOpCount;
  case M_OTHER_OP: return OtherOpCount;
  }
  return 0.0;
}

void AutoLinAS_ClearIO (void)
{
  MemoryIOCount = 0.0;
  CacheIOCount = 0.0;
  DiskIOCount = 0.0;
}

void AutoLinAS_AddIO (int level, double count)
{
  switch (level) {
  case M_MEMORY: MemoryIOCount += count;  break;
  case M_CACHE: CacheIOCount += count;  break;
  case M_DISK: DiskIOCount += count;  break;
  }
  TotalIOCount += count;
}

double AutoLinAS_GetIO (int level)
{
  switch (level) {
  case M_MEMORY: return MemoryIOCount;
  case M_CACHE: return CacheIOCount;
  case M_DISK: return DiskIOCount;
  }
  return 0.0;
}

