/* post_es_2d_tri.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>

#include "../lib/autolinas_predefined.h"



/* Node */

/* 2D plane strain, linear elastic */

int Node_nNodes;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* displacement analysis result */
double Node_uVec[Node_maxNodes][Node_nDofs];

/* stress analysis result */
double Node_sigma[Node_maxNodes];



/* Element */

/* 2D plane strain, linear elastic */

int Element_nElements;

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



void ReadMesh (FILE* fp)
{
  int iElement;
  int iNode_Elem;
  int iNode;
  
  fscanf (fp, "%d", &Element_nElements);
  fprintf (stderr, " nElements %d \n", Element_nElements);
  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d ", 
	      &Element_nodeIds[iElement][iNode_Elem]);
    }
  }

  fscanf (fp, "%d", &Node_nNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf %lf", 
	    &Node_XVec[iNode][0],
	    &Node_XVec[iNode][1]);
  }
}

void ReadResult (FILE* fp)
{
  int iNode;

  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf %lf", 
	    &Node_uVec[iNode][0], &Node_uVec[iNode][1]);
  }
}

void ReadData (int argc, char* argv[])
{
  FILE* fp;

  if (argc < 3) {
    fprintf (stderr, 
	     " post_es_2d_tri input_file_name output_file_name \n");
    exit (1);
  }
  
  fp = fopen (argv[1], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open input file  \n");
    exit (1);
  }
  ReadMesh (fp);
  fclose (fp);

  fp = fopen (argv[2], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open output file  \n");
    exit (1);
  }
  ReadResult (fp);
  fclose (fp);
}



double MinRange = 0.0;
double MaxRange = 1000.0;

double Scale = 3.0;

int MeshIsShown = 1;

#define PLOT_NONE 0
#define PLOT_CONTOUR 1
#define PLOT_LINE_CONTOUR 2
#define PLOT_DEFORMATION 3

int PlotType = PLOT_NONE;



void PlotMesh (void) 
{
  int iElement;

  AutoGL_SetColor (1, 1, 1);
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    int nodeId1 = Element_nodeIds[iElement][1];
    int nodeId2 = Element_nodeIds[iElement][2];
    AutoGL_DrawLine (Node_XVec[nodeId0][0], Node_XVec[nodeId0][1], 0,
		     Node_XVec[nodeId1][0], Node_XVec[nodeId1][1], 0);
    AutoGL_DrawLine (Node_XVec[nodeId1][0], Node_XVec[nodeId1][1], 0,
		     Node_XVec[nodeId2][0], Node_XVec[nodeId2][1], 0);
    AutoGL_DrawLine (Node_XVec[nodeId2][0], Node_XVec[nodeId2][1], 0,
		     Node_XVec[nodeId0][0], Node_XVec[nodeId0][1], 0);
  }
}

void PlotContour (void) 
{
  int iElement;

  AutoGL_DrawContourMap (80, MinRange, MaxRange);

  AutoGL_SetColor (1, 1, 1);
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    int nodeId1 = Element_nodeIds[iElement][1];
    int nodeId2 = Element_nodeIds[iElement][2];
    double grades[3];

    grades[0] = (Node_sigma[nodeId0] - MinRange) / (MaxRange - MinRange);
    grades[1] = (Node_sigma[nodeId1] - MinRange) / (MaxRange - MinRange);
    grades[2] = (Node_sigma[nodeId2] - MinRange) / (MaxRange - MinRange);

    AutoGL_DrawContourTriangle 
      (Node_XVec[nodeId0][0], Node_XVec[nodeId0][1], 0, grades[0],
       Node_XVec[nodeId1][0], Node_XVec[nodeId1][1], 0, grades[1], 
       Node_XVec[nodeId2][0], Node_XVec[nodeId2][1], 0, grades[2]);
  }
}

void PlotLineContour (void) 
{
  int nGrades = 20;
  int iGrade;

  AutoGL_DrawLineContourMap (80, MinRange, MaxRange, nGrades);

  for (iGrade = 1; iGrade < nGrades; iGrade++) {
    double grade = 1.0 / nGrades * iGrade;
    double levelValue = MinRange + (MaxRange - MinRange) * grade;
    double red, green, blue;
    int iElement;

    AutoGL_GetContourColor (&red, &green, &blue,
			    grade);
    AutoGL_SetColor (red, green, blue);
  
    for (iElement = 0; iElement < Element_nElements; iElement++) {
      int nodeId0 = Element_nodeIds[iElement][0];
      int nodeId1 = Element_nodeIds[iElement][1];
      int nodeId2 = Element_nodeIds[iElement][2];
      double triPoints[3][3];
      double triValues[3];

      triPoints[0][0] = Node_XVec[nodeId0][0];
      triPoints[0][1] = Node_XVec[nodeId0][1];
      triPoints[0][2] = 0.0;
      triPoints[1][0] = Node_XVec[nodeId1][0];
      triPoints[1][1] = Node_XVec[nodeId1][1];
      triPoints[1][2] = 0.0;
      triPoints[2][0] = Node_XVec[nodeId2][0];
      triPoints[2][1] = Node_XVec[nodeId2][1];
      triPoints[2][2] = 0.0;

      triValues[0] = Node_sigma[nodeId0];
      triValues[1] = Node_sigma[nodeId1];
      triValues[2] = Node_sigma[nodeId2];
      
      AutoGL_DrawTriangleIsoline
	(levelValue, triPoints, triValues);
    }
  }
}

void PlotDeformation (void) 
{
  int iElement;

  AutoGL_SetColor (0, 1, 1);
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int nodeId0 = Element_nodeIds[iElement][0];
    int nodeId1 = Element_nodeIds[iElement][1];
    int nodeId2 = Element_nodeIds[iElement][2];
    double p0X = Node_XVec[nodeId0][0] + Node_uVec[nodeId0][0] * Scale;
    double p1X = Node_XVec[nodeId1][0] + Node_uVec[nodeId1][0] * Scale;
    double p2X = Node_XVec[nodeId2][0] + Node_uVec[nodeId2][0] * Scale;
    double p0Y = Node_XVec[nodeId0][1] + Node_uVec[nodeId0][1] * Scale;
    double p1Y = Node_XVec[nodeId1][1] + Node_uVec[nodeId1][1] * Scale;
    double p2Y = Node_XVec[nodeId2][1] + Node_uVec[nodeId2][1] * Scale;

    AutoGL_DrawLine (p0X, p0Y, 0,
		     p1X, p1Y, 0);
    AutoGL_DrawLine (p1X, p1Y, 0,
		     p2X, p2Y, 0);
    AutoGL_DrawLine (p2X, p2Y, 0,
		     p0X, p0Y, 0);
  }
}

void RedrawView (void) 
{
  if (MeshIsShown) {
    PlotMesh ();
  }

  switch (PlotType) {
  case PLOT_NONE:
    break;
  case PLOT_CONTOUR:
    PlotContour ();
    break;
  case PLOT_LINE_CONTOUR:
    PlotLineContour ();
    break;
  case PLOT_DEFORMATION:
    PlotDeformation ();
    break;
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();  
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  ReadData (argc, argv);

  AutoGL_SetViewSize (30); 
  AutoGL_SetViewCenter (15, 10, 0); 

  AutoGL_SetViewRedrawCallback (RedrawView);   
  AutoGL_SetDefaultCallbackInMode2D (NULL);

  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode2D ();

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");
  AutoGL_AddBoolean (&MeshIsShown, "MeshIsShown");
  AutoGL_AddInteger (&PlotType, "PlotType");
  AutoGL_AddIntegerItem ("none");      
  AutoGL_AddIntegerItem ("contour"); 
  AutoGL_AddIntegerItem ("line_contour"); 
  AutoGL_AddIntegerItem ("deformation");    
  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");
  AutoGL_AddReal (&Scale, "Scale");
  AutoGL_AddCallback (UpdateModel, "UpdateModel");
}






