/* pre_2d_tri.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../lib/autolinas_predefined.h"



int Node_nNodes;
int Element_nElements;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



#define X_SIZE 1.0
#define Y_SIZE 1.0

#define EPS 0.0001

int XDivision;
int YDivision;

void MakeMesh (void)
{
  int iNode;
  int iElement;
  int x, y;
  double xSize, ySize;
  
  xSize = X_SIZE;
  ySize = Y_SIZE;
  
  iNode = 0;
  for (y = 0; y < YDivision + 1; y++) {
    for (x = 0; x < XDivision + 1; x++) {
      Node_XVec[iNode][0] = xSize * x;
      Node_XVec[iNode][1] = ySize * y;
      iNode++;
    }
  }
  Node_nNodes = iNode;
  
  iElement = 0;
  for (y = 0; y < YDivision; y++) {
    for (x = 0; x < XDivision; x++) {
      Element_nodeIds[iElement][0] = x + y * (XDivision + 1);
      Element_nodeIds[iElement][1] = (x + 1) + (y + 1) * (XDivision + 1);
      Element_nodeIds[iElement][2] = x + (y + 1) * (XDivision + 1);
      iElement++;

      Element_nodeIds[iElement][0] = x + y * (XDivision + 1);
      Element_nodeIds[iElement][1] = (x + 1) + y * (XDivision + 1);
      Element_nodeIds[iElement][2] = (x + 1) + (y + 1) * (XDivision + 1);
      iElement++;
    }
  }
  Element_nElements = iElement;
  
}

void WriteMesh (FILE* fp)
{
  int iElement;
  int iNode;
  
  fprintf (fp, "%6d\n", Element_nElements);
  fprintf (stderr, " nElements %d \n", Element_nElements);
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    fprintf (fp, "%6d %6d %6d\n", 
	     Element_nodeIds[iElement][0],
	     Element_nodeIds[iElement][1],
	     Element_nodeIds[iElement][2]);
  }

  fprintf (fp, "%6d\n", Node_nNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%8.4f %8.4f\n", 
	     Node_XVec[iNode][0],
	     Node_XVec[iNode][1]);
  }
}

int main (int argc, char* argv[])
{
  FILE* fp;

  if (argc < 4) {
    fprintf (stderr, 
	     " pre_2d_tri output_file_name xdiv ydiv \n");
    exit (1);
  }
  
  XDivision = atoi (argv[2]);
  YDivision = atoi (argv[3]);

  MakeMesh ();

  fp = fopen (argv[1], "w");
  if (fp == NULL) {
    fprintf (stderr, " cannot open output file  \n");
    exit (1);
  }

  WriteMesh (fp);

  fclose (fp);

  return 0;
}





