/* pre_3d_tetra.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../lib/autolinas_predefined.h"



int Node_nNodes;
int Element_nElements;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



#define X_SIZE 1.0
#define Y_SIZE 1.0
#define Z_SIZE 1.0

#define EPS 0.0001

int XDivision;
int YDivision;
int ZDivision;

int HexToTetraMap[6][4] = {
  { 0, 1, 3, 5 }, 
  { 0, 3, 7, 5 }, 
  { 0, 4, 5, 7 }, 
  { 0, 3, 2, 7 }, 
  { 0, 2, 4, 7 }, 
  { 2, 4, 7, 6 }
};

int Grid (int x, int y, int z)
{
  return x 
    + y * (XDivision + 1) 
    + z * (XDivision + 1) * (YDivision + 1);
}

void MakeMesh (void)
{
  int iNode;
  int iElement;
  int x, y, z;
  
  iNode = 0;
  for (z = 0; z < ZDivision + 1; z++) {
    for (y = 0; y < YDivision + 1; y++) {
      for (x = 0; x < XDivision + 1; x++) {
	Node_XVec[iNode][0] = X_SIZE * x;
	Node_XVec[iNode][1] = Y_SIZE * y;
	Node_XVec[iNode][2] = Z_SIZE * z;
	iNode++;
      }
    }
  }
  Node_nNodes = iNode;
  
  iElement = 0;
  for (z = 0; z < ZDivision; z++) {
    for (y = 0; y < YDivision; y++) {
      for (x = 0; x < XDivision; x++) {
	int nodeIds[8];
	int i, v;

	nodeIds[0] = Grid (x + 0, y + 0, z + 0);
	nodeIds[1] = Grid (x + 1, y + 0, z + 0);
	nodeIds[2] = Grid (x + 0, y + 1, z + 0);
	nodeIds[3] = Grid (x + 1, y + 1, z + 0);
	nodeIds[4] = Grid (x + 0, y + 0, z + 1);
	nodeIds[5] = Grid (x + 1, y + 0, z + 1);
	nodeIds[6] = Grid (x + 0, y + 1, z + 1);
	nodeIds[7] = Grid (x + 1, y + 1, z + 1);

	for (i = 0; i < 6; i++) {
	  for (v = 0; v < 4; v++) {
	    Element_nodeIds[iElement][v] 
	      = nodeIds[HexToTetraMap[i][v]];
	  }
	  iElement++;
	}
      }
    }
  }
  Element_nElements = iElement;
}

void WriteMesh (FILE* fp)
{
  int iElement;
  int iNode;
  
  fprintf (fp, "%6d\n", Element_nElements);
  fprintf (stderr, " nElements %d \n", Element_nElements);
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    fprintf (fp, "%6d %6d %6d %6d\n", 
	     Element_nodeIds[iElement][0],
	     Element_nodeIds[iElement][1],
	     Element_nodeIds[iElement][2],
	     Element_nodeIds[iElement][3]);
  }

  fprintf (fp, "%6d\n", Node_nNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%8.4f %8.4f %8.4f\n", 
	     Node_XVec[iNode][0],
	     Node_XVec[iNode][1],
	     Node_XVec[iNode][2]);
  }
}

int main (int argc, char* argv[])
{
  FILE* fp;

  if (argc < 5) {
    fprintf (stderr, 
	     " pre_3d_tetra output_file_name xdiv ydiv zdiv \n");
    exit (1);
  }
  
  XDivision = atoi (argv[2]);
  YDivision = atoi (argv[3]);
  ZDivision = atoi (argv[4]);

  MakeMesh ();

  fp = fopen (argv[1], "w");
  if (fp == NULL) {
    fprintf (stderr, " cannot open output file  \n");
    exit (1);
  }

  WriteMesh (fp);

  fclose (fp);

  return 0;
}





