/* pre_3d_tetra_2nd.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../lib/autolinas_predefined.h"



int Node_nNodes;
int Element_nElements;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



#define X_SIZE 1.0
#define Y_SIZE 1.0
#define Z_SIZE 1.0

#define EPS 0.0001

int XDivision;
int YDivision;
int ZDivision;

int HexToTetraMap[6][10] = {
  { 2, 8, 0, 20, 5, 1, 11, 4, 10, 14 }, 
  { 0, 20, 26, 18, 10, 13, 9, 23,22, 19 }, 
  { 0, 8, 26, 20, 4, 13, 10, 17, 23, 14 }, 
  { 0, 26, 6, 18, 13, 3, 9, 16, 12, 22 }, 
  { 0, 26, 8, 6, 13, 4, 3, 17, 7, 16 }, 
  { 24, 26, 18, 6, 25, 21, 15, 22, 12, 16 }
};

int Grid (int x, int y, int z)
{
  return x 
    + y * (XDivision * 2 + 1) 
    + z * (XDivision * 2 + 1) * (YDivision * 2 + 1);
}

void MakeMesh (void)
{
  int iNode;
  int iElement;
  int x, y, z;
  
  iNode = 0;
  for (z = 0; z < ZDivision * 2 + 1; z++) {
    for (y = 0; y < YDivision * 2 + 1; y++) {
      for (x = 0; x < XDivision * 2 + 1; x++) {
	Node_XVec[iNode][0] = X_SIZE / 2 * x;
	Node_XVec[iNode][1] = Y_SIZE / 2 * y;
	Node_XVec[iNode][2] = Z_SIZE / 2 * z;
	iNode++;
      }
    }
  }
  Node_nNodes = iNode;
  
  iElement = 0;
  for (z = 0; z < ZDivision * 2; z += 2) {
    for (y = 0; y < YDivision * 2; y += 2) {
      for (x = 0; x < XDivision * 2; x += 2) {
	int nodeIds[27];
	int i, v;

	nodeIds[ 0] = Grid (x + 0, y + 0, z + 0);
	nodeIds[ 1] = Grid (x + 1, y + 0, z + 0);
	nodeIds[ 2] = Grid (x + 2, y + 0, z + 0);
	nodeIds[ 3] = Grid (x + 0, y + 1, z + 0);
	nodeIds[ 4] = Grid (x + 1, y + 1, z + 0);
	nodeIds[ 5] = Grid (x + 2, y + 1, z + 0);
	nodeIds[ 6] = Grid (x + 0, y + 2, z + 0);
	nodeIds[ 7] = Grid (x + 1, y + 2, z + 0);
	nodeIds[ 8] = Grid (x + 2, y + 2, z + 0);

	nodeIds[ 9] = Grid (x + 0, y + 0, z + 1);
	nodeIds[10] = Grid (x + 1, y + 0, z + 1);
	nodeIds[11] = Grid (x + 2, y + 0, z + 1);
	nodeIds[12] = Grid (x + 0, y + 1, z + 1);
	nodeIds[13] = Grid (x + 1, y + 1, z + 1);
	nodeIds[14] = Grid (x + 2, y + 1, z + 1);
	nodeIds[15] = Grid (x + 0, y + 2, z + 1);
	nodeIds[16] = Grid (x + 1, y + 2, z + 1);
	nodeIds[17] = Grid (x + 2, y + 2, z + 1);

	nodeIds[18] = Grid (x + 0, y + 0, z + 2);
	nodeIds[19] = Grid (x + 1, y + 0, z + 2);
	nodeIds[20] = Grid (x + 2, y + 0, z + 2);
	nodeIds[21] = Grid (x + 0, y + 1, z + 2);
	nodeIds[22] = Grid (x + 1, y + 1, z + 2);
	nodeIds[23] = Grid (x + 2, y + 1, z + 2);
	nodeIds[24] = Grid (x + 0, y + 2, z + 2);
	nodeIds[25] = Grid (x + 1, y + 2, z + 2);
	nodeIds[26] = Grid (x + 2, y + 2, z + 2);

	for (i = 0; i < 6; i++) {
	  for (v = 0; v < 10; v++) {
	    Element_nodeIds[iElement][v] 
	      = nodeIds[HexToTetraMap[i][v]];
	  }
	  iElement++;
	}
      }
    }
  }
  Element_nElements = iElement;
}

void WriteMesh (FILE* fp)
{
  int iElement;
  int iNode;
  
  fprintf (fp, "%6d\n", Element_nElements);
  fprintf (stderr, " nElements %d \n", Element_nElements);
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    fprintf (fp, "%6d %6d %6d %6d %6d %6d %6d %6d %6d %6d\n", 
	     Element_nodeIds[iElement][0],
	     Element_nodeIds[iElement][1],
	     Element_nodeIds[iElement][2],
	     Element_nodeIds[iElement][3],
	     Element_nodeIds[iElement][4],
	     Element_nodeIds[iElement][5],
	     Element_nodeIds[iElement][6],
	     Element_nodeIds[iElement][7],
	     Element_nodeIds[iElement][8],
	     Element_nodeIds[iElement][9]);
  }

  fprintf (fp, "%6d\n", Node_nNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%8.4f %8.4f %8.4f\n", 
	     Node_XVec[iNode][0],
	     Node_XVec[iNode][1],
	     Node_XVec[iNode][2]);
  }
}

int main (int argc, char* argv[])
{
  FILE* fp;

  if (argc < 5) {
    fprintf (stderr, 
	     " pre_3d_tetra_2nd output_file_name xdiv ydiv zdiv \n");
    exit (1);
  }
  
  XDivision = atoi (argv[2]);
  YDivision = atoi (argv[3]);
  ZDivision = atoi (argv[4]);

  MakeMesh ();

  fp = fopen (argv[1], "w");
  if (fp == NULL) {
    fprintf (stderr, " cannot open output file  \n");
    exit (1);
  }

  WriteMesh (fp);

  fclose (fp);

  return 0;
}





