/* autolinas_dscg_em.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dscg_inside.h"



/* symmetric matrix */
/* iterative method, conjugate gradient */
/* diagonal scaling preconditioner */
/* unassembled matrix, element matrix storage */

/*

  # # #
  # # #  -->  # - # - # 
  # # #       - x - x x 
              # - # - # 
  x x x       - x - x x 
  x x x  -->  # x # x #+x
  x x x      

  full matrix (element matrix storage)

*/



/* module variables */

static int System_nMatrixSize;

static double SM[Element_maxElements][Element_nDofs][Element_nDofs];



void AutoLinAS_INSIDE_InitializeMatrix (void)
{
  System_nMatrixSize 
    = System_nElements * Element_nDofs * Element_nDofs;

#ifdef MEASURE
  fprintf (stderr, " nDofs %d \n", System_nDofs);
  fprintf (stderr, " nElements %d \n", System_nElements);
  fprintf (stderr, " element matrix size %d \n", 
	   Element_nDofs * Element_nDofs);
  fprintf (stderr, " EBEMatrix size %d \n", 
	   System_nMatrixSize);
#endif
}

void AutoLinAS_INSIDE_Clear (void)
{
  int iElement;
  int iDof_Elem, jDof_Elem;

  for (iElement = 0; iElement < System_nElements; iElement++) {

    for (iDof_Elem = 0; iDof_Elem < Element_nDofs; iDof_Elem++) {
      for (jDof_Elem = 0; jDof_Elem < Element_nDofs; jDof_Elem++) {
	SM[iElement][iDof_Elem][jDof_Elem] = 0.0;
      }
    }
  }
}

void AutoLinAS_INSIDE_AssembleElementMatrix (int iElement)
{
  double em[Element_nDofs][Element_nDofs];
  int iDof_Elem, jDof_Elem;

  assert(System_elementMatrix != NULL);

  (*System_elementMatrix) (iElement, 
			   em);

  for (iDof_Elem = 0; iDof_Elem < Element_nDofs; iDof_Elem++) {
    for (jDof_Elem = 0; jDof_Elem < Element_nDofs; jDof_Elem++) {
      SM[iElement][iDof_Elem][jDof_Elem] 
	= em[iDof_Elem][jDof_Elem];
    }
  }
}

void AutoLinAS_INSIDE_MultiplyMatrixAndVector
(double sv0[System_maxDofs], 
 double sv1[System_maxDofs])
{
  int iDof;
  int iElement;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    sv1[iDof] = 0.0;
  }

  for (iElement = 0; iElement < System_nElements; iElement++) {
    double tmp0[Element_nDofs];
    double tmp1[Element_nDofs];
    int iNode_Elem;
    int iDof_Node;
    int iDof_Elem, jDof_Elem;

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int iNode = (*System_elementNodeId) (iElement, iNode_Elem);
      
      for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	tmp0[iNode_Elem * Node_nDofs + iDof_Node] 
	  = sv0[iNode * Node_nDofs + iDof_Node];
      }
    }

    for (iDof_Elem = 0; iDof_Elem < Element_nDofs; iDof_Elem++) {
      tmp1[iDof_Elem] = 0.0;

      for (jDof_Elem = 0; jDof_Elem < Element_nDofs; jDof_Elem++) {
	tmp1[iDof_Elem] 
	  += SM[iElement][iDof_Elem][jDof_Elem] * tmp0[jDof_Elem];
      }
    }

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int iNode = (*System_elementNodeId) (iElement, iNode_Elem);
      
      for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	sv1[iNode * Node_nDofs + iDof_Node] 
	  += tmp1[iNode_Elem * Node_nDofs + iDof_Node];
      }
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {
      sv1[iDof] = sv0[iDof];
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize + System_nDofs);
  AutoLinAS_AddIO (M_CACHE, System_nMatrixSize 
		   + System_nMatrixSize / Element_nDofs);
#endif
}

void AutoLinAS_INSIDE_SetDirichletCondition (void)
{
  int iDof;
  int iElement;
  int iDof_Elem, jDof_Elem;
  int jNode_Elem;
  int jDof_Node;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVModified[iDof] = 0.0;
  }

  for (iElement = 0; iElement < System_nElements; iElement++) {

    for (iDof_Elem = 0; iDof_Elem < Element_nDofs; iDof_Elem++) {
      int iNode = (*System_elementNodeId) (iElement, iDof_Elem / Node_nDofs);
      int iDof = iNode * Node_nDofs + iDof_Elem % Node_nDofs;

      if (SVFixedFlag[iDof]) {
	double tmp0[Element_nDofs];

	for (jDof_Elem = 0; jDof_Elem < Element_nDofs; jDof_Elem++) {
	  tmp0[jDof_Elem] 
	    = -SVFixed[iDof] * SM[iElement][jDof_Elem][iDof_Elem];
	}

	for (jNode_Elem = 0; jNode_Elem < Element_nNodes; jNode_Elem++) {
	  int jNode = (*System_elementNodeId) (iElement, jNode_Elem);
	  
	  for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	    SVModified[jNode * Node_nDofs + jDof_Node] 
	      += tmp0[jNode_Elem * Node_nDofs + jDof_Node];
	  }
	}
      }
    }
  }

  for (iElement = 0; iElement < System_nElements; iElement++) {

    for (iDof_Elem = 0; iDof_Elem < Element_nDofs; iDof_Elem++) {
      int iNode = (*System_elementNodeId) (iElement, iDof_Elem / Node_nDofs);
      int iDof = iNode * Node_nDofs + iDof_Elem % Node_nDofs;

      if (SVFixedFlag[iDof]) {

	for (jDof_Elem = 0; jDof_Elem < Element_nDofs; jDof_Elem++) {
	  SM[iElement][iDof_Elem][jDof_Elem] = 0.0;
	  SM[iElement][jDof_Elem][iDof_Elem] = 0.0;
	}
      }
    }
  }
}

void AutoLinAS_INSIDE_MakeScalingVector (void)
{
  int iDof;
  int iElement;
  int iDof_Elem;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVScale[iDof] = 0.0;
  }

  for (iElement = 0; iElement < System_nElements; iElement++) {

    for (iDof_Elem = 0; iDof_Elem < Element_nDofs; iDof_Elem++) {
      int iNode = (*System_elementNodeId) (iElement, iDof_Elem / Node_nDofs);
      int iDof = iNode * Node_nDofs + iDof_Elem % Node_nDofs;

      SVScale[iDof] += SM[iElement][iDof_Elem][iDof_Elem];
    }
  }
  
  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {
      SVScale[iDof] = 1.0;
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVScale[iDof] = sqrt (SVScale[iDof]);
  }
}

void AutoLinAS_INSIDE_ScaleMatrix (void)
{
  int iElement;
  int iDof_Elem, jDof_Elem;

  for (iElement = 0; iElement < System_nElements; iElement++) {

    for (iDof_Elem = 0; iDof_Elem < Element_nDofs; iDof_Elem++) {
      int iNode = (*System_elementNodeId) (iElement, iDof_Elem / Node_nDofs);
      int iDof = iNode * Node_nDofs + iDof_Elem % Node_nDofs;

      for (jDof_Elem = 0; jDof_Elem < Element_nDofs; jDof_Elem++) {
	int jNode = (*System_elementNodeId) (iElement, jDof_Elem / Node_nDofs);
	int jDof = jNode * Node_nDofs + jDof_Elem % Node_nDofs;

	SM[iElement][iDof_Elem][jDof_Elem] 
	  /= SVScale[iDof] * SVScale[jDof];
      }
    }
  }
}

