/* autolinas_dscg_full.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dscg_inside.h"



/* symmetric matrix */
/* iterative method, conjugate gradient */
/* diagonal scaling preconditioner */
/* full matrix */



/* module variables */

static double SM[System_maxDofs][System_maxDofs];



void AutoLinAS_INSIDE_InitializeMatrix (void)
{
}

void AutoLinAS_INSIDE_Clear (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    for (jDof = 0; jDof < System_nDofs; jDof++) {
      SM[iDof][jDof] = 0.0;
    }
  }
}

void AutoLinAS_INSIDE_AssembleElementMatrix (int iElement)
{
  double em[Element_nDofs][Element_nDofs];
  int iNode_Elem, jNode_Elem;
  int iDof_Node, jDof_Node;
  
  assert(System_elementMatrix != NULL);

  (*System_elementMatrix) (iElement, 
			   em);

  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = (*System_elementNodeId) (iElement, iNode_Elem);
    
    for (jNode_Elem = 0; jNode_Elem < Element_nNodes; jNode_Elem++) {
      int jNode = (*System_elementNodeId) (iElement, jNode_Elem);
      
      for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  int iDof = iNode * Node_nDofs + iDof_Node;
	  int jDof = jNode * Node_nDofs + jDof_Node;
	  int iDof_Elem = iNode_Elem * Node_nDofs + iDof_Node;
	  int jDof_Elem = jNode_Elem * Node_nDofs + jDof_Node;
	  
	  SM[iDof][jDof] += em[iDof_Elem][jDof_Elem];
	}
      }
    }
  }
}

void AutoLinAS_INSIDE_MultiplyMatrixAndVector
(double sv0[System_maxDofs], 
 double sv1[System_maxDofs])
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    sv1[iDof] = 0.0;

    for (jDof = 0; jDof < System_nDofs; jDof++) {
      sv1[iDof] += SM[iDof][jDof] * sv0[jDof];
    }
  }
}

void AutoLinAS_INSIDE_SetDirichletCondition (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVModified[iDof] = 0.0;
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {
      double value = SVFixed[iDof];

      for (jDof = 0; jDof < System_nDofs; jDof++) {
	SVModified[jDof] -= value * SM[jDof][iDof];
      }
      
      for (jDof = 0; jDof < System_nDofs; jDof++) {
	SM[iDof][jDof] = 0.0;
	SM[jDof][iDof] = 0.0;
      }

      SM[iDof][iDof] = 1.0;
    }
  }
}

void AutoLinAS_INSIDE_MakeScalingVector (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVScale[iDof] = sqrt (SM[iDof][iDof]);
  }
}

void AutoLinAS_INSIDE_ScaleMatrix (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVScale[iDof] = sqrt (SM[iDof][iDof]);
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    for (jDof = 0; jDof < System_nDofs; jDof++) {
      SM[iDof][jDof] /= SVScale[iDof] * SVScale[jDof];
    }
  }
}

