/* autolinas_ldl_skyline.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_ldl_inside.h"
#include "autolinas_ldl_skyline_inside.h"



/*

  *
  * *
  - * *
  - - * *
  - - - * *
  - - - * * *
  - - - * * * *
  - - - - * * * *
  - - - - - - * * *
  - - - - - - * * * *

  symmetric, lower triangle (skyline storage)

*/



/* module variables */

int System_nMatrixSize;
int System_minIndexs[System_maxDofs];
int System_maxIndexs[System_maxDofs];
int System_ij[System_maxDofs];

double SM[System_maxMatrixSize];

/* 
   SM[System_ij[iDof] + jDof];
   if (System_minIndexs[iDof] <= jDof && jDof <= iDof)
*/



static void CheckIndexRange (int iElement)
{
  int iNode_Elem;
  int iDof_Node;
  int minNodeNumber;
  int maxNodeNumber;
  int minIndex;
  int maxIndex;
  
  minNodeNumber = (*System_elementNodeId) (iElement, 0);
  maxNodeNumber = minNodeNumber;
  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = (*System_elementNodeId) (iElement, iNode_Elem);

    if (iNode < minNodeNumber) {
      minNodeNumber = iNode;
    }
    if (maxNodeNumber < iNode) {
      maxNodeNumber = iNode;
    }
  }
  minIndex = minNodeNumber * Node_nDofs;
  maxIndex = maxNodeNumber * Node_nDofs + Node_nDofs - 1;
  
  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = (*System_elementNodeId) (iElement, iNode_Elem);

    if (minIndex <= System_minIndexs[iNode * Node_nDofs]) {
      for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	System_minIndexs[iNode * Node_nDofs + iDof_Node] = minIndex;
      }
    }
    if (System_maxIndexs[iNode * Node_nDofs] <= maxIndex) {
      for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	System_maxIndexs[iNode * Node_nDofs + iDof_Node] = maxIndex;
      }
    }
  }
}

void AutoLinAS_INSIDE_InitializeMatrix (void)
{
  int iDof;
  int iElement;

  for (iDof = 0 ; iDof < System_nDofs; iDof++) {
    System_minIndexs[iDof] = iDof; 
    System_maxIndexs[iDof] = iDof;
  }
  for (iElement = 0; iElement < System_nElements; iElement++) {
    CheckIndexRange (iElement);
  }

  System_ij[0] = 0;
  for (iDof = 1; iDof < System_nDofs; iDof++) {
    System_ij[iDof] 
      = System_ij[iDof - 1] + iDof - System_minIndexs[iDof];
  }
  System_nMatrixSize = System_ij[System_nDofs - 1] + System_nDofs;
  assert(System_nMatrixSize <= System_maxMatrixSize);

#ifdef MEASURE
  fprintf (stderr, " nDofs %d \n", System_nDofs);
  fprintf (stderr, " mean bandwidth %d \n", 
	   System_nMatrixSize / System_nDofs);
  fprintf (stderr, " SkylineMatrix size %d \n", 
	   System_nMatrixSize);
#endif
}

void AutoLinAS_INSIDE_Clear (void)
{
  int i;

  for (i = 0; i < System_nMatrixSize; i++) {
    SM[i] = 0.0;
  }
}

void AutoLinAS_INSIDE_AssembleElementMatrix (int iElement)
{
  double em[Element_nDofs][Element_nDofs];
  int iNode_Elem, jNode_Elem;
  int iDof_Node, jDof_Node;
  
  assert(System_elementMatrix != NULL);

  (*System_elementMatrix) (iElement, 
			   em);
  
  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = (*System_elementNodeId) (iElement, iNode_Elem);

    for (jNode_Elem = 0; jNode_Elem < Element_nNodes; jNode_Elem++) {
      int jNode = (*System_elementNodeId) (iElement, jNode_Elem);
      
      for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  int iDof_System = iNode * Node_nDofs + iDof_Node;
	  int jDof_System = jNode * Node_nDofs + jDof_Node;
	  int iDof_Elem = iNode_Elem * Node_nDofs + iDof_Node;
	  int jDof_Elem = jNode_Elem * Node_nDofs + jDof_Node;

	  if (jDof_System <= iDof_System) {
	    SM[System_ij[iDof_System] + jDof_System] 
	      += em[iDof_Elem][jDof_Elem];
	  }
	}
      }
    }
  }
}

void AutoLinAS_INSIDE_SetDirichletCondition (void)
{
  int iDof;
  
  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVModified[iDof] = 0.0;
  }
  
  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {
      double value = SVFixed[iDof];
      double coefficient = SM[System_ij[iDof] + iDof] * PENALTY_NUMBER;
      
      SM[System_ij[iDof] + iDof] = coefficient;
      SVModified[iDof] = value * coefficient;
    }
  }
}

