/* autolinas_ldl_skyline_block_mt.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_multithread_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_ldl_inside.h"
#include "autolinas_ldl_skyline_inside.h"



/* symmetric matrix */
/* direct method, LtDL decomposition */
/* profile form (IJK), block */
/* sparse matrix, skyline storage, lower triangle */



/* module initialization */

#if 0
#define BLOCK_SIZE 36
#else
#define BLOCK_SIZE 72
#endif

static int System_nBlocks;
static int System_minBlockIndexs[System_maxDofs / BLOCK_SIZE + 1];

static double SB0[BLOCK_SIZE][BLOCK_SIZE];  /* source block 0 */
static double DB[BLOCK_SIZE][BLOCK_SIZE];   /* destination block 0 */



static void CheckBlock (void)
{
  int bi;
  int i;

  System_nBlocks = (System_nDofs + BLOCK_SIZE - 1) / BLOCK_SIZE;
  for (bi = 0; bi < System_nBlocks; bi++) {
    int iMin = System_nDofs;

    for (i = 0; i < BLOCK_SIZE; i++) {
      int I = bi * BLOCK_SIZE + i;

      if (I < System_nDofs) {
	if (System_minIndexs[I] < iMin) {
	  iMin = System_minIndexs[I];
	}
      }
    }
    System_minBlockIndexs[bi] = iMin / BLOCK_SIZE;
  }
}

static void ReadDiagonal 
(int bi,
 double A[BLOCK_SIZE])
{
  int i;

  for (i = 0; i < BLOCK_SIZE; i++) {
    int I = bi * BLOCK_SIZE + i;

    if (I < System_nDofs) {
      A[i] = SM[System_ij[I] + I];
    } else {
      A[i] = 1.0;
    }
  }

#ifdef MEASURE
  AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * 2);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE);
#endif
}

static void ReadBlock 
(int bi, int bj,
 double B[BLOCK_SIZE][BLOCK_SIZE])
{
  int i, j;

  for (i = 0; i < BLOCK_SIZE; i++) {
    int I = bi * BLOCK_SIZE + i;

    if (I < System_nDofs) {
      for (j = 0; j < BLOCK_SIZE; j++) {
	int J = bj * BLOCK_SIZE + j;

	if (System_minIndexs[I] <= J && J <= I) {
	  B[i][j] = SM[System_ij[I] + J];
	} else {
	  B[i][j] = 0.0;
	}
      }
    } else {
      for (j = 0; j < BLOCK_SIZE; j++) {
	B[i][j] = 0.0;
      }
      B[i][i] = 1.0;
    }
  }

#ifdef MEASURE
  if (bi == bj) {
    AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * BLOCK_SIZE / 2);
  } else {
    AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * BLOCK_SIZE);
  }
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE);
#endif
}

static void WriteBlock 
(int bi, int bj,
 double B[BLOCK_SIZE][BLOCK_SIZE])
{
  int i, j;

  for (i = 0; i < BLOCK_SIZE; i++) {
    int I = bi * BLOCK_SIZE + i;

    if (I < System_nDofs) {
      for (j = 0; j < BLOCK_SIZE; j++) {
	int J = bj * BLOCK_SIZE + j;

	if (System_minIndexs[I] <= J && J <= I) {
	  SM[System_ij[I] + J] = B[i][j];
	}
      }
    }
  }

#ifdef MEASURE
  if (bi == bj) {
    AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * BLOCK_SIZE / 2);
  } else {
    AutoLinAS_AddIO (M_MEMORY, BLOCK_SIZE * BLOCK_SIZE);
  }
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE);
#endif
}

/* DB (bj, bj), DB (bj, bj) -> DB (bj, bj) */
static void DoBlock_SJJ_SJJ_DJJ (void)
{
  int i, j, l;

  for (j = 0; j < BLOCK_SIZE; j += 3) {
    double sum0 = 0.0;
    double sum1 = 0.0;
    double sum2 = 0.0;

    for (l = 0; l < j; l++) {
      sum0 += DB[l][l] * DB[j + 0][l] * DB[j + 0][l];
      sum1 += DB[l][l] * DB[j + 0][l] * DB[j + 1][l];
      sum2 += DB[l][l] * DB[j + 0][l] * DB[j + 2][l];
    }

    DB[j + 0][j + 0] -= sum0;

    DB[j + 1][j + 0] -= sum1;
    DB[j + 1][j + 0] /= DB[j + 0][j + 0];

    DB[j + 2][j + 0] -= sum2;
    DB[j + 2][j + 0] /= DB[j + 0][j + 0];
    
    sum1 = 0.0;
    sum2 = 0.0;

    for (l = 0; l < j + 1; l++) {
      sum1 += DB[l][l] * DB[j + 1][l] * DB[j + 1][l];
      sum2 += DB[l][l] * DB[j + 1][l] * DB[j + 2][l];
    }

    DB[j + 1][j + 1] -= sum1;

    DB[j + 2][j + 1] -= sum2;
    DB[j + 2][j + 1] /= DB[j + 1][j + 1];

    sum2 = 0.0;

    for (l = 0; l < j + 2; l++) {
      sum2 += DB[l][l] * DB[j + 2][l] * DB[j + 2][l];
    }

    DB[j + 2][j + 2] -= sum2;

    for (i = j + 3; i < BLOCK_SIZE; i += 3) {
      double sum0_0 = 0.0;
      double sum0_1 = 0.0;
      double sum0_2 = 0.0;

      double sum1_0 = 0.0;
      double sum1_1 = 0.0;
      double sum1_2 = 0.0;

      double sum2_0 = 0.0;
      double sum2_1 = 0.0;
      double sum2_2 = 0.0;

      for (l = 0; l < j; l++) {
	double da = DB[l][l];
	double da_sb0_0 = da * DB[j + 0][l];
	double da_sb0_1 = da * DB[j + 1][l];
	double da_sb0_2 = da * DB[j + 2][l];
	double db0 = DB[i + 0][l];
	double db1 = DB[i + 1][l];
	double db2 = DB[i + 2][l];

	sum0_0 += da_sb0_0 * db0;
	sum0_1 += da_sb0_0 * db1;
	sum0_2 += da_sb0_0 * db2;

	sum1_0 += da_sb0_1 * db0;
	sum1_1 += da_sb0_1 * db1;
	sum1_2 += da_sb0_1 * db2;

	sum2_0 += da_sb0_2 * db0;
	sum2_1 += da_sb0_2 * db1;
	sum2_2 += da_sb0_2 * db2;
      }

      DB[i + 0][j + 0] -= sum0_0;
      DB[i + 1][j + 0] -= sum0_1;
      DB[i + 2][j + 0] -= sum0_2;
      DB[i + 0][j + 0] /= DB[j + 0][j + 0];
      DB[i + 1][j + 0] /= DB[j + 0][j + 0];
      DB[i + 2][j + 0] /= DB[j + 0][j + 0];

      sum1_0 += DB[j + 0][j + 0] * DB[j + 1][j + 0] * DB[i + 0][j + 0];
      sum1_1 += DB[j + 0][j + 0] * DB[j + 1][j + 0] * DB[i + 1][j + 0];
      sum1_2 += DB[j + 0][j + 0] * DB[j + 1][j + 0] * DB[i + 2][j + 0];

      DB[i + 0][j + 1] -= sum1_0;
      DB[i + 1][j + 1] -= sum1_1;
      DB[i + 2][j + 1] -= sum1_2;
      DB[i + 0][j + 1] /= DB[j + 1][j + 1];
      DB[i + 1][j + 1] /= DB[j + 1][j + 1];
      DB[i + 2][j + 1] /= DB[j + 1][j + 1];

      sum2_0 += DB[j + 0][j + 0] * DB[j + 2][j + 0] * DB[i + 0][j + 0];
      sum2_1 += DB[j + 0][j + 0] * DB[j + 2][j + 0] * DB[i + 1][j + 0];
      sum2_2 += DB[j + 0][j + 0] * DB[j + 2][j + 0] * DB[i + 2][j + 0];

      sum2_0 += DB[j + 1][j + 1] * DB[j + 2][j + 1] * DB[i + 0][j + 1];
      sum2_1 += DB[j + 1][j + 1] * DB[j + 2][j + 1] * DB[i + 1][j + 1];
      sum2_2 += DB[j + 1][j + 1] * DB[j + 2][j + 1] * DB[i + 2][j + 1];

      DB[i + 0][j + 2] -= sum2_0;
      DB[i + 1][j + 2] -= sum2_1;
      DB[i + 2][j + 2] -= sum2_2;
      DB[i + 0][j + 2] /= DB[j + 2][j + 2];
      DB[i + 1][j + 2] /= DB[j + 2][j + 2];
      DB[i + 2][j + 2] /= DB[j + 2][j + 2];

    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 36 * 9);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 36 * 12);
  AutoLinAS_AddOp (M_OTHER_OP, BLOCK_SIZE * BLOCK_SIZE / 2);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 36 * 7);
#endif
}

/* SB0 (bj, bj), DB (bi, bj) -> DB (bi, bj) */
static void DoBlock_SJJ_SIJ_DIJ (void)
{
  int i, j, l;

  for (j = 0; j < BLOCK_SIZE; j += 3) {

    for (i = 0; i < BLOCK_SIZE; i += 3) {
      double sum0_0 = 0.0;
      double sum0_1 = 0.0;
      double sum0_2 = 0.0;

      double sum1_0 = 0.0;
      double sum1_1 = 0.0;
      double sum1_2 = 0.0;

      double sum2_0 = 0.0;
      double sum2_1 = 0.0;
      double sum2_2 = 0.0;

      for (l = 0; l < j; l++) {
	double da = SB0[l][l];
	double da_sb0_0 = da * SB0[j + 0][l];
	double da_sb0_1 = da * SB0[j + 1][l];
	double da_sb0_2 = da * SB0[j + 2][l];
	double db0 = DB[i + 0][l];
	double db1 = DB[i + 1][l];
	double db2 = DB[i + 2][l];

	sum0_0 += da_sb0_0 * db0;
	sum0_1 += da_sb0_0 * db1;
	sum0_2 += da_sb0_0 * db2;

	sum1_0 += da_sb0_1 * db0;
	sum1_1 += da_sb0_1 * db1;
	sum1_2 += da_sb0_1 * db2;

	sum2_0 += da_sb0_2 * db0;
	sum2_1 += da_sb0_2 * db1;
	sum2_2 += da_sb0_2 * db2;
      }

      DB[i + 0][j + 0] -= sum0_0;
      DB[i + 1][j + 0] -= sum0_1;
      DB[i + 2][j + 0] -= sum0_2;
      DB[i + 0][j + 0] /= SB0[j + 0][j + 0];
      DB[i + 1][j + 0] /= SB0[j + 0][j + 0];
      DB[i + 2][j + 0] /= SB0[j + 0][j + 0];

      sum1_0 += SB0[j][j] * SB0[j + 1][j] * DB[i + 0][j];
      sum1_1 += SB0[j][j] * SB0[j + 1][j] * DB[i + 1][j];
      sum1_2 += SB0[j][j] * SB0[j + 1][j] * DB[i + 2][j];

      DB[i + 0][j + 1] -= sum1_0;
      DB[i + 1][j + 1] -= sum1_1;
      DB[i + 2][j + 1] -= sum1_2;
      DB[i + 0][j + 1] /= SB0[j + 1][j + 1];
      DB[i + 1][j + 1] /= SB0[j + 1][j + 1];
      DB[i + 2][j + 1] /= SB0[j + 1][j + 1];

      sum2_0 += SB0[j + 0][j + 0] * SB0[j + 2][j + 0] * DB[i + 0][j + 0];
      sum2_1 += SB0[j + 0][j + 0] * SB0[j + 2][j + 0] * DB[i + 1][j + 0];
      sum2_2 += SB0[j + 0][j + 0] * SB0[j + 2][j + 0] * DB[i + 2][j + 0];

      sum2_0 += SB0[j + 1][j + 1] * SB0[j + 2][j + 1] * DB[i + 0][j + 1];
      sum2_1 += SB0[j + 1][j + 1] * SB0[j + 2][j + 1] * DB[i + 1][j + 1];
      sum2_2 += SB0[j + 1][j + 1] * SB0[j + 2][j + 1] * DB[i + 2][j + 1];

      DB[i + 0][j + 2] -= sum2_0;
      DB[i + 1][j + 2] -= sum2_1;
      DB[i + 2][j + 2] -= sum2_2;
      DB[i + 0][j + 2] /= SB0[j + 2][j + 2];
      DB[i + 1][j + 2] /= SB0[j + 2][j + 2];
      DB[i + 2][j + 2] /= SB0[j + 2][j + 2];

    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 18 * 9);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 18 * 12);
  AutoLinAS_AddOp (M_OTHER_OP, BLOCK_SIZE * BLOCK_SIZE);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 18 * 7);
#endif
}



/* ---------------- start multi-threaded part --------------------------- */

#define MAX_THREAD 8

static double MT_DA[MAX_THREAD][BLOCK_SIZE];               /* diagonal array */
static double MT_A0[MAX_THREAD][BLOCK_SIZE];               /* temporary array */
static double MT_A1[MAX_THREAD][BLOCK_SIZE];               /* temporary array */
static double MT_A2[MAX_THREAD][BLOCK_SIZE];               /* temporary array */
static double MT_SB0[MAX_THREAD][BLOCK_SIZE][BLOCK_SIZE];  /* source block 0 */
static double MT_SB1[MAX_THREAD][BLOCK_SIZE][BLOCK_SIZE];  /* source block 1 */
static double MT_DB[MAX_THREAD][BLOCK_SIZE][BLOCK_SIZE];   /* destination block 0 */



static void ClearMT_DB (int myThreadId)
{
  int i, j;

  for (i = 0; i < BLOCK_SIZE; i++) {
    for (j = 0; j < BLOCK_SIZE; j++) {
      MT_DB[myThreadId][i][j] = 0.0;
    }
  }
}

/* SB0 (bj, bl), SB0 (bj, bl) -> DB (bj, bj) */
static void DoBlockMT_SJL_SJL_DJJ 
(int myThreadId)
{
  int i, j, l;

  for (j = 0; j < BLOCK_SIZE; j += 3) {

    for (l = 0; l < BLOCK_SIZE; l++) {
      double da = MT_DA[myThreadId][l];

      MT_A0[myThreadId][l] = da * MT_SB0[myThreadId][j + 0][l];
      MT_A1[myThreadId][l] = da * MT_SB0[myThreadId][j + 1][l];
      MT_A2[myThreadId][l] = da * MT_SB0[myThreadId][j + 2][l];
    }

    for (i = j; i < BLOCK_SIZE; i += 3) {
      double sum0_0 = 0.0;
      double sum0_1 = 0.0;
      double sum0_2 = 0.0;

      double sum1_0 = 0.0;
      double sum1_1 = 0.0;
      double sum1_2 = 0.0;

      double sum2_0 = 0.0;
      double sum2_1 = 0.0;
      double sum2_2 = 0.0;

      for (l = 0; l < BLOCK_SIZE; l++) {
	double da_sb0_0 = MT_A0[myThreadId][l];
	double da_sb0_1 = MT_A1[myThreadId][l];
	double da_sb0_2 = MT_A2[myThreadId][l];
	double sb1_0 = MT_SB0[myThreadId][i + 0][l];
	double sb1_1 = MT_SB0[myThreadId][i + 1][l];
	double sb1_2 = MT_SB0[myThreadId][i + 2][l];

	sum0_0 += da_sb0_0 * sb1_0;
	sum0_1 += da_sb0_0 * sb1_1;
	sum0_2 += da_sb0_0 * sb1_2;
	
	sum1_0 += da_sb0_1 * sb1_0;
	sum1_1 += da_sb0_1 * sb1_1;
	sum1_2 += da_sb0_1 * sb1_2;
	
	sum2_0 += da_sb0_2 * sb1_0;
	sum2_1 += da_sb0_2 * sb1_1;
	sum2_2 += da_sb0_2 * sb1_2;
      }

      if (i == j) {
	MT_DB[myThreadId][i + 0][j + 0] -= sum0_0;
	MT_DB[myThreadId][i + 1][j + 0] -= sum0_1;
	MT_DB[myThreadId][i + 2][j + 0] -= sum0_2;
	
	MT_DB[myThreadId][i + 1][j + 1] -= sum1_1;
	MT_DB[myThreadId][i + 2][j + 1] -= sum1_2;
	
	MT_DB[myThreadId][i + 2][j + 2] -= sum2_2;
      } else {
	MT_DB[myThreadId][i + 0][j + 0] -= sum0_0;
	MT_DB[myThreadId][i + 1][j + 0] -= sum0_1;
	MT_DB[myThreadId][i + 2][j + 0] -= sum0_2;
	
	MT_DB[myThreadId][i + 0][j + 1] -= sum1_0;
	MT_DB[myThreadId][i + 1][j + 1] -= sum1_1;
	MT_DB[myThreadId][i + 2][j + 1] -= sum1_2;
	
	MT_DB[myThreadId][i + 0][j + 2] -= sum2_0;
	MT_DB[myThreadId][i + 1][j + 2] -= sum2_1;
	MT_DB[myThreadId][i + 2][j + 2] -= sum2_2;
      }
    }
  }

#if 0
#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 18 * 9);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 18 * 9);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE / 6 * 3);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 18 * 6);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE / 6 * 7);
#endif
#endif
}

/* SB0 (bj, bl), SB1 (bi, bl) -> DB (bi, bj) */
static void DoBlockMT_SJL_SIL_DIJ 
(int myThreadId)
{
  int i, j, l;

  for (j = 0; j < BLOCK_SIZE; j += 3) {

    for (l = 0; l < BLOCK_SIZE; l++) {
      double da = MT_DA[myThreadId][l];

      MT_A0[myThreadId][l] = da * MT_SB0[myThreadId][j + 0][l];
      MT_A1[myThreadId][l] = da * MT_SB0[myThreadId][j + 1][l];
      MT_A2[myThreadId][l] = da * MT_SB0[myThreadId][j + 2][l];
    }

    for (i = 0; i < BLOCK_SIZE; i += 3) {
      double sum0_0 = 0.0;
      double sum0_1 = 0.0;
      double sum0_2 = 0.0;

      double sum1_0 = 0.0;
      double sum1_1 = 0.0;
      double sum1_2 = 0.0;

      double sum2_0 = 0.0;
      double sum2_1 = 0.0;
      double sum2_2 = 0.0;

      for (l = 0; l < BLOCK_SIZE; l++) {
	double da_sb0_0 = MT_A0[myThreadId][l];
	double da_sb0_1 = MT_A1[myThreadId][l];
	double da_sb0_2 = MT_A2[myThreadId][l];
	double sb1_0 = MT_SB1[myThreadId][i + 0][l];
	double sb1_1 = MT_SB1[myThreadId][i + 1][l];
	double sb1_2 = MT_SB1[myThreadId][i + 2][l];

	sum0_0 += da_sb0_0 * sb1_0;
	sum0_1 += da_sb0_0 * sb1_1;
	sum0_2 += da_sb0_0 * sb1_2;

	sum1_0 += da_sb0_1 * sb1_0;
	sum1_1 += da_sb0_1 * sb1_1;
	sum1_2 += da_sb0_1 * sb1_2;

	sum2_0 += da_sb0_2 * sb1_0;
	sum2_1 += da_sb0_2 * sb1_1;
	sum2_2 += da_sb0_2 * sb1_2;
      }

      MT_DB[myThreadId][i + 0][j + 0] -= sum0_0;
      MT_DB[myThreadId][i + 1][j + 0] -= sum0_1;
      MT_DB[myThreadId][i + 2][j + 0] -= sum0_2;

      MT_DB[myThreadId][i + 0][j + 1] -= sum1_0;
      MT_DB[myThreadId][i + 1][j + 1] -= sum1_1;
      MT_DB[myThreadId][i + 2][j + 1] -= sum1_2;

      MT_DB[myThreadId][i + 0][j + 2] -= sum2_0;
      MT_DB[myThreadId][i + 1][j + 2] -= sum2_1;
      MT_DB[myThreadId][i + 2][j + 2] -= sum2_2;
    }
  }

#if 0
#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 9 * 9);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 9 * 9);
  AutoLinAS_AddOp (M_MUL_OP, BLOCK_SIZE * BLOCK_SIZE / 3 * 3);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE * BLOCK_SIZE / 9 * 6);
  AutoLinAS_AddIO (M_CACHE, BLOCK_SIZE * BLOCK_SIZE / 3 * 7);
#endif
#endif
}

static void AssembleDB (int nThreads)
{
  int iThread;
  int i, j;

  for (iThread = 0; iThread < nThreads; iThread++) {

    for (i = 0; i < BLOCK_SIZE; i++) {
      for (j = 0; j < BLOCK_SIZE; j++) {
	DB[i][j] += MT_DB[iThread][i][j];
      }
    }
  }
}

static void LDLDecompositionMT 
(int myThreadId, int nThreads)
{
  int bi, bj, bl;

  for (bj = 0; bj < System_nBlocks; bj++) {

    AutoLinAS_BarrierThreads ();
    
    ClearMT_DB (myThreadId); 

    for (bl = System_minBlockIndexs[bj] + myThreadId; 
	 bl < bj; 
	 bl += nThreads) {
      ReadDiagonal (bl, 
		    MT_DA[myThreadId]);
      ReadBlock (bj, bl, 
		 MT_SB0[myThreadId]);
      DoBlockMT_SJL_SJL_DJJ (myThreadId); 
      /* DA (bl), SB0 (bj, bl), SB0 (bj, bl) -> DB (bj, bj) */
    }
    
    AutoLinAS_BarrierThreads ();

    if (myThreadId == 0) {
      ReadBlock (bj, bj, 
		 DB);

      AssembleDB (nThreads); 

      DoBlock_SJJ_SJJ_DJJ (); 
      /* DB (bj, bj), DB (bj, bj) -> DB (bj, bj) */
      WriteBlock (bj, bj, DB);
    }

    for (bi = bj + 1; bi < System_nBlocks; bi++) {
      if (bj < System_minBlockIndexs[bi]) {
	continue;
      }

      AutoLinAS_BarrierThreads ();

      ClearMT_DB (myThreadId); 

      for (bl = System_minBlockIndexs[bj] + myThreadId; 
	   bl < bj; 
	   bl += nThreads) {
	if (bl < System_minBlockIndexs[bi]) {
	  continue;
	}
	
	ReadDiagonal (bl, 
		      MT_DA[myThreadId]);
	ReadBlock (bj, bl, 
		   MT_SB0[myThreadId]);
	ReadBlock (bi, bl, 
		   MT_SB1[myThreadId]);
	DoBlockMT_SJL_SIL_DIJ (myThreadId); 
	/* DA (bl), SB0 (bj, bl), SB1 (bi, bl) -> DB (bi, bj) */
      }

      AutoLinAS_BarrierThreads ();

      if (myThreadId == 0) {
	ReadBlock (bi, bj, 
		   DB);

	AssembleDB (nThreads); 

	ReadBlock (bj, bj, 
		   SB0);
	DoBlock_SJJ_SIJ_DIJ (); 
	/* SB0 (bj, bj), DB (bi, bj) -> DB (bi, bj) */
	WriteBlock (bi, bj, DB);
      }
    }
  }

  AutoLinAS_BarrierThreads ();
}

/* ---------------- end multi-threaded part --------------------------- */



void AutoLinAS_INSIDE_LDLDecomposition (void)
{
  CheckBlock ();
  
  AutoLinAS_RunThreads (2, LDLDecompositionMT);
}

void AutoLinAS_INSIDE_ForwardReduction (void)
{
  int i, j;

#if (Node_nDofs == 1)
  for (i = 0; i < System_nDofs; i++) {
    double sum = 0.0;

    for (j = System_minIndexs[i]; j < i; j++) {
      sum += SM[System_ij[i] + j] * SV1[j];
    }
    SV1[i] -= sum;
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 2);
#endif

#endif
#if (Node_nDofs == 2)

  for (i = 0; i < System_nDofs; i += 2) {
    double sum0 = 0.0;
    double sum1 = 0.0;

    for (j = System_minIndexs[i]; j < i; j++) {
      sum0 += SM[System_ij[i + 0] + j] * SV1[j];
      sum1 += SM[System_ij[i + 1] + j] * SV1[j];
    }
    SV1[i + 0] -= sum0;

    sum1 += SM[System_ij[i + 1] + i] * SV1[i];
    SV1[i + 1] -= sum1;
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 1.5);
#endif

#endif
#if (Node_nDofs == 3)

  for (i = 0; i < System_nDofs; i += 3) {
    double sum0 = 0.0;
    double sum1 = 0.0;
    double sum2 = 0.0;

    for (j = System_minIndexs[i]; j < i; j++) {
      sum0 += SM[System_ij[i + 0] + j] * SV1[j];
      sum1 += SM[System_ij[i + 1] + j] * SV1[j];
      sum2 += SM[System_ij[i + 2] + j] * SV1[j];
    }
    SV1[i + 0] -= sum0;

    sum1 += SM[System_ij[i + 1] + i] * SV1[i];
    sum2 += SM[System_ij[i + 2] + i] * SV1[i];
    SV1[i + 1] -= sum1;

    sum2 += SM[System_ij[i + 2] + i + 1] * SV1[i + 1];
    SV1[i + 2] -= sum2;
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 1.333);
#endif

#endif

}

void AutoLinAS_INSIDE_BackwardSubstitution (void)
{
  int i, j;
  int maxBandWidth;

  maxBandWidth = -System_minIndexs[0];
  for (i = 1; i < System_nDofs; i++) {
    if (maxBandWidth < i - System_minIndexs[i]) {
      maxBandWidth = i - System_minIndexs[i];
    }
  }

  SV1[System_nDofs - 1] 
    /= SM[System_ij[System_nDofs - 1] + System_nDofs - 1];

  for (i = System_nDofs - 2; i >= 0; i--) {
    int jMin;

    jMin = i + maxBandWidth;
    if (jMin > System_nDofs - 1) {
      jMin = System_nDofs - 1;
    } else {
      while (System_minIndexs[jMin] > i && jMin != i) {
	jMin--;
      }
    }
    SV1[i] /= SM[System_ij[i] + i];

    for (j = i + 1; j <= jMin; j++) {
      if (i >= System_minIndexs[j]) {
        SV1[i] -= SM[System_ij[j] + i] * SV1[j];
      }
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_OTHER_OP, System_nDofs);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 2);
#endif
}

