/* autolinas_ldl_skyline_profile_tuned.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_ldl_inside.h"
#include "autolinas_ldl_skyline_inside.h"



/* symmetric matrix */
/* direct method, LtDL decomposition */
/* profile form (IJK) */
/* sparse matrix, skyline storage, lower triangle */



/* ***TUNED*** */
#if (Node_nDofs == 1)

void AutoLinAS_INSIDE_LDLDecomposition (void)
{
  static double k_xx[System_maxDofs];
  static double k_xx_jx[System_maxDofs];
  int iDof, jDof, kDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    k_xx[iDof] = SM[System_ij[iDof] + iDof];
  }

#ifdef MEASURE
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 2);
#endif

  for (jDof = 0; jDof < System_nDofs; jDof++) {
    double sum = 0.0;

    for (kDof = System_minIndexs[jDof]; kDof < jDof; kDof++) {
      sum += k_xx[kDof] 
	* SM[System_ij[jDof] + kDof] 
	* SM[System_ij[jDof] + kDof];
    }

    SM[System_ij[jDof] + jDof] -= sum;
    k_xx[jDof] = SM[System_ij[jDof] + jDof];

    for (kDof = System_minIndexs[jDof]; kDof < jDof; kDof++) {
      k_xx_jx[kDof] = k_xx[kDof] * SM[System_ij[jDof] + kDof];
    }

    for (iDof = jDof + 1; iDof < System_nDofs; iDof++) {
      double sum = 0.0;
      int iMax;

      if (jDof < System_minIndexs[iDof]) {
	continue;
      }

      if (System_minIndexs[iDof] > System_minIndexs[jDof]) {
	iMax = System_minIndexs[iDof];
      } else {
	iMax = System_minIndexs[jDof];
      }

      for (kDof = iMax; kDof < jDof; kDof++) {
	sum += k_xx_jx[kDof] * SM[System_ij[iDof] + kDof];
      }

#ifdef MEASURE
      if (0 < jDof - iMax) {
	int count = jDof - iMax;

	AutoLinAS_AddOp (M_ADD_OP, count * 1);
	AutoLinAS_AddOp (M_MUL_OP, count * 1);
	AutoLinAS_AddIO (M_MEMORY, count * 2);
      }
#endif

      SM[System_ij[iDof] + jDof] -= sum;
      SM[System_ij[iDof] + jDof] /= k_xx[jDof];

#ifdef MEASURE
      AutoLinAS_AddOp (M_ADD_OP, 2);
      AutoLinAS_AddOp (M_MUL_OP, 2);
      AutoLinAS_AddOp (M_OTHER_OP, 1);
      AutoLinAS_AddIO (M_MEMORY, 6);
#endif
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nDofs);
  AutoLinAS_AddOp (M_OTHER_OP, System_nDofs);
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 3);
#endif
}

#endif
#if (Node_nDofs == 2)

void AutoLinAS_INSIDE_LDLDecomposition (void)
{
  static double k_xx[System_maxDofs];
  static double k_xx_jx0[System_maxDofs];
  static double k_xx_jx1[System_maxDofs];
  int iDof, jDof, kDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    k_xx[iDof] = SM[System_ij[iDof] + iDof];
  }

#ifdef MEASURE
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 2);
#endif

  for (jDof = 0; jDof < System_nDofs; jDof += 2) {
    double sum0 = 0.0;
    double sum1 = 0.0;

    for (kDof = System_minIndexs[jDof]; kDof < jDof; kDof++) {
      sum0 += k_xx[kDof] 
	* SM[System_ij[jDof + 0] + kDof] 
	* SM[System_ij[jDof + 0] + kDof];
      sum1 += k_xx[kDof] 
	* SM[System_ij[jDof + 0] + kDof] 
	* SM[System_ij[jDof + 1] + kDof];
    }

    SM[System_ij[jDof + 0] + jDof + 0] -= sum0;
    k_xx[jDof + 0] = SM[System_ij[jDof + 0] + jDof + 0];

    SM[System_ij[jDof + 1] + jDof + 0] -= sum1;
    SM[System_ij[jDof + 1] + jDof + 0] /= k_xx[jDof + 0];
    
    sum1 = 0.0;

    for (kDof = System_minIndexs[jDof]; kDof < jDof + 1; kDof++) {
      sum1 += k_xx[kDof] 
	* SM[System_ij[jDof + 1] + kDof] 
	* SM[System_ij[jDof + 1] + kDof];
    }

    SM[System_ij[jDof + 1] + jDof + 1] -= sum1;
    k_xx[jDof + 1] = SM[System_ij[jDof + 1] + jDof + 1];

    for (kDof = System_minIndexs[jDof]; kDof < jDof; kDof++) {
      k_xx_jx0[kDof] = k_xx[kDof] * SM[System_ij[jDof + 0] + kDof];
      k_xx_jx1[kDof] = k_xx[kDof] * SM[System_ij[jDof + 1] + kDof];
    }

    for (iDof = jDof + 2; iDof < System_nDofs; iDof += 2) {
      double sum0_0 = 0.0;
      double sum1_0 = 0.0;
      double sum0_1 = 0.0;
      double sum1_1 = 0.0;
      int iMax;

      if (jDof < System_minIndexs[iDof]) {
	continue;
      }

      if (System_minIndexs[iDof] > System_minIndexs[jDof]) {
	iMax = System_minIndexs[iDof];
      } else {
	iMax = System_minIndexs[jDof];
      }

      for (kDof = iMax; kDof < jDof; kDof++) {
	double da_sb0_0 = k_xx_jx0[kDof];
	double da_sb0_1 = k_xx_jx1[kDof];
	double db0 = SM[System_ij[iDof + 0] + kDof];
	double db1 = SM[System_ij[iDof + 1] + kDof];

	sum0_0 += da_sb0_0 * db0;
	sum0_1 += da_sb0_0 * db1;

	sum1_0 += da_sb0_1 * db0;
	sum1_1 += da_sb0_1 * db1;
      }

#ifdef MEASURE
      if (0 < jDof - iMax) {
	int count = jDof - iMax;

	AutoLinAS_AddOp (M_ADD_OP, count * 4);
	AutoLinAS_AddOp (M_MUL_OP, count * 4);
	AutoLinAS_AddIO (M_MEMORY, count * 4);
      }
#endif

      SM[System_ij[iDof + 0] + jDof + 0] -= sum0_0;
      SM[System_ij[iDof + 1] + jDof + 0] -= sum0_1;
      SM[System_ij[iDof + 0] + jDof + 0] /= k_xx[jDof + 0];
      SM[System_ij[iDof + 1] + jDof + 0] /= k_xx[jDof + 0];
      
      sum1_0 += k_xx[jDof + 0] 
	* SM[System_ij[jDof + 1] + jDof + 0] 
	* SM[System_ij[iDof + 0] + jDof + 0];
      sum1_1 += k_xx[jDof + 0] 
	* SM[System_ij[jDof + 1] + jDof + 0] 
	* SM[System_ij[iDof + 1] + jDof + 0];
      
      SM[System_ij[iDof + 0] + jDof + 1] -= sum1_0;
      SM[System_ij[iDof + 1] + jDof + 1] -= sum1_1;
      SM[System_ij[iDof + 0] + jDof + 1] /= k_xx[jDof + 1];
      SM[System_ij[iDof + 1] + jDof + 1] /= k_xx[jDof + 1];

#ifdef MEASURE
      AutoLinAS_AddOp (M_ADD_OP, 9);
      AutoLinAS_AddOp (M_MUL_OP, 10);
      AutoLinAS_AddOp (M_OTHER_OP, 4);
      AutoLinAS_AddIO (M_MEMORY, 23);
#endif
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nDofs * 3);
  AutoLinAS_AddOp (M_OTHER_OP, System_nDofs);
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 8);
#endif
}

#endif
#if (Node_nDofs == 3)

void AutoLinAS_INSIDE_LDLDecomposition (void)
{
  static double k_xx[System_maxDofs];
  static double k_xx_jx0[System_maxDofs];
  static double k_xx_jx1[System_maxDofs];
  static double k_xx_jx2[System_maxDofs];
  int iDof, jDof, kDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    k_xx[iDof] = SM[System_ij[iDof] + iDof];
  }

#ifdef MEASURE
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 2);
#endif

  for (jDof = 0; jDof < System_nDofs; jDof += 3) {
    double sum0 = 0.0;
    double sum1 = 0.0;
    double sum2 = 0.0;

    for (kDof = System_minIndexs[jDof]; kDof < jDof; kDof++) {
      sum0 += k_xx[kDof] 
	* SM[System_ij[jDof + 0] + kDof] 
	* SM[System_ij[jDof + 0] + kDof];
      sum1 += k_xx[kDof] 
	* SM[System_ij[jDof + 0] + kDof] 
	* SM[System_ij[jDof + 1] + kDof];
      sum2 += k_xx[kDof] 
	* SM[System_ij[jDof + 0] + kDof] 
	* SM[System_ij[jDof + 2] + kDof];
    }

    SM[System_ij[jDof + 0] + jDof + 0] -= sum0;
    k_xx[jDof + 0] = SM[System_ij[jDof + 0] + jDof + 0];

    SM[System_ij[jDof + 1] + jDof + 0] -= sum1;
    SM[System_ij[jDof + 1] + jDof + 0] /= k_xx[jDof + 0];
    
    SM[System_ij[jDof + 2] + jDof + 0] -= sum2;
    SM[System_ij[jDof + 2] + jDof + 0] /= k_xx[jDof + 0];
    
    sum1 = 0.0;
    sum2 = 0.0;

    for (kDof = System_minIndexs[jDof]; kDof < jDof + 1; kDof++) {
      sum1 += k_xx[kDof] 
	* SM[System_ij[jDof + 1] + kDof] 
	* SM[System_ij[jDof + 1] + kDof];
      sum2 += k_xx[kDof] 
	* SM[System_ij[jDof + 1] + kDof] 
	* SM[System_ij[jDof + 2] + kDof];
    }

    SM[System_ij[jDof + 1] + jDof + 1] -= sum1;
    k_xx[jDof + 1] = SM[System_ij[jDof + 1] + jDof + 1];

    SM[System_ij[jDof + 2] + jDof + 1] -= sum2;
    SM[System_ij[jDof + 2] + jDof + 1] /= k_xx[jDof + 1];
    
    sum2 = 0.0;

    for (kDof = System_minIndexs[jDof]; kDof < jDof + 2; kDof++) {
      sum2 += k_xx[kDof] 
	* SM[System_ij[jDof + 2] + kDof] 
	* SM[System_ij[jDof + 2] + kDof];
    }

    SM[System_ij[jDof + 2] + jDof + 2] -= sum2;
    k_xx[jDof + 2] = SM[System_ij[jDof + 2] + jDof + 2];

    for (kDof = System_minIndexs[jDof]; kDof < jDof; kDof++) {
      k_xx_jx0[kDof] = k_xx[kDof] * SM[System_ij[jDof + 0] + kDof];
      k_xx_jx1[kDof] = k_xx[kDof] * SM[System_ij[jDof + 1] + kDof];
      k_xx_jx2[kDof] = k_xx[kDof] * SM[System_ij[jDof + 2] + kDof];
    }

    for (iDof = jDof + 3; iDof < System_nDofs; iDof += 3) {
      double sum0_0 = 0.0;
      double sum1_0 = 0.0;
      double sum2_0 = 0.0;
      double sum0_1 = 0.0;
      double sum1_1 = 0.0;
      double sum2_1 = 0.0;
      double sum0_2 = 0.0;
      double sum1_2 = 0.0;
      double sum2_2 = 0.0;
      int iMax;

      if (jDof < System_minIndexs[iDof]) {
	continue;
      }

      if (System_minIndexs[iDof] > System_minIndexs[jDof]) {
	iMax = System_minIndexs[iDof];
      } else {
	iMax = System_minIndexs[jDof];
      }

      for (kDof = iMax; kDof < jDof; kDof++) {
	double da_sb0_0 = k_xx_jx0[kDof];
	double da_sb0_1 = k_xx_jx1[kDof];
	double da_sb0_2 = k_xx_jx2[kDof];
	double db0 = SM[System_ij[iDof + 0] + kDof];
	double db1 = SM[System_ij[iDof + 1] + kDof];
	double db2 = SM[System_ij[iDof + 2] + kDof];

	sum0_0 += da_sb0_0 * db0;
	sum0_1 += da_sb0_0 * db1;
	sum0_2 += da_sb0_0 * db2;

	sum1_0 += da_sb0_1 * db0;
	sum1_1 += da_sb0_1 * db1;
	sum1_2 += da_sb0_1 * db2;

	sum2_0 += da_sb0_2 * db0;
	sum2_1 += da_sb0_2 * db1;
	sum2_2 += da_sb0_2 * db2;
      }

#ifdef MEASURE
      if (0 < jDof - iMax) {
	int count = jDof - iMax;

	AutoLinAS_AddOp (M_ADD_OP, count * 9);
	AutoLinAS_AddOp (M_MUL_OP, count * 9);
	AutoLinAS_AddIO (M_MEMORY, count * 6);
      }
#endif

      SM[System_ij[iDof + 0] + jDof + 0] -= sum0_0;
      SM[System_ij[iDof + 1] + jDof + 0] -= sum0_1;
      SM[System_ij[iDof + 2] + jDof + 0] -= sum0_2;
      SM[System_ij[iDof + 0] + jDof + 0] /= k_xx[jDof + 0];
      SM[System_ij[iDof + 1] + jDof + 0] /= k_xx[jDof + 0];
      SM[System_ij[iDof + 2] + jDof + 0] /= k_xx[jDof + 0];
      
      sum1_0 += k_xx[jDof + 0] 
	* SM[System_ij[jDof + 1] + jDof + 0] 
	* SM[System_ij[iDof + 0] + jDof + 0];
      sum1_1 += k_xx[jDof + 0] 
	* SM[System_ij[jDof + 1] + jDof + 0] 
	* SM[System_ij[iDof + 1] + jDof + 0];
      sum1_2 += k_xx[jDof + 0] 
	* SM[System_ij[jDof + 1] + jDof + 0] 
	* SM[System_ij[iDof + 2] + jDof + 0];
      
      SM[System_ij[iDof + 0] + jDof + 1] -= sum1_0;
      SM[System_ij[iDof + 1] + jDof + 1] -= sum1_1;
      SM[System_ij[iDof + 2] + jDof + 1] -= sum1_2;
      SM[System_ij[iDof + 0] + jDof + 1] /= k_xx[jDof + 1];
      SM[System_ij[iDof + 1] + jDof + 1] /= k_xx[jDof + 1];
      SM[System_ij[iDof + 2] + jDof + 1] /= k_xx[jDof + 1];

      sum2_0 += k_xx[jDof + 0] 
	* SM[System_ij[jDof + 2] + jDof + 0] 
	* SM[System_ij[iDof + 0] + jDof + 0];
      sum2_1 += k_xx[jDof + 0] 
	* SM[System_ij[jDof + 2] + jDof + 0] 
	* SM[System_ij[iDof + 1] + jDof + 0];
      sum2_2 += k_xx[jDof + 0] 
	* SM[System_ij[jDof + 2] + jDof + 0] 
	* SM[System_ij[iDof + 2] + jDof + 0];

      sum2_0 += k_xx[jDof + 1] 
	* SM[System_ij[jDof + 2] + jDof + 1] 
	* SM[System_ij[iDof + 0] + jDof + 1];
      sum2_1 += k_xx[jDof + 1] 
	* SM[System_ij[jDof + 2] + jDof + 1] 
	* SM[System_ij[iDof + 1] + jDof + 1];
      sum2_2 += k_xx[jDof + 1] 
	* SM[System_ij[jDof + 2] + jDof + 1] 
	* SM[System_ij[iDof + 2] + jDof + 1];
      
      SM[System_ij[iDof + 0] + jDof + 2] -= sum2_0;
      SM[System_ij[iDof + 1] + jDof + 2] -= sum2_1;
      SM[System_ij[iDof + 2] + jDof + 2] -= sum2_2;
      SM[System_ij[iDof + 0] + jDof + 2] /= k_xx[jDof + 2];
      SM[System_ij[iDof + 1] + jDof + 2] /= k_xx[jDof + 2];
      SM[System_ij[iDof + 2] + jDof + 2] /= k_xx[jDof + 2];

#ifdef MEASURE
      AutoLinAS_AddOp (M_ADD_OP, 24);
      AutoLinAS_AddOp (M_MUL_OP, 30);
      AutoLinAS_AddOp (M_OTHER_OP, 9);
      AutoLinAS_AddIO (M_MEMORY, 40);
#endif
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nDofs * 6);
  AutoLinAS_AddOp (M_OTHER_OP, System_nDofs * 3);
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 12);
#endif
}

#endif

/* ***TUNED*** */
#if (Node_nDofs == 1)

void AutoLinAS_INSIDE_ForwardReduction (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    double sum = 0.0;

    for (jDof = System_minIndexs[iDof]; jDof < iDof; jDof++) {
      sum += SM[System_ij[iDof] + jDof] * SV1[jDof];
    }
    SV1[iDof] -= sum;
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 2);
#endif
}

#endif
#if (Node_nDofs == 2)

void AutoLinAS_INSIDE_ForwardReduction (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof += 2) {
    double sum0 = 0.0;
    double sum1 = 0.0;

    for (jDof = System_minIndexs[iDof]; jDof < iDof; jDof++) {
      sum0 += SM[System_ij[iDof + 0] + jDof] * SV1[jDof];
      sum1 += SM[System_ij[iDof + 1] + jDof] * SV1[jDof];
    }
    SV1[iDof + 0] -= sum0;

    sum1 += SM[System_ij[iDof + 1] + iDof] * SV1[iDof];
    SV1[iDof + 1] -= sum1;
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 1.5);
#endif
}

#endif
#if (Node_nDofs == 3)

void AutoLinAS_INSIDE_ForwardReduction (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof += 3) {
    double sum0 = 0.0;
    double sum1 = 0.0;
    double sum2 = 0.0;

    for (jDof = System_minIndexs[iDof]; jDof < iDof; jDof++) {
      sum0 += SM[System_ij[iDof + 0] + jDof] * SV1[jDof];
      sum1 += SM[System_ij[iDof + 1] + jDof] * SV1[jDof];
      sum2 += SM[System_ij[iDof + 2] + jDof] * SV1[jDof];
    }
    SV1[iDof + 0] -= sum0;

    sum1 += SM[System_ij[iDof + 1] + iDof] * SV1[iDof];
    sum2 += SM[System_ij[iDof + 2] + iDof] * SV1[iDof];
    SV1[iDof + 1] -= sum1;

    sum2 += SM[System_ij[iDof + 2] + iDof + 1] * SV1[iDof + 1];
    SV1[iDof + 2] -= sum2;
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 1.333);
#endif
}

#endif

void AutoLinAS_INSIDE_BackwardSubstitution (void)
{
  int iDof, jDof;
  int maxBandWidth;

  maxBandWidth = -System_minIndexs[0];
  for (iDof = 1; iDof < System_nDofs; iDof++) {
    if (maxBandWidth < iDof - System_minIndexs[iDof]) {
      maxBandWidth = iDof - System_minIndexs[iDof];
    }
  }

  SV1[System_nDofs - 1] 
    /= SM[System_ij[System_nDofs - 1] + System_nDofs - 1];

  for (iDof = System_nDofs - 2; 0 <= iDof; iDof--) {
    int jMin;

    jMin = iDof + maxBandWidth;
    if (System_nDofs - 1 < jMin) {
      jMin = System_nDofs - 1;
    } else {
      while (iDof < System_minIndexs[jMin] && jMin != iDof) {
	jMin--;
      }
    }

    SV1[iDof] /= SM[System_ij[iDof] + iDof];

    for (jDof = iDof + 1; jDof <= jMin; jDof++) {
      if (System_minIndexs[jDof] <= iDof) {
        SV1[iDof] -= SM[System_ij[jDof] + iDof] * SV1[jDof];
      }
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_OTHER_OP, System_nDofs);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 2);
#endif
}

