/* vector_renumber_2d_tri.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../lib/autolinas_predefined.h"



/* Node */

/* 2D plane strain, linear elastic */

int Node_nNodes;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

int Node_hasMarks[Node_maxNodes];



/* Element */

/* 2D plane strain, linear elastic */

int Element_nElements;

/* old connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];

int Element_colors[Element_maxElements];
int Element_oldElementIds[Element_maxElements];

/* new, renumbered connectivity */
int Element_newNodeIds[Element_maxElements][Element_nNodes];



/* Color */

#define MAX_COLORS 100
int Color_nColors;
int Color_nElements[MAX_COLORS];



void ReadMesh (FILE* fp)
{
  int iElement;
  int iNode_Elem;
  int iNode;
  
  fscanf (fp, "%d", &Element_nElements);
  fprintf (stderr, " nElements %d \n", Element_nElements);
  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d ", 
	      &Element_nodeIds[iElement][iNode_Elem]);
    }
  }

  fscanf (fp, "%d", &Node_nNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf %lf", 
	    &Node_XVec[iNode][0],
	    &Node_XVec[iNode][1]);
  }
}

void Renumber ()
{
  int currentColor;
  int nMarkedElements;
  int iColor;
  int iElement;
  int iNode_Elem;
  int iNode;

  /* perform multi-coloring of elements */

  currentColor = 0;
  nMarkedElements = 0;

  for (iColor = 0; iColor < MAX_COLORS; iColor++) {
    Color_nElements[iColor] = 0;
  }

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    Element_colors[iElement] = -1;
  }
    
  for (;;) {
    if (nMarkedElements == Element_nElements) break;
    assert(currentColor < MAX_COLORS);

    for (iNode = 0; iNode < Node_nNodes; iNode++) {
      Node_hasMarks[iNode] = 0;
    }
    
    for (iElement = 0; iElement < Element_nElements; iElement++) {
      int anyNodeOfTheElementHasMark;
      
      if (0 <= Element_colors[iElement]) continue;

      anyNodeOfTheElementHasMark = 0;
      for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
	int nodeId = Element_nodeIds[iElement][iNode_Elem];

	if (Node_hasMarks[nodeId]) {
	  anyNodeOfTheElementHasMark = 1;
	  break;
	}
      }
      
      if (!anyNodeOfTheElementHasMark) {
	for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
	  int nodeId = Element_nodeIds[iElement][iNode_Elem];
	  
	  Node_hasMarks[nodeId] = 1;
	}
	Element_colors[iElement] = currentColor;
	Color_nElements[currentColor]++;
	Element_oldElementIds[nMarkedElements] = iElement;
	nMarkedElements++;
      }
    }
    
    currentColor++;
  }

  Color_nColors = currentColor;

  fprintf (stderr, " nColors %d \n", Color_nColors);



  /* set up <Element_newNodeIds> */

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    int oldElementId = Element_oldElementIds[iElement];

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      Element_newNodeIds[iElement][iNode_Elem]
	= Element_nodeIds[oldElementId][iNode_Elem];
    }
  }
}

void WriteMesh (FILE* fp)
{
  int iElement;
  int iNode_Elem;
  int iNode;
  
  fprintf (fp, "%d \n", Element_nElements);
  for (iElement = 0; iElement < Element_nElements; iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fprintf (fp, "%d ", 
	       Element_newNodeIds[iElement][iNode_Elem]);
    }
    fprintf (fp, "\n");
  }

  fprintf (fp, "%d \n", Node_nNodes);
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%f %f \n", 
	     Node_XVec[iNode][0],
	     Node_XVec[iNode][1]);
  }
}

void WriteColor (FILE* fp)
{
  int iColor;

  fprintf (fp, "%d \n", Color_nColors);
  for (iColor = 0; iColor < Color_nColors; iColor++) {
    fprintf (fp, "%d ", 
	     Color_nElements[iColor]);
  }
  fprintf (fp, "\n");
}

int main (int argc, char* argv[])
{
  FILE* fp;

  if (argc < 4) {
    fprintf (stderr, 
	     " vector_renumber_2d_tri input_mesh output_mesh color_file \n");
    exit (1);
  }
  
  fp = fopen (argv[1], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open input mesh file  \n");
    exit (1);
  }
  ReadMesh (fp);
  fclose (fp);

  Renumber ();

  fp = fopen (argv[2], "w");
  WriteMesh (fp);
  fclose (fp);

  fp = fopen (argv[3], "w");
  WriteColor (fp);
  fclose (fp);

  return 0;
}



