/* make_advoutput.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <Adv/AdvDocument.h>

#include "../lib/autolinas_predefined.h"



int Domain_nDomains;
int Node_nNodes;
int Element_nElements;

/* displacement analysis result */
double Node_uVec[Node_maxNodes][Node_nDofs];

/* equivalent stress analysis result */
double Node_sigma[Node_maxNodes];



void ReadInputFile (char *fileName)
{
  FILE *fp;
  int iElement;
  int iNode_Elem;
      
  fp = fopen (fileName, "r");
  assert(fp);

  /* element connectivity */
  
  fscanf (fp, "%d", &Element_nElements);
  
  for (iElement = 0; iElement < Element_nElements; iElement++) {
    
    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      int dummy;
      
      fscanf (fp, "%d", &dummy);
    }
    
    fprintf (fp, "\n");
  }
  
  /* nodal coordinate */
  
  fscanf (fp, "%d", &Node_nNodes);

  fclose (fp);
}

void ReadOutputFile (char *fileName)
{
  FILE *fp;
  int iNode;
      
  fp = fopen (fileName, "r");
  assert(fp);

  /* displacement */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf %lf %lf", 
	    &Node_uVec[iNode][0],
	    &Node_uVec[iNode][1],
	    &Node_uVec[iNode][2]);
  }
  
  /* equivalent stress */
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fscanf (fp, "%lf", 
	    &Node_sigma[iNode]);
  }

  fclose (fp);
}

void WriteDisplacement (AdvDocFile *advFile)
{
  AdvDocument* doc;
  adv_off_t offset;
  int iNode;

  doc = adv_dio_create (advFile, adv_dio_make_documentid ("Displacement"));

  adv_dio_set_property (doc, "content_type", "FEGenericAttribute");
  adv_dio_set_property_int32 (doc, "num_items", Node_nNodes);
  adv_dio_set_property (doc, "fega_type", "AllNodeVariable");
  adv_dio_set_property (doc, "format", "f8f8f8");
  adv_dio_set_property (doc, "label", "Displacement");
  adv_dio_set_property_int32 (doc, "index_byte", 4);
  
  offset = 0;
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    offset += adv_dio_write_float64 
      (doc, offset, Node_uVec[iNode][0]);
    offset += adv_dio_write_float64 
      (doc, offset, Node_uVec[iNode][1]);
    offset += adv_dio_write_float64 
      (doc, offset, Node_uVec[iNode][2]);
  }

  adv_dio_close (doc);
}

void WriteEquivalentStress (AdvDocFile *advFile)
{
  AdvDocument* doc;
  adv_off_t offset;
  int iNode;

  doc = adv_dio_create (advFile, adv_dio_make_documentid 
			("NodalEquivalentStress"));

  adv_dio_set_property (doc, "content_type", "FEGenericAttribute");
  adv_dio_set_property_int32 (doc, "num_items", Node_nNodes);
  adv_dio_set_property (doc, "fega_type", "AllNodeVariable");
  adv_dio_set_property (doc, "format", "f8");
  adv_dio_set_property (doc, "label", "NodalEquivalentStress");
  adv_dio_set_property_int32 (doc, "index_byte", 4);
  
  offset = 0;
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    offset += adv_dio_write_float64 
      (doc, offset, Node_sigma[iNode]);
  }

  adv_dio_close (doc);
}

void WriteAdvIOFile (char* fileName)
{
  AdvDocFile *advFile;

  advFile = adv_dio_file_open (fileName, "c");

  WriteDisplacement (advFile);
  WriteEquivalentStress (advFile);

  adv_dio_file_close (advFile);
}

int main (int argc, char* argv[])
{
  int iDomain;

  if (argc < 4) {
    fprintf (stderr, 
	     " make_advoutput n_domains input_file output_file \n");
    exit (1);
  }
  
  Domain_nDomains = atoi (argv[1]);

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    char fileName[1024];

    sprintf (fileName, "%s.%d", argv[2], iDomain);
    ReadInputFile (fileName);

    sprintf (fileName, "%s.%d", argv[3], iDomain);
    ReadOutputFile (fileName);

    sprintf (fileName, "advhddm_out_%d.adv", iDomain);
    WriteAdvIOFile (fileName);
  }

  return 0;
}

