/* header file for autolinas.c */

typedef int (*ElementNodeIdCallbackType) 
  (int iElement, int iNode_Elem);

typedef void (*ElementMatrixCallbackType) 
  (int iElement, 
   double em[Element_nDofs][Element_nDofs]);

void AutoLinAS_Initialize 
(int nNodes, 
 int nElements, 
 ElementNodeIdCallbackType callback);

/* {x_fixed} = {0} */
void AutoLinAS_ClearConstraint (void);

/* {x_fixed}[iNode][iDof_Node] = value */ 
void AutoLinAS_SetConstraint
(int iNode, int iDof_Node, double value);

/* {b} = {0} */
void AutoLinAS_ClearRightHandSide (void);

/* {b}[iNode][iDof_Node] += value */ 
void AutoLinAS_AddValueIntoRightHandSide
(int iNode, int iDof_Node, double value);

/* {b} += {ev} (element-wize right hand side vector) */
void AutoLinAS_AddElementVectorIntoRightHandSide
(int iElement, double ev[Element_nDofs]);

/* {x_initial} = {0} */
void AutoLinAS_ClearInitialResult (void);

/* {x_initial} = {x} */
void AutoLinAS_CopyResultToInitialResult (void);

/* {x_initial}[iNode][iDof_Node] = value */ 
void AutoLinAS_SetInitialResult
(int iNode, int iDof_Node, double value);

/* get component of {x} */
double AutoLinAS_Result (int iNode, int iDof_Node);

/* [A] += [em] (element-wize coefficient matrix) */
void AutoLinAS_SetElementMatrixCallback 
(ElementMatrixCallbackType callback);

/* apply constraint and prepare solving */
void AutoLinAS_Fix (void);

/* solve {x} of equation [A]{x} == {b} */
/* with constraint {x_fixed} */
void AutoLinAS_Solve (void);

/* solve {x} of equation [A]{x} == {b} */
/* with constraint {x_fixed} */
/* initial value of {x} is {x_initial} */
void AutoLinAS_SolveUsingInitialResult (void);



