/* autolinas_dscg.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autonosim.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dscg_inside.h"



#if 0
#define DEBUG_MSG
#endif



/* module variables */

double SVModified[System_maxDofs];

double SVScale[System_maxDofs];



static void SolveConjugateGradientWithDiagonalScaling (void)
{
  static double x[System_maxDofs];
  static double b[System_maxDofs];
  static double Kx[System_maxDofs];
  static double r[System_maxDofs];
  static double p[System_maxDofs];
  static double Kp[System_maxDofs];
  double bb;
  double rr_old;
  double rr;
  int iDof;
#ifdef MEASURE
  int iterationCount = 0;
#endif
  int nGlobalNodes = AutoNOSim_NGlobalEntitys ();

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    b[iDof] = SV1[iDof];
    x[iDof] = SVInitial[iDof];
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    b[iDof] /= SVScale[iDof];  
    x[iDof] *= SVScale[iDof];
  }

  AutoLinAS_INSIDE_MultiplyMatrixAndVector (x, 
					    Kx);

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    r[iDof] = b[iDof] - Kx[iDof];
    p[iDof] = r[iDof];
  }

  bb = AutoNOSim_ScalarProduct (Node_nDofs, b, b);
  rr_old = AutoNOSim_ScalarProduct (Node_nDofs, r, r);

#ifdef DEBUG_MSG
  fprintf (stderr, " bb %e \n", bb);
  fprintf (stderr, " rr_old %e \n", rr_old);
#endif

  rr = 0.0;

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nDofs * 3);
  AutoLinAS_AddOp (M_MUL_OP, System_nDofs * 3);
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 11);
#endif

  do {
    double rp;
    double pKp;
    double alpha;
    double beta;

    AutoLinAS_INSIDE_MultiplyMatrixAndVector (p,
					      Kp);

#ifdef MEASURE
    AutoLinAS_ClearTime ();
#endif

    AutoNOSim_GetTwoScalarProducts (Node_nDofs, 
				    r, p, &rp, 
				    p, Kp, &pKp);

#ifdef DEBUG_MSG
    fprintf (stderr, " rp %e \n", rp);
    fprintf (stderr, " pKp %e \n", pKp);
#endif

    alpha = rp / pKp;

    for (iDof = 0; iDof < System_nDofs; iDof++) {
      x[iDof] += p[iDof] * alpha;
      r[iDof] -= Kp[iDof] * alpha;
    }

    rr = AutoNOSim_ScalarProduct (Node_nDofs, r, r);

#ifdef DEBUG_MSG
    fprintf (stderr, " rr %e \n", rr);
#endif

    beta = rr / rr_old;
    rr_old = rr;

    for (iDof = 0; iDof < System_nDofs; iDof++) {
      p[iDof] = r[iDof] + p[iDof] * beta;
    }

#ifdef MEASURE
    AutoLinAS_CountTime (4);
#endif

#ifdef MEASURE
    AutoLinAS_AddOp (M_ADD_OP, System_nDofs * 6);
    AutoLinAS_AddOp (M_MUL_OP, System_nDofs * 6);
    AutoLinAS_AddIO (M_MEMORY, System_nDofs * 12);
#endif

#ifdef MEASURE
    iterationCount++;
    if (iterationCount % 100 == 0) {
      fprintf (stderr, " iteration %d \n", iterationCount);
      fprintf (stderr, " norm (bb) %e norm (rr) %e \n", 
	       sqrt (bb / (Node_nDofs * nGlobalNodes)), 
	       sqrt (rr / (Node_nDofs * nGlobalNodes)));
    }
#endif

  } while (sqrt (bb / (Node_nDofs * nGlobalNodes)) * 0.000001 
	   < sqrt (rr / (Node_nDofs * nGlobalNodes)));

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    x[iDof] /= SVScale[iDof];
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SV1[iDof] = x[iDof];
  }

#ifdef MEASURE
  fprintf (stderr, " total iteration %d \n", iterationCount);
#endif

#ifdef MEASURE
  AutoLinAS_AddOp (M_OTHER_OP, System_nDofs);
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 4);
#endif

#ifdef MEASURE
  AutoLinAS_DivideTimeCounters (iterationCount);
#endif
}

void AutoLinAS_Initialize 
(int nNodes, 
 int nElements, 
 ElementNodeIdCallbackType callback)
{
  AutoLinAS_INSIDE_InitializeSystem (nNodes, nElements, callback);
  AutoLinAS_INSIDE_InitializeMatrix ();
}

void AutoLinAS_Fix (void)
{
  int iElement;

  AutoLinAS_INSIDE_Clear ();

  for (iElement = 0; iElement < System_nElements; iElement++) {
    AutoLinAS_INSIDE_AssembleElementMatrix (iElement);
  }

  AutoLinAS_INSIDE_SetDirichletCondition ();
  AutoLinAS_INSIDE_MakeScalingVector ();
  AutoLinAS_INSIDE_ScaleMatrix ();
}

void AutoLinAS_Solve (void)
{
  AutoLinAS_ClearInitialResult ();
  AutoLinAS_SolveUsingInitialResult ();
}

void AutoLinAS_SolveUsingInitialResult (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SV1[iDof] = SV0[iDof];
    SV1[iDof] += SVModified[iDof];
    if (SVFixedFlag[iDof]) {
      SV1[iDof] = SVFixed[iDof];
    }
  }

  AutoNOSim_Barrier ();

#ifdef MEASURE_TIME
  AutoLinAS_InitializeMeasure ();
#endif

  SolveConjugateGradientWithDiagonalScaling ();

#ifdef MEASURE_TIME
  AutoLinAS_TerminateMeasure ();

  AutoLinAS_PrintTimeCounters ();
#endif
}

