/* autolinas_dscg_mt.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autonosim.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_multithread_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dscg_inside.h"



/* module variables */

double SVModified[System_maxDofs];

double SVScale[System_maxDofs];



extern void AutoLinAS_INSIDE_MultiplyMatrixAndVector_MT
(int myThreadId, int nThreads, 
 int startNodeId, int endNodeId,
 double sv0[System_maxDofs], 
 double sv1[System_maxDofs]);

static void SolveConjugateGradientWithDiagonalScaling 
(int myThreadId, int nThreads, 
 int startNodeId, int endNodeId)
{
  static double x[System_maxDofs];
  static double b[System_maxDofs];
  static double Kx[System_maxDofs];
  static double r[System_maxDofs];
  static double p[System_maxDofs];
  static double Kp[System_maxDofs];
  double bb;
  double rr_old;
  double rr;
  int iDof;
#ifdef MEASURE
  int iterationCount = 0;
#endif
  int nGlobalNodes = AutoNOSim_NGlobalEntitys ();
  int startDofId = startNodeId * Node_nDofs;
  int endDofId = endNodeId * Node_nDofs;

  for (iDof = startDofId; iDof < endDofId; iDof++) {
    b[iDof] = SV1[iDof];
    x[iDof] = SVInitial[iDof];
  }

  for (iDof = startDofId; iDof < endDofId; iDof++) {
    b[iDof] /= SVScale[iDof];  
    x[iDof] *= SVScale[iDof];
  }

  AutoLinAS_INSIDE_MultiplyMatrixAndVector_MT 
    (myThreadId, nThreads, startNodeId, endNodeId, 
     x, 
     Kx);

  for (iDof = startDofId; iDof < endDofId; iDof++) {
    r[iDof] = b[iDof] - Kx[iDof];
    p[iDof] = r[iDof];
  }

  bb = 0.0;
  rr_old = 0.0;
  for (iDof = startDofId; iDof < endDofId; iDof++) {
    if (!AutoNOSim_EntityIsExcluded (iDof / Node_nDofs)) {
      bb += b[iDof] * b[iDof];
      rr_old += r[iDof] * r[iDof];
    }
  }

  AutoLinAS_GetThreadSum2 (&bb, &rr_old);
  if (myThreadId == 0) {
    AutoNOSim_GetSum2 (&bb, &rr_old);
  } else {
    bb = 0.0;
    rr_old = 0.0;
  }
  AutoLinAS_GetThreadSum2 (&bb, &rr_old);

  rr = 0.0;

#ifdef MEASURE
  if (myThreadId == 0) {
    AutoLinAS_AddOp (M_ADD_OP, System_nDofs * 3);
    AutoLinAS_AddOp (M_MUL_OP, System_nDofs * 3);
    AutoLinAS_AddIO (M_MEMORY, System_nDofs * 11);
  }
#endif

  do {
    double rp;
    double pKp;
    double alpha;
    double beta;

    AutoLinAS_INSIDE_MultiplyMatrixAndVector_MT 
      (myThreadId, nThreads, startNodeId, endNodeId, 
       p,
       Kp);
    
#ifdef MEASURE
    if (myThreadId == 0) {
      AutoLinAS_ClearTime ();
    }
#endif

    rp = 0.0;
    pKp = 0.0;
    for (iDof = startDofId; iDof < endDofId; iDof++) {
      if (!AutoNOSim_EntityIsExcluded (iDof / Node_nDofs)) {
	rp += r[iDof] * p[iDof];
	pKp += p[iDof] * Kp[iDof];
      }
    }

    AutoLinAS_GetThreadSum2 (&rp, &pKp);
    if (myThreadId == 0) {
      AutoNOSim_GetSum2 (&rp, &pKp);
    } else {
      rp = 0.0;
      pKp = 0.0;
    }
    AutoLinAS_GetThreadSum2 (&rp, &pKp);

    alpha = rp / pKp;

    for (iDof = startDofId; iDof < endDofId; iDof++) {
      x[iDof] += p[iDof] * alpha;
      r[iDof] -= Kp[iDof] * alpha;
    }

    rr = 0.0;
    for (iDof = startDofId; iDof < endDofId; iDof++) {
      if (!AutoNOSim_EntityIsExcluded (iDof / Node_nDofs)) {
	rr += r[iDof] * r[iDof];
      }
    }

    AutoLinAS_GetThreadSum (&rr);
    if (myThreadId == 0) {
      AutoNOSim_GetSum (&rr);
    } else {
      rr = 0.0;
    }
    AutoLinAS_GetThreadSum (&rr);

    beta = rr / rr_old;
    rr_old = rr;

    for (iDof = startDofId; iDof < endDofId; iDof++) {
      p[iDof] = r[iDof] + p[iDof] * beta;
    }

#ifdef MEASURE
    if (myThreadId == 0) {
      AutoLinAS_CountTime (4);
    }
#endif
      
#ifdef MEASURE
    if (myThreadId == 0) {
      AutoLinAS_AddOp (M_ADD_OP, System_nDofs * 6);
      AutoLinAS_AddOp (M_MUL_OP, System_nDofs * 6);
      AutoLinAS_AddIO (M_MEMORY, System_nDofs * 12);
    }
#endif

#ifdef MEASURE
    iterationCount++;
    if (myThreadId == 0) {
      if (iterationCount % 100 == 0) {
	fprintf (stderr, " iteration %d \n", iterationCount);
	fprintf (stderr, " norm (bb) %e norm (rr) %e \n", 
		 sqrt (bb / (Node_nDofs * nGlobalNodes)), 
		 sqrt (rr / (Node_nDofs * nGlobalNodes)));
      }
    }
#endif

  } while (sqrt (bb / (Node_nDofs * nGlobalNodes)) * 0.000001 
	   < sqrt (rr / (Node_nDofs * nGlobalNodes)));
  
  AutoLinAS_BarrierThreads ();

  for (iDof = startDofId; iDof < endDofId; iDof++) {
    x[iDof] /= SVScale[iDof];
  }

  for (iDof = startDofId; iDof < endDofId; iDof++) {
    SV1[iDof] = x[iDof];
  }

  AutoLinAS_BarrierThreads ();

#ifdef MEASURE
  if (myThreadId == 0) {
    fprintf (stderr, " total iteration %d \n", iterationCount);
  }
#endif

#ifdef MEASURE
  if (myThreadId == 0) {
    AutoLinAS_AddOp (M_OTHER_OP, System_nDofs);
    AutoLinAS_AddIO (M_MEMORY, System_nDofs * 4);
  }
#endif

#ifdef MEASURE
  if (myThreadId == 0) {
    AutoLinAS_DivideTimeCounters (iterationCount);
  }
#endif
}

void AutoLinAS_Initialize 
(int nNodes, 
 int nElements, 
 ElementNodeIdCallbackType callback)
{
  AutoLinAS_INSIDE_InitializeSystem (nNodes, nElements, callback);
  AutoLinAS_INSIDE_InitializeMatrix ();
}

void AutoLinAS_Fix (void)
{
  int iElement;

  AutoLinAS_INSIDE_Clear ();

  for (iElement = 0; iElement < System_nElements; iElement++) {
    AutoLinAS_INSIDE_AssembleElementMatrix (iElement);
  }

  AutoLinAS_INSIDE_SetDirichletCondition ();
  AutoLinAS_INSIDE_MakeScalingVector ();
  AutoLinAS_INSIDE_ScaleMatrix ();
}

void AutoLinAS_Solve (void)
{
  AutoLinAS_ClearInitialResult ();
  AutoLinAS_SolveUsingInitialResult ();
}

static void DscgThreadFunction (int myThreadId, int nThreads)
{
  int startNodeId, endNodeId;

  startNodeId = System_nNodes / nThreads * myThreadId;
  endNodeId = System_nNodes / nThreads * (myThreadId + 1);
  if (myThreadId == nThreads - 1) {
    endNodeId = System_nNodes;
  }



  fprintf (stderr, 
	   " myThreadId %d nThreads %d startNodeId %d endNodeId %d \n", 
	   myThreadId, nThreads, 
	   startNodeId, endNodeId);
  


  SolveConjugateGradientWithDiagonalScaling 
    (myThreadId, nThreads,
     startNodeId, endNodeId);
}

void AutoLinAS_SolveUsingInitialResult (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SV1[iDof] = SV0[iDof];
    SV1[iDof] += SVModified[iDof];
    if (SVFixedFlag[iDof]) {
      SV1[iDof] = SVFixed[iDof];
    }
  }

  AutoNOSim_Barrier ();

#ifdef MEASURE_TIME
  AutoLinAS_InitializeMeasure ();
#endif

  AutoLinAS_RunThreads (2, DscgThreadFunction);

#ifdef MEASURE_TIME
  AutoLinAS_TerminateMeasure ();

  AutoLinAS_PrintTimeCounters ();
#endif
}

