/* autolinas_dscg_nonzero_ol.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autonosim.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dscg_inside.h"



/* symmetric matrix */
/* iterative method, conjugate gradient */
/* diagonal scaling preconditioner */
/* sparse matrix, non-zero component storage */

/* tuned using DOF x DOF loop unrolling */
/* overlapping of communication and calculation */

/*

  # * - - * - - - - -        #**
  * # * * - - - - - -        *#**
  - * # - * - - - - -        *#*
  - * - # - * * - - -        *#**
  * - * - # * - * * -   --\  **#***
  - - - * * # - - - -   --/  **#
  - - - * - - # - * *        *#**
  - - - - * - - # * *        *#**
  - - - - * - * * # -        ***#
  - - - - - - * * - #        **#

  full matrix (non-zero component storage)

*/



/* module variables */

#if (Space_nDimensions == 2)

#if (Element_nNodes == 3) 
#define System_maxNodesOfOneRow 60
#define System_nAverageNodesOfOneRow 20
#endif

#endif

#if (Space_nDimensions == 3)

#if (Element_nNodes == 4) 
#define System_maxNodesOfOneRow 120
#define System_nAverageNodesOfOneRow 20
#endif

#if (Element_nNodes == 10) 
#define System_maxNodesOfOneRow 240
#define System_nAverageNodesOfOneRow 40
#endif

#endif

static int System_nIndexs[Node_maxNodes];
static int System_indexTmp[Node_maxNodes][System_maxNodesOfOneRow];
static int System_minIndexs[Node_maxNodes];
static int System_maxIndexs[Node_maxNodes];
static int System_index[Node_maxNodes * System_nAverageNodesOfOneRow];
static int System_nMatrixSize;

/* ***TUNED*** */
static double SM[Node_maxNodes * System_nAverageNodesOfOneRow 
		 * Node_nDofs * Node_nDofs];

static int Node_isOnBoundary[Node_maxNodes];



static void RegisterIndex (int iElement)
{
  int iNode_Elem, jNode_Elem;
  int k;

  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = (*System_elementNodeId) (iElement, iNode_Elem);

    for (jNode_Elem = 0; jNode_Elem < Element_nNodes; jNode_Elem++) {
      int jNode = (*System_elementNodeId) (iElement, jNode_Elem);
      int found = 0;

      for (k = 0; k < System_nIndexs[iNode]; k++) {
	if (System_indexTmp[iNode][k] == jNode) { 
	  found = 1;  break; 
	}
      }
      if (!found) {
	assert(System_nIndexs[iNode] < System_maxNodesOfOneRow);
	
	System_indexTmp[iNode][System_nIndexs[iNode]] = jNode;
	System_nIndexs[iNode]++;
      }
    }
  }
}

static void SortIndex (int iNode)
{
  static int dest[System_maxNodesOfOneRow];
  int i, j;
  
  for (i = 0; i < System_nIndexs[iNode]; i++) {
    int value = System_indexTmp[iNode][i];
    int value_index = i;

    for (j = 0; j < i; j++) {
      if (value < dest[j]) {
	value_index = j;
	break;
      }
    }
    for (j = i - 1; value_index <= j; j--) {
      dest[j + 1] = dest[j];
    }
    dest[value_index] = value;
  }
  for (i = 0; i < System_nIndexs[iNode]; i++) {
    System_indexTmp[iNode][i] = dest[i];
  }
}

static void CompressIndex ()
{
  int iNode;
  int k;
  int count;
  
  count = 0;
  for (iNode = 0; iNode < System_nNodes; iNode++) {
    System_minIndexs[iNode] = count;

    for (k = 0; k < System_nIndexs[iNode]; k++) {
      System_index[count] = System_indexTmp[iNode][k];
      count++;
    }

    System_maxIndexs[iNode] = count;
  }
}

void AutoLinAS_INSIDE_InitializeMatrix (void)
{
  int iNode;
  int iElement;
  int maxIndex, meanIndex;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    System_nIndexs[iNode] = 0; 
  }
  for (iElement = 0; iElement < System_nElements; iElement++) {
    RegisterIndex (iElement);
  }
  for (iNode = 0; iNode < System_nNodes; iNode++) {
    SortIndex (iNode); 
  }
  CompressIndex ();

  System_nMatrixSize = 0;
  maxIndex = 0;
  for (iNode = 0; iNode < System_nNodes; iNode++) {
    if (maxIndex < System_nIndexs[iNode]) {
      maxIndex = System_nIndexs[iNode];
    }
    System_nMatrixSize += System_nIndexs[iNode]; 
  }
  meanIndex = System_nMatrixSize / System_nNodes;
  System_nMatrixSize *= Node_nDofs * Node_nDofs;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    Node_isOnBoundary[iNode] 
      = AutoNOSim_EntityIsOnBoundary (iNode);
  }

#ifdef MEASURE
  fprintf (stderr, " nDofs %d \n", System_nDofs);
  fprintf (stderr, " maxIndex/node %d \n", maxIndex);
  fprintf (stderr, " meanIndex/node %d \n", meanIndex);
  fprintf (stderr, " NonZeroMatrix size %d \n", System_nMatrixSize);
#endif
}

void AutoLinAS_INSIDE_Clear (void)
{
  int iNode;
  int k;
  int iDof_Node, jDof_Node;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (k = System_minIndexs[iNode]; 
	 k < System_maxIndexs[iNode]; 
	 k++) {

      for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  SM[k * Node_nDofs * Node_nDofs 
	     + iDof_Node * Node_nDofs
	     + jDof_Node] = 0.0;
	}
      }
    }
  }
}

void AutoLinAS_INSIDE_AssembleElementMatrix (int iElement)
{
  double em[Element_nDofs][Element_nDofs];
  int iNode_Elem, jNode_Elem;
  int k;
  int iDof_Node, jDof_Node;
  
  assert(System_elementMatrix != NULL);

  (*System_elementMatrix) (iElement, 
			   em);

  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = (*System_elementNodeId) (iElement, iNode_Elem);

    for (jNode_Elem = 0; jNode_Elem < Element_nNodes; jNode_Elem++) {
      int jNode = (*System_elementNodeId) (iElement, jNode_Elem);
      
      for (k = System_minIndexs[iNode]; 
	   k < System_maxIndexs[iNode]; 
	   k++) {
	if (System_index[k] == jNode) {

	  for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
	    for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	      int iDof_Elem = iNode_Elem * Node_nDofs + iDof_Node;
	      int jDof_Elem = jNode_Elem * Node_nDofs + jDof_Node;

	      SM[k * Node_nDofs * Node_nDofs 
		 + iDof_Node * Node_nDofs 
		 + jDof_Node] 
		+= em[iDof_Elem][jDof_Elem];
	    }
	  }
	  break;
	}
      }      
    }
  }
}

/* ***TUNED*** */
void AutoLinAS_INSIDE_MultiplyMatrixAndVector
(double sv0[System_maxDofs], 
 double sv1[System_maxDofs])
{
  int iNode;
  int k;
#ifdef MEASURE
  int count0;
#endif

#ifdef MEASURE
  AutoLinAS_ClearTime ();
#endif

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    int iDof = iNode * Node_nDofs;
    
#if (Node_nDofs == 1)
    double sv1_isum = 0.0;
#endif
#if (Node_nDofs == 2)
    double sv1_isum0 = 0.0;
    double sv1_isum1 = 0.0;
#endif
#if (Node_nDofs == 3)
    double sv1_isum0 = 0.0;
    double sv1_isum1 = 0.0;
    double sv1_isum2 = 0.0;
#endif

    if (!Node_isOnBoundary[iNode]) continue;

    for (k = System_minIndexs[iNode]; 
	 k < System_maxIndexs[iNode]; 
	 k++) {
      int jNode = System_index[k];
      
#if (Node_nDofs == 1)
      sv1_isum += SM[k] * sv0[jNode];
#endif
#if (Node_nDofs == 2)
      double sm_0_0 = SM[k * 4 + 0];
      double sm_0_1 = SM[k * 4 + 1];
      double sm_1_0 = SM[k * 4 + 2];
      double sm_1_1 = SM[k * 4 + 3];
      
      double sv0_j0 = sv0[jNode * 2 + 0];
      double sv0_j1 = sv0[jNode * 2 + 1];
      
      sv1_isum0 
	+= sm_0_0 * sv0_j0 
	+ sm_0_1 * sv0_j1;
      sv1_isum1 
	+= sm_1_0 * sv0_j0 
	+ sm_1_1 * sv0_j1;
#endif
#if (Node_nDofs == 3)
      double sm_0_0 = SM[k * 9 + 0];
      double sm_0_1 = SM[k * 9 + 1];
      double sm_0_2 = SM[k * 9 + 2];
      double sm_1_0 = SM[k * 9 + 3];
      double sm_1_1 = SM[k * 9 + 4];
      double sm_1_2 = SM[k * 9 + 5];
      double sm_2_0 = SM[k * 9 + 6];
      double sm_2_1 = SM[k * 9 + 7];
      double sm_2_2 = SM[k * 9 + 8];
      
      double sv0_j0 = sv0[jNode * 3 + 0];
      double sv0_j1 = sv0[jNode * 3 + 1];
      double sv0_j2 = sv0[jNode * 3 + 2];
      
      sv1_isum0 
	+= sm_0_0 * sv0_j0 
	+ sm_0_1 * sv0_j1
	+ sm_0_2 * sv0_j2;
      sv1_isum1 
	+= sm_1_0 * sv0_j0 
	+ sm_1_1 * sv0_j1
	+ sm_1_2 * sv0_j2;
      sv1_isum2 
	+= sm_2_0 * sv0_j0 
	+ sm_2_1 * sv0_j1
	+ sm_2_2 * sv0_j2;
#endif
    }

#if (Node_nDofs == 1)
    sv1[iDof] = sv1_isum;
#endif
#if (Node_nDofs == 2)
    sv1[iDof + 0] = sv1_isum0;
    sv1[iDof + 1] = sv1_isum1;
#endif
#if (Node_nDofs == 3)
    sv1[iDof + 0] = sv1_isum0;
    sv1[iDof + 1] = sv1_isum1;
    sv1[iDof + 2] = sv1_isum2;
#endif
  }

#ifdef MEASURE
  AutoLinAS_CountTime (0);
#endif

#ifdef MEASURE
  AutoLinAS_ClearTime ();
#endif

  AutoNOSim_PreExchange (Node_nDofs, sv1);

#ifdef MEASURE
  AutoLinAS_CountTime (1);
#endif

#ifdef MEASURE
  AutoLinAS_ClearTime ();
#endif

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    int iDof = iNode * Node_nDofs;
    
#if (Node_nDofs == 1)
    double sv1_isum = 0.0;
#endif
#if (Node_nDofs == 2)
    double sv1_isum0 = 0.0;
    double sv1_isum1 = 0.0;
#endif
#if (Node_nDofs == 3)
    double sv1_isum0 = 0.0;
    double sv1_isum1 = 0.0;
    double sv1_isum2 = 0.0;
#endif

    if (Node_isOnBoundary[iNode]) continue;

    for (k = System_minIndexs[iNode]; 
	 k < System_maxIndexs[iNode]; 
	 k++) {
      int jNode = System_index[k];
      
#if (Node_nDofs == 1)
      sv1_isum += SM[k] * sv0[jNode];
#endif
#if (Node_nDofs == 2)
      double sm_0_0 = SM[k * 4 + 0];
      double sm_0_1 = SM[k * 4 + 1];
      double sm_1_0 = SM[k * 4 + 2];
      double sm_1_1 = SM[k * 4 + 3];
      
      double sv0_j0 = sv0[jNode * 2 + 0];
      double sv0_j1 = sv0[jNode * 2 + 1];
      
      sv1_isum0 
	+= sm_0_0 * sv0_j0 
	+ sm_0_1 * sv0_j1;
      sv1_isum1 
	+= sm_1_0 * sv0_j0 
	+ sm_1_1 * sv0_j1;
#endif
#if (Node_nDofs == 3)
      double sm_0_0 = SM[k * 9 + 0];
      double sm_0_1 = SM[k * 9 + 1];
      double sm_0_2 = SM[k * 9 + 2];
      double sm_1_0 = SM[k * 9 + 3];
      double sm_1_1 = SM[k * 9 + 4];
      double sm_1_2 = SM[k * 9 + 5];
      double sm_2_0 = SM[k * 9 + 6];
      double sm_2_1 = SM[k * 9 + 7];
      double sm_2_2 = SM[k * 9 + 8];
      
      double sv0_j0 = sv0[jNode * 3 + 0];
      double sv0_j1 = sv0[jNode * 3 + 1];
      double sv0_j2 = sv0[jNode * 3 + 2];
      
      sv1_isum0 
	+= sm_0_0 * sv0_j0 
	+ sm_0_1 * sv0_j1
	+ sm_0_2 * sv0_j2;
      sv1_isum1 
	+= sm_1_0 * sv0_j0 
	+ sm_1_1 * sv0_j1
	+ sm_1_2 * sv0_j2;
      sv1_isum2 
	+= sm_2_0 * sv0_j0 
	+ sm_2_1 * sv0_j1
	+ sm_2_2 * sv0_j2;
#endif
    }

#if (Node_nDofs == 1)
    sv1[iDof] = sv1_isum;
#endif
#if (Node_nDofs == 2)
    sv1[iDof + 0] = sv1_isum0;
    sv1[iDof + 1] = sv1_isum1;
#endif
#if (Node_nDofs == 3)
    sv1[iDof + 0] = sv1_isum0;
    sv1[iDof + 1] = sv1_isum1;
    sv1[iDof + 2] = sv1_isum2;
#endif
  }

#ifdef MEASURE
  AutoLinAS_CountTime (2);
#endif

#ifdef MEASURE
  AutoLinAS_ClearTime ();
#endif

  AutoNOSim_PostExchange (AUTONOSIM_OP_SUM, Node_nDofs, sv1);

#ifdef MEASURE
  AutoLinAS_CountTime (3);
#endif

#ifdef MEASURE
  count0 = System_nMatrixSize / (Node_nDofs * Node_nDofs);

  AutoLinAS_AddOp 
    (M_ADD_OP, count0 * Node_nDofs * Node_nDofs);
  AutoLinAS_AddOp 
    (M_MUL_OP, count0 * Node_nDofs * Node_nDofs);

  AutoLinAS_AddOp 
    (M_MEMORY, count0 / (sizeof (double) / sizeof (int)));
  AutoLinAS_AddIO 
    (M_MEMORY, count0 * (Node_nDofs * Node_nDofs + Node_nDofs));
  AutoLinAS_AddIO 
    (M_MEMORY, System_nNodes * Node_nDofs);
#endif

}

void AutoLinAS_INSIDE_SetDirichletCondition (void)
{
  int iNode;
  int k;
  int iDof_Node, jDof_Node;

  {
    int iDof;

    for (iDof = 0; iDof < System_nDofs; iDof++) {
      SVModified[iDof] = 0.0;
    }
  }

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;

      if (SVFixedFlag[iDof]) {
	
	for (k = System_minIndexs[iNode]; 
	     k < System_maxIndexs[iNode]; 
	     k++) {
	  int jNode = System_index[k];
	  
	  for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	    int jDof = jNode * Node_nDofs + jDof_Node;
	    
	    SVModified[jDof] 
	      -= SM[k * Node_nDofs * Node_nDofs 
		    + iDof_Node * Node_nDofs
		    + jDof_Node] 
	      * SVFixed[iDof];
	  }
	}
      }
    }
  }

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      
      for (k = System_minIndexs[iNode]; 
	   k < System_maxIndexs[iNode]; 
	   k++) {
	int jNode = System_index[k];
	
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  int jDof = jNode * Node_nDofs + jDof_Node;
	  
	  if (SVFixedFlag[jDof]) {
	    SM[k * Node_nDofs * Node_nDofs
	       + iDof_Node * Node_nDofs
	       + jDof_Node] = 0.0;
	  }
	}
      }
    }
  }

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;
      
      if (SVFixedFlag[iDof]) {
	
	for (k = System_minIndexs[iNode]; 
	     k < System_maxIndexs[iNode]; 
	     k++) {
	  
	  for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	    SM[k * Node_nDofs * Node_nDofs 
	       + iDof_Node * Node_nDofs 
	       + jDof_Node] = 0.0;
	  }
	}
      }
    }
  }

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;

      if (SVFixedFlag[iDof]) {
	
	for (k = System_minIndexs[iNode]; 
	     k < System_maxIndexs[iNode]; 
	     k++) {
	  int jNode = System_index[k];
	  
	  for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	    int jDof = jNode * Node_nDofs + jDof_Node;
	    
	    if (jDof == iDof) {
	      SM[k * Node_nDofs * Node_nDofs
		 + iDof_Node * Node_nDofs
		 + jDof_Node] = 1.0;
	    }
	  }
	}
      }
    }
  }

  AutoNOSim_Exchange (AUTONOSIM_OP_SUM, Node_nDofs, SVModified);
  AutoNOSim_Exchange (AUTONOSIM_OP_SUM, Node_nDofs, SVFixed);
}

static double GetValue (int iDof, int jDof)
{
  int iNode, jNode;
  int iDof_Node, jDof_Node;
  int k;

  iNode = iDof / Node_nDofs;
  iDof_Node = iDof % Node_nDofs;

  jNode = jDof / Node_nDofs;
  jDof_Node = jDof % Node_nDofs;

  for (k = System_minIndexs[iNode]; 
       k < System_maxIndexs[iNode]; 
       k++) {
    if (System_index[k] == jNode) {
      return SM[k * Node_nDofs * Node_nDofs 
		+ iDof_Node * Node_nDofs
		+ jDof_Node];
    }
  }
  return 0.0;
}

void AutoLinAS_INSIDE_MakeScalingVector (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVScale[iDof] = GetValue (iDof, iDof);
  }

  AutoNOSim_Exchange (AUTONOSIM_OP_SUM, Node_nDofs, SVScale);

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVScale[iDof] = sqrt (SVScale[iDof]);
  }
}

void AutoLinAS_INSIDE_ScaleMatrix (void)
{
  int iNode;
  int k;
  int iDof_Node, jDof_Node;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;

      for (k = System_minIndexs[iNode]; 
	   k < System_maxIndexs[iNode]; 
	   k++) {
	int jNode = System_index[k];
	
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  int jDof = jNode * Node_nDofs + jDof_Node;
	  
	  SM[k * Node_nDofs * Node_nDofs 
	     + iDof_Node * Node_nDofs
	     + jDof_Node] 
	    /= SVScale[iDof] * SVScale[jDof];
	}
      }
    }
  }
}

