/* autolinas_multithread_pthread.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <pthread.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_multithread_inside.h"



#define MAX_THREAD 1024

typedef struct {
  int myThreadId;
  int nThreads;
} ThreadArgument;

static int NThreads;
static pthread_t Threads[MAX_THREAD];
static ThreadArgument ThreadArguments[MAX_THREAD];

static AutoLinAS_ThreadFunction UserThreadFunction;

static int BarrierCount;
static pthread_mutex_t BarrierMutex;
static pthread_cond_t BarrierCond;
static double Sum0;
static double Sum1;



static void PthreadCallback (ThreadArgument *argument)
{
  (*UserThreadFunction) (argument->myThreadId, argument->nThreads);
}

void AutoLinAS_RunThreads 
(int nThreads, 
 AutoLinAS_ThreadFunction function)
{
  int iThread;

  NThreads = nThreads;
  UserThreadFunction = function;

  BarrierCount = NThreads;
  pthread_mutex_init (&BarrierMutex, NULL);
  pthread_cond_init (&BarrierCond, NULL);

  for (iThread = 0; iThread < NThreads; iThread++) {
    ThreadArguments[iThread].myThreadId = iThread;
    ThreadArguments[iThread].nThreads = NThreads;

    pthread_create (&Threads[iThread],
		    NULL,
		    (void *)PthreadCallback,
		    (void *)&ThreadArguments[iThread]);
  }
  
  for (iThread = 0; iThread < NThreads; iThread++) {
    pthread_join (Threads[iThread], NULL);
  }
}

void AutoLinAS_BarrierThreads (void)
{
  pthread_mutex_lock (&BarrierMutex);
  if (1 < BarrierCount) {
    BarrierCount--;
    pthread_cond_wait (&BarrierCond, &BarrierMutex);
  } else {
    BarrierCount = NThreads;
    pthread_cond_broadcast (&BarrierCond);
  }
  pthread_mutex_unlock (&BarrierMutex);
}

void AutoLinAS_GetThreadSum 
(double *sum_IO)
{
  AutoLinAS_BarrierThreads ();
  Sum0 = 0.0;
  AutoLinAS_BarrierThreads ();

  pthread_mutex_lock (&BarrierMutex);
  Sum0 += *sum_IO;
  if (1 < BarrierCount) {
    BarrierCount--;
    pthread_cond_wait (&BarrierCond, &BarrierMutex);
  } else {
    BarrierCount = NThreads;
    pthread_cond_broadcast (&BarrierCond);
  }
  pthread_mutex_unlock (&BarrierMutex);

  *sum_IO = Sum0;
  AutoLinAS_BarrierThreads ();
}

void AutoLinAS_GetThreadSum2 
(double *sum0_IO, double *sum1_IO)
{
  AutoLinAS_BarrierThreads ();
  Sum0 = 0.0;
  Sum1 = 0.0;
  AutoLinAS_BarrierThreads ();

  pthread_mutex_lock (&BarrierMutex);
  Sum0 += *sum0_IO;
  Sum1 += *sum1_IO;
  if (1 < BarrierCount) {
    BarrierCount--;
    pthread_cond_wait (&BarrierCond, &BarrierMutex);
  } else {
    BarrierCount = NThreads;
    pthread_cond_broadcast (&BarrierCond);
  }
  pthread_mutex_unlock (&BarrierMutex);

  *sum0_IO = Sum0;
  *sum1_IO = Sum1;
  AutoLinAS_BarrierThreads ();
}

