/* autolinas_system.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"



int System_nNodes;
int System_nElements;
int System_nDofs;

ElementNodeIdCallbackType System_elementNodeId;
ElementMatrixCallbackType System_elementMatrix;

int SVFixedFlag[System_maxDofs];  /* {x_fixed} */
double SVFixed[System_maxDofs];

double SVInitial[System_maxDofs];  /* {x_initial} */

double SV0[System_maxDofs];   /* {b} */

double SV1[System_maxDofs];   /* {x} */



void AutoLinAS_INSIDE_InitializeSystem 
(int nNodes, 
 int nElements, 
 ElementNodeIdCallbackType callback)
{
  assert(0 <= nNodes);
  assert(nNodes < Node_maxNodes);
  assert(0 <= nElements);
  assert(nNodes < Element_maxElements);
  assert(callback != NULL);

  System_nNodes = nNodes;
  System_nElements = nElements;
  System_nDofs = Node_nDofs * System_nNodes;
  System_elementNodeId = callback;
}

void AutoLinAS_ClearConstraint (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVFixedFlag[iDof] = 0;
    SVFixed[iDof] = 0.0;
  }
}

void AutoLinAS_SetConstraint
(int iNode, int iDof_Node, double value)
{
  int iDof = iNode * Node_nDofs + iDof_Node;

  assert(0 <= iNode);
  assert(iNode < System_nNodes);
  assert(0 <= iDof_Node);
  assert(iDof_Node < Node_nDofs);

  SVFixedFlag[iDof] = 1;
  SVFixed[iDof] = value;
}

void AutoLinAS_ClearRightHandSide (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SV0[iDof] = 0.0;
  }
}

void AutoLinAS_AddValueIntoRightHandSide
(int iNode, int iDof_Node, double value)
{
  int iDof = iNode * Node_nDofs + iDof_Node;

  assert(0 <= iNode);
  assert(iNode < System_nNodes);
  assert(0 <= iDof_Node);
  assert(iDof_Node < Node_nDofs);

  SV0[iDof] += value;
}

void AutoLinAS_AddElementVectorIntoRightHandSide
(int iElement, double ev[Element_nDofs])
{
  int iNode_Elem;
  int iDof_Node;

  assert(0 <= iElement);
  assert(iElement < System_nElements);

  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = (*System_elementNodeId) (iElement, iNode_Elem);
    
    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      int iDof = iNode * Node_nDofs + iDof_Node;
      int iDof_Elem = iNode_Elem * Node_nDofs + iDof_Node;
      
      SV0[iDof] += ev[iDof_Elem];
    }
  }
}

void AutoLinAS_ClearInitialResult (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVInitial[iDof] = 0.0;
  }
}

void AutoLinAS_CopyResultToInitialResult (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVInitial[iDof] = SV1[iDof];
  }
}

void AutoLinAS_SetInitialResult
(int iNode, int iDof_Node, double value)
{
  int iDof = iNode * Node_nDofs + iDof_Node;

  assert(0 <= iNode);
  assert(iNode < System_nNodes);
  assert(0 <= iDof_Node);
  assert(iDof_Node < Node_nDofs);

  SVInitial[iDof] += value;
}

double AutoLinAS_Result (int iNode, int iDof_Node)
{
  int iDof = iNode * Node_nDofs + iDof_Node;

  assert(0 <= iNode);
  assert(iNode < System_nNodes);
  assert(0 <= iDof_Node);
  assert(iDof_Node < Node_nDofs);

  return SV1[iDof];
}

void AutoLinAS_SetElementMatrixCallback 
(ElementMatrixCallbackType callback)
{
  assert(callback != NULL);

  System_elementMatrix = callback;
}



