/* post_es_2d_tri.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>

#include "../lib/autolinas_predefined.h"



int Domain_nDomains;



/* Node */

/* 2D plane strain, linear elastic */

int Node_nNodes[Domain_maxDomains];

/* global coordinate */
double Node_XVec[Domain_maxDomains][Node_maxNodes][Space_nDimensions];

/* displacement analysis result */
double Node_uVec[Domain_maxDomains][Node_maxNodes][Node_nDofs];

/* stress analysis result */
double Node_sigma[Domain_maxDomains][Node_maxNodes];



/* Element */

/* 2D plane strain, linear elastic */

int Element_nElements[Domain_maxDomains];

/* connectivity */
int Element_nodeIds[Domain_maxDomains][Element_maxElements][Element_nNodes];



void ReadMesh (FILE* fp, int iDomain)
{
  int iElement;
  int iNode_Elem;
  int iNode;
  
  fscanf (fp, "%d", &Element_nElements[iDomain]);
  assert(Element_nElements[iDomain] < Element_maxElements);
  fprintf (stderr, " nElements %d \n", Element_nElements[iDomain]);

  for (iElement = 0; 
       iElement < Element_nElements[iDomain]; 
       iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d ", 
	      &Element_nodeIds[iDomain][iElement][iNode_Elem]);
    }
  }

  fscanf (fp, "%d", &Node_nNodes[iDomain]);
  assert(Node_nNodes[iDomain] < Node_maxNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes[iDomain]);

  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    fscanf (fp, "%lf %lf", 
	    &Node_XVec[iDomain][iNode][0],
	    &Node_XVec[iDomain][iNode][1]);
  }
}

void ReadResult (FILE* fp, int iDomain)
{
  int iNode;

  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    fscanf (fp, "%lf %lf", 
	    &Node_uVec[iDomain][iNode][0], 
	    &Node_uVec[iDomain][iNode][1]);
  }
}

void ReadData (int argc, char* argv[])
{
  int iDomain;

  if (argc < 4) {
    fprintf (stderr, 
	     " post input_file output_file n_domains \n");
    exit (1);
  }

  Domain_nDomains = atoi (argv[3]);
  assert(Domain_nDomains <= Domain_maxDomains);
  
  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    char fileName[256];
    FILE* fp;

    sprintf (fileName, "%s.%d", argv[1], iDomain);
    fp = fopen (fileName, "r");
    if (fp == NULL) {
      fprintf (stderr, " cannot open input file  \n");
      exit (1);
    }
    ReadMesh (fp, iDomain);
    fclose (fp);
  }

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    char fileName[256];
    FILE* fp;

    sprintf (fileName, "%s.%d", argv[2], iDomain);
    fp = fopen (fileName, "r");
    if (fp == NULL) {
      fprintf (stderr, " cannot open output file  \n");
      exit (1);
    }
    ReadResult (fp, iDomain);
    fclose (fp);
  }
}



double MinRange = 0.0;
double MaxRange = 1000.0;

double Scale = 3.0;

int MeshIsShown = 1;

#define PLOT_NONE 0
#define PLOT_CONTOUR 1
#define PLOT_LINE_CONTOUR 2
#define PLOT_DEFORMATION 3

int PlotType = PLOT_NONE;



void PlotMesh (int iDomain) 
{
  int iElement;

  AutoGL_SetColor (1, 1, 1);

  for (iElement = 0; 
       iElement < Element_nElements[iDomain]; 
       iElement++) {
    int nodeId0 = Element_nodeIds[iDomain][iElement][0];
    int nodeId1 = Element_nodeIds[iDomain][iElement][1];
    int nodeId2 = Element_nodeIds[iDomain][iElement][2];
    AutoGL_DrawLine (Node_XVec[iDomain][nodeId0][0], 
		     Node_XVec[iDomain][nodeId0][1], 
		     0,
		     Node_XVec[iDomain][nodeId1][0], 
		     Node_XVec[iDomain][nodeId1][1], 
		     0);
    AutoGL_DrawLine (Node_XVec[iDomain][nodeId1][0], 
		     Node_XVec[iDomain][nodeId1][1], 
		     0,
		     Node_XVec[iDomain][nodeId2][0], 
		     Node_XVec[iDomain][nodeId2][1], 
		     0);
    AutoGL_DrawLine (Node_XVec[iDomain][nodeId2][0], 
		     Node_XVec[iDomain][nodeId2][1], 
		     0,
		     Node_XVec[iDomain][nodeId0][0], 
		     Node_XVec[iDomain][nodeId0][1], 
		     0);
  }
}

void PlotContour (int iDomain) 
{
  int iElement;

  AutoGL_DrawContourMap (80, MinRange, MaxRange);

  AutoGL_SetColor (1, 1, 1);

  for (iElement = 0; 
       iElement < Element_nElements[iDomain]; 
       iElement++) {
    int nodeId0 = Element_nodeIds[iDomain][iElement][0];
    int nodeId1 = Element_nodeIds[iDomain][iElement][1];
    int nodeId2 = Element_nodeIds[iDomain][iElement][2];
    double grades[3];

    grades[0] = (Node_sigma[iDomain][nodeId0] - MinRange) 
      / (MaxRange - MinRange);
    grades[1] = (Node_sigma[iDomain][nodeId1] - MinRange) 
      / (MaxRange - MinRange);
    grades[2] = (Node_sigma[iDomain][nodeId2] - MinRange) 
      / (MaxRange - MinRange);

    AutoGL_DrawContourTriangle 
      (Node_XVec[iDomain][nodeId0][0], 
       Node_XVec[iDomain][nodeId0][1], 
       0, 
       grades[0],
       Node_XVec[iDomain][nodeId1][0], 
       Node_XVec[iDomain][nodeId1][1], 
       0, 
       grades[1], 
       Node_XVec[iDomain][nodeId2][0], 
       Node_XVec[iDomain][nodeId2][1], 
       0, 
       grades[2]);
  }
}

void PlotLineContour (int iDomain) 
{
  int nGrades = 20;
  int iGrade;

  AutoGL_DrawLineContourMap (80, MinRange, MaxRange, nGrades);

  for (iGrade = 1; 
       iGrade < nGrades; 
       iGrade++) {
    double grade = 1.0 / nGrades * iGrade;
    double levelValue = MinRange + (MaxRange - MinRange) * grade;
    double red, green, blue;
    int iElement;

    AutoGL_GetContourColor (&red, &green, &blue,
			    grade);
    AutoGL_SetColor (red, green, blue);
  
    for (iElement = 0; 
	 iElement < Element_nElements[iDomain]; 
	 iElement++) {
      int nodeId0 = Element_nodeIds[iDomain][iElement][0];
      int nodeId1 = Element_nodeIds[iDomain][iElement][1];
      int nodeId2 = Element_nodeIds[iDomain][iElement][2];
      double triPoints[3][3];
      double triValues[3];

      triPoints[0][0] = Node_XVec[iDomain][nodeId0][0];
      triPoints[0][1] = Node_XVec[iDomain][nodeId0][1];
      triPoints[0][2] = 0.0;
      triPoints[1][0] = Node_XVec[iDomain][nodeId1][0];
      triPoints[1][1] = Node_XVec[iDomain][nodeId1][1];
      triPoints[1][2] = 0.0;
      triPoints[2][0] = Node_XVec[iDomain][nodeId2][0];
      triPoints[2][1] = Node_XVec[iDomain][nodeId2][1];
      triPoints[2][2] = 0.0;

      triValues[0] = Node_sigma[iDomain][nodeId0];
      triValues[1] = Node_sigma[iDomain][nodeId1];
      triValues[2] = Node_sigma[iDomain][nodeId2];
      
      AutoGL_DrawTriangleIsoline
	(levelValue, triPoints, triValues);
    }
  }
}

void PlotDeformation (int iDomain) 
{
  int iElement;

  AutoGL_SetColor (0, 1, 1);

  for (iElement = 0; 
       iElement < Element_nElements[iDomain]; 
       iElement++) {
    int nodeId0 = Element_nodeIds[iDomain][iElement][0];
    int nodeId1 = Element_nodeIds[iDomain][iElement][1];
    int nodeId2 = Element_nodeIds[iDomain][iElement][2];
    double p0X = Node_XVec[iDomain][nodeId0][0] 
      + Node_uVec[iDomain][nodeId0][0] * Scale;
    double p1X = Node_XVec[iDomain][nodeId1][0] 
      + Node_uVec[iDomain][nodeId1][0] * Scale;
    double p2X = Node_XVec[iDomain][nodeId2][0] 
      + Node_uVec[iDomain][nodeId2][0] * Scale;
    double p0Y = Node_XVec[iDomain][nodeId0][1] 
      + Node_uVec[iDomain][nodeId0][1] * Scale;
    double p1Y = Node_XVec[iDomain][nodeId1][1] 
      + Node_uVec[iDomain][nodeId1][1] * Scale;
    double p2Y = Node_XVec[iDomain][nodeId2][1] 
      + Node_uVec[iDomain][nodeId2][1] * Scale;

    AutoGL_DrawLine (p0X, p0Y, 0,
		     p1X, p1Y, 0);
    AutoGL_DrawLine (p1X, p1Y, 0,
		     p2X, p2Y, 0);
    AutoGL_DrawLine (p2X, p2Y, 0,
		     p0X, p0Y, 0);
  }
}

void RedrawView (void) 
{
  int iDomain;

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    if (MeshIsShown) {
      PlotMesh (iDomain);
    }
    
    switch (PlotType) {
    case PLOT_NONE:
      break;
    case PLOT_CONTOUR:
      PlotContour (iDomain);
      break;
    case PLOT_LINE_CONTOUR:
      PlotLineContour (iDomain);
      break;
    case PLOT_DEFORMATION:
      PlotDeformation (iDomain);
      break;
    }
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();  
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  ReadData (argc, argv);

  AutoGL_SetViewSize (30); 
  AutoGL_SetViewCenter (15, 10, 0); 

  AutoGL_SetViewRedrawCallback (RedrawView);   
  AutoGL_SetDefaultCallbackInMode2D (NULL);

  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode2D ();

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");
  AutoGL_AddBoolean (&MeshIsShown, "MeshIsShown");
  AutoGL_AddInteger (&PlotType, "PlotType");
  AutoGL_AddIntegerItem ("none");      
  AutoGL_AddIntegerItem ("contour"); 
  AutoGL_AddIntegerItem ("line_contour"); 
  AutoGL_AddIntegerItem ("deformation");    
  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");
  AutoGL_AddReal (&Scale, "Scale");
  AutoGL_AddCallback (UpdateModel, "UpdateModel");
}






