/* post_es_3d_tetra.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>

#include "../lib/autolinas_predefined.h"



int Domain_nDomains;



/* Node */

/* 3D solid, linear elastic */

int Node_nNodes[Domain_maxDomains];

/* global coordinate */
double Node_XVec[Domain_maxDomains][Node_maxNodes][Space_nDimensions];

/* displacement boundary condition */
int Node_u_BCVecFlag[Domain_maxDomains][Node_maxNodes][Node_nDofs];
double Node_u_BCVec[Domain_maxDomains][Node_maxNodes][Node_nDofs];

/* concentrated load boundary condition */
int Node_f_BCVecFlag[Domain_maxDomains][Node_maxNodes][Node_nDofs];
double Node_f_BCVec[Domain_maxDomains][Node_maxNodes][Node_nDofs];

/* displacement analysis result */
double Node_uVec[Domain_maxDomains][Node_maxNodes][Node_nDofs];

/* equivalent stress analysis result */
double Node_sigma_eq[Domain_maxDomains][Node_maxNodes];



/* Element */

/* 3D solid, linear elastic */

int Element_nElements[Domain_maxDomains];

/* connectivity */
int Element_nodeIds[Domain_maxDomains][Element_maxElements][Element_nNodes];



void ReadMesh (FILE* fp, int iDomain)
{
  int iElement;
  int iNode_Elem;
  int iNode;
  
  fscanf (fp, "%d", &Element_nElements[iDomain]);
  assert(Element_nElements[iDomain] < Element_maxElements);
  fprintf (stderr, " nElements %d \n", Element_nElements[iDomain]);

  for (iElement = 0; 
       iElement < Element_nElements[iDomain]; 
       iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d ", 
	      &Element_nodeIds[iDomain][iElement][iNode_Elem]);
    }
  }

  fscanf (fp, "%d", &Node_nNodes[iDomain]);
  assert(Node_nNodes[iDomain] < Node_maxNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes[iDomain]);

  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    fscanf (fp, "%lf %lf %lf", 
	    &Node_XVec[iDomain][iNode][0],
	    &Node_XVec[iDomain][iNode][1], 
	    &Node_XVec[iDomain][iNode][2]);
  }
}

void ReadCondition (FILE* fp, int iDomain)
{
  int iNode;
  int iDof_Node;
  int count;
  int iBc;
  
  for (iNode = 0; iNode < Node_nNodes[iDomain]; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iDomain][iNode][iDof_Node] = 0;
      Node_u_BCVec[iDomain][iNode][iDof_Node] = 0.0;
      Node_f_BCVecFlag[iDomain][iNode][iDof_Node] = 0;
      Node_f_BCVec[iDomain][iNode][iDof_Node] = 0.0;
    }
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " displacement BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_u_BCVecFlag[iDomain][iNode][iDof_Node] = 1;
    Node_u_BCVec[iDomain][iNode][iDof_Node] = value;
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " load BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_f_BCVecFlag[iDomain][iNode][iDof_Node] = 1;
    Node_f_BCVec[iDomain][iNode][iDof_Node] = value;
  }
}

void ReadResult (FILE* fp, int iDomain)
{
  int iNode;

  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    fscanf (fp, "%lf %lf %lf", 
	    &Node_uVec[iDomain][iNode][0], 
	    &Node_uVec[iDomain][iNode][1],
	    &Node_uVec[iDomain][iNode][2]);
  }

  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    fscanf (fp, "%lf", 
	    &Node_sigma_eq[iDomain][iNode]);
  }
}

void ReadData (int argc, char* argv[])
{
  int iDomain;

  if (argc < 4) {
    fprintf (stderr, 
	     " post_es_3d_tetra input_file output_file n_domains \n");
    exit (1);
  }
  
  Domain_nDomains = atoi (argv[3]);
  assert(Domain_nDomains <= Domain_maxDomains);

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    char fileName[256];
    FILE* fp;

    sprintf (fileName, "%s.%d", argv[1], iDomain);
    fp = fopen (fileName, "r");
    if (fp == NULL) {
      fprintf (stderr, " cannot open input file  \n");
      exit (1);
    }
    ReadMesh (fp, iDomain);
    ReadCondition (fp, iDomain);
    fclose (fp);
  }

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    char fileName[256];
    FILE* fp;

    sprintf (fileName, "%s.%d", argv[2], iDomain);
    fp = fopen (fileName, "r");
    if (fp == NULL) {
      fprintf (stderr, " cannot open output file  \n");
      exit (1);
    }
    ReadResult (fp, iDomain);
    fclose (fp);
  }
}



double MinRange = 0.0;
double MaxRange = 1.0;

double Scale = 2.0;

int MeshIsShown = 1;
int DeformationIsShown;
int ContourIsShown;
int DisplacementBcIsShown;
int LoadBcIsShown;



static void SetRange ()
{
  int iDomain;
  int iNode;
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double size;
  double midX, midY, midZ;
  double maxLength;

  AutoGL_AssignVector3D 
    (&minX, &minY, &minZ,
     Node_XVec[0][0][0], Node_XVec[0][0][1], Node_XVec[0][0][2]);
  AutoGL_AssignVector3D 
    (&maxX, &maxY, &maxZ,
     Node_XVec[0][0][0], Node_XVec[0][0][1], Node_XVec[0][0][2]);
  
  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    for (iNode = 0; iNode < Node_nNodes[iDomain]; iNode++) {
      double x = Node_XVec[iDomain][iNode][0];
      double y = Node_XVec[iDomain][iNode][1];
      double z = Node_XVec[iDomain][iNode][2];
      
      AutoGL_GetMinVector3D (&minX, &minY, &minZ,
			     minX, minY, minZ,
			     x, y, z);
      AutoGL_GetMaxVector3D (&maxX, &maxY, &maxZ,
			     maxX, maxY, maxZ,
			     x, y, z);
    }
  }

  size = AutoGL_GetVectorDistance3D 
    (minX, minY, minZ,
     maxX, maxY, maxZ);
  AutoGL_AssignVector3D 
    (&midX, &midY, &midZ,
     (minX + maxX) / 2, 
     (minY + maxY) / 2,
     (minZ + maxZ) / 2);

  AutoGL_SetViewSize (size / 2); 
  AutoGL_SetViewCenter (midX, midY, midZ); 

  MinRange = Node_sigma_eq[0][0]; 
  MaxRange = Node_sigma_eq[0][0]; 
  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    for (iNode = 0; iNode < Node_nNodes[iDomain]; iNode++) {
      if (Node_sigma_eq[iDomain][iNode] < MinRange) {
	MinRange = Node_sigma_eq[iDomain][iNode];
      }
      if (MaxRange < Node_sigma_eq[iDomain][iNode]) {
	MaxRange = Node_sigma_eq[iDomain][iNode];
      }
    }
  }

  maxLength = 0.0;
  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    for (iNode = 0; iNode < Node_nNodes[iDomain]; iNode++) {
      double ux = Node_uVec[iDomain][iNode][0];
      double uy = Node_uVec[iDomain][iNode][1];
      double uz = Node_uVec[iDomain][iNode][2]; 
      double length = AutoGL_GetVectorLength3D (ux, uy, uz);
      if (maxLength < length) {
	maxLength = length;
      }
    }
  }
    
  Scale = size / maxLength * 0.1;
}

void PlotModel (int iDomain)
{
  int iElement;

  for (iElement = 0; iElement < Element_nElements[iDomain]; iElement++) {
    int nodeId0 = Element_nodeIds[iDomain][iElement][0];
    int nodeId1 = Element_nodeIds[iDomain][iElement][1];
    int nodeId2 = Element_nodeIds[iDomain][iElement][2];
    int nodeId3 = Element_nodeIds[iDomain][iElement][3];

    double p0X = Node_XVec[iDomain][nodeId0][0];
    double p0Y = Node_XVec[iDomain][nodeId0][1];
    double p0Z = Node_XVec[iDomain][nodeId0][2];

    double p1X = Node_XVec[iDomain][nodeId1][0];
    double p1Y = Node_XVec[iDomain][nodeId1][1];
    double p1Z = Node_XVec[iDomain][nodeId1][2];

    double p2X = Node_XVec[iDomain][nodeId2][0];
    double p2Y = Node_XVec[iDomain][nodeId2][1];
    double p2Z = Node_XVec[iDomain][nodeId2][2];

    double p3X = Node_XVec[iDomain][nodeId3][0];
    double p3Y = Node_XVec[iDomain][nodeId3][1];
    double p3Z = Node_XVec[iDomain][nodeId3][2];

    if (DeformationIsShown) {
      p0X += Node_uVec[iDomain][nodeId0][0] * Scale;
      p0Y += Node_uVec[iDomain][nodeId0][1] * Scale;
      p0Z += Node_uVec[iDomain][nodeId0][2] * Scale;
      
      p1X += Node_uVec[iDomain][nodeId1][0] * Scale;
      p1Y += Node_uVec[iDomain][nodeId1][1] * Scale;
      p1Z += Node_uVec[iDomain][nodeId1][2] * Scale;
      
      p2X += Node_uVec[iDomain][nodeId2][0] * Scale;
      p2Y += Node_uVec[iDomain][nodeId2][1] * Scale;
      p2Z += Node_uVec[iDomain][nodeId2][2] * Scale;
      
      p3X += Node_uVec[iDomain][nodeId3][0] * Scale;
      p3Y += Node_uVec[iDomain][nodeId3][1] * Scale;
      p3Z += Node_uVec[iDomain][nodeId3][2] * Scale;
    }

    if (MeshIsShown) {
      AutoGL_SetColor (1, 0, 0);
      AutoGL_DrawLine (p0X, p0Y, p0Z,
		       p1X, p1Y, p1Z);
      AutoGL_DrawLine (p1X, p1Y, p1Z,
		       p2X, p2Y, p2Z);
      AutoGL_DrawLine (p2X, p2Y, p2Z,
		       p0X, p0Y, p0Z);
      AutoGL_DrawLine (p0X, p0Y, p0Z,
		       p3X, p3Y, p3Z);
      AutoGL_DrawLine (p1X, p1Y, p1Z,
		       p3X, p3Y, p3Z);
      AutoGL_DrawLine (p2X, p2Y, p2Z,
		       p3X, p3Y, p3Z);
    }

    if (ContourIsShown) {
      double value0 = (Node_sigma_eq[iDomain][nodeId0] - MinRange)
	/ (MaxRange - MinRange);
      double value1 = (Node_sigma_eq[iDomain][nodeId1] - MinRange)
	/ (MaxRange - MinRange);
      double value2 = (Node_sigma_eq[iDomain][nodeId2] - MinRange)
	/ (MaxRange - MinRange);
      double value3 = (Node_sigma_eq[iDomain][nodeId3] - MinRange)
	/ (MaxRange - MinRange);
      
      AutoGL_DrawContourTriangle (p0X, p0Y, p0Z, value0, 
				  p1X, p1Y, p1Z, value1, 
				  p2X, p2Y, p2Z, value2);
      AutoGL_DrawContourTriangle (p0X, p0Y, p0Z, value0,
				  p1X, p1Y, p1Z, value1,
				  p3X, p3Y, p3Z, value3);
      AutoGL_DrawContourTriangle (p1X, p1Y, p1Z, value1,
				  p2X, p2Y, p2Z, value2,
				  p3X, p3Y, p3Z, value3);
      AutoGL_DrawContourTriangle (p2X, p2Y, p2Z, value2,
				  p0X, p0Y, p0Z, value0,
				  p3X, p3Y, p3Z, value3);
    } else {
      AutoGL_SetColor (0, 1, 1);
      AutoGL_DrawTriangle (p0X, p0Y, p0Z,
			   p1X, p1Y, p1Z,
			   p2X, p2Y, p2Z);
      AutoGL_DrawTriangle (p0X, p0Y, p0Z,
			   p1X, p1Y, p1Z,
			   p3X, p3Y, p3Z);
      AutoGL_DrawTriangle (p1X, p1Y, p1Z,
			   p2X, p2Y, p2Z,
			   p3X, p3Y, p3Z);
      AutoGL_DrawTriangle (p2X, p2Y, p2Z,
			   p0X, p0Y, p0Z,
			   p3X, p3Y, p3Z);
    }
  }
}

void PlotDisplacementBc (int iDomain) 
{
  int iNode;

  AutoGL_SetColor (0, 1, 0);
  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    double x = Node_XVec[iDomain][iNode][0];
    double y = Node_XVec[iDomain][iNode][1];
    double z = Node_XVec[iDomain][iNode][2];

    if (Node_u_BCVecFlag[iDomain][iNode][0]
	&& Node_u_BCVecFlag[iDomain][iNode][1]
	&& Node_u_BCVecFlag[iDomain][iNode][2]) {
      
      AutoGL_DrawCubeMark3D (x, y, z, 5, 0);
    } else {
      if (Node_u_BCVecFlag[iDomain][iNode][0]) {
	double arrowLength = Node_u_BCVec[iDomain][iNode][0] * Scale;
	
	AutoGL_DrawLine (x, y, z, 
			 x + arrowLength, y, z);
      } 
      if (Node_u_BCVecFlag[iDomain][iNode][1]) {
	double arrowLength = Node_u_BCVec[iDomain][iNode][1] * Scale;
	
	AutoGL_DrawLine (x, y, z, 
			 x, y + arrowLength, z);
      } 
      if (Node_u_BCVecFlag[iDomain][iNode][2]) {
	double arrowLength = Node_u_BCVec[iDomain][iNode][2] * Scale;
	
	AutoGL_DrawLine (x, y, z, 
			 x, y, z + arrowLength);
      }
    }
  }
}

void PlotLoadBc (int iDomain) 
{
  int iNode;

  AutoGL_SetColor (0, 0, 1);
  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    double x = Node_XVec[iDomain][iNode][0];
    double y = Node_XVec[iDomain][iNode][1];
    double z = Node_XVec[iDomain][iNode][2];

    if (Node_f_BCVecFlag[iDomain][iNode][0]) {
      double arrowLength = Node_f_BCVec[iDomain][iNode][0] * Scale;
      
      AutoGL_DrawLine (x, y, z, 
		       x + arrowLength, y, z);
    }
    if (Node_f_BCVecFlag[iDomain][iNode][1]) {
      double arrowLength = Node_f_BCVec[iDomain][iNode][1] * Scale;
      
      AutoGL_DrawLine (x, y, z, 
		       x, y + arrowLength, z);
    }
    if (Node_f_BCVecFlag[iDomain][iNode][2]) {
      double arrowLength = Node_f_BCVec[iDomain][iNode][2] * Scale;
      
      AutoGL_DrawLine (x, y, z, 
		       x, y, z + arrowLength);
    }
  }
}

void RedrawView (void) 
{
  int iDomain;

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    PlotModel (iDomain);
    if (DisplacementBcIsShown) {
      PlotDisplacementBc (iDomain);
    }
    if (LoadBcIsShown) {
      PlotLoadBc (iDomain);
    }
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();  
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  ReadData (argc, argv);
  SetRange ();

  AutoGL_SetViewRedrawCallback (RedrawView);   
  AutoGL_SetDefaultCallbackInMode3D (NULL);

  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();  

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");
  AutoGL_AddBoolean (&MeshIsShown, "MeshIsShown");
  AutoGL_SetLabel ("mesh");
  AutoGL_AddBoolean (&DeformationIsShown, "DeformationIsShown");
  AutoGL_SetLabel ("deformation");
  AutoGL_AddBoolean (&ContourIsShown, "ContourIsShown");
  AutoGL_SetLabel ("contour");
  AutoGL_AddBoolean (&DisplacementBcIsShown, "DisplacementBcIsShown");
  AutoGL_SetLabel ("disp. bc");
  AutoGL_AddBoolean (&LoadBcIsShown, "LoadBcIsShown");
  AutoGL_SetLabel ("load bc");
  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");
  AutoGL_AddReal (&Scale, "Scale");
  AutoGL_AddCallback (UpdateModel, "UpdateModel");
}






