/* postfgr_es_3d_tetra.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>

#include "../lib/autolinas_predefined.h"



#define Face_maxFaces 100000

int Domain_nDomains;



/* Node */

/* 3D solid, linear elastic */

int Node_nNodes[Domain_maxDomains];

/* global coordinate */
double Node_XVec[Domain_maxDomains][Node_maxNodes][Space_nDimensions];

/* displacement boundary condition */
int Node_u_BCVecFlag[Domain_maxDomains][Node_maxNodes][Node_nDofs];
double Node_u_BCVec[Domain_maxDomains][Node_maxNodes][Node_nDofs];

/* concentrated load boundary condition */
int Node_f_BCVecFlag[Domain_maxDomains][Node_maxNodes][Node_nDofs];
double Node_f_BCVec[Domain_maxDomains][Node_maxNodes][Node_nDofs];

/* displacement analysis result */
double Node_uVec[Domain_maxDomains][Node_maxNodes][Node_nDofs];

/* equivalent stress analysis result */
double Node_sigma_eq[Domain_maxDomains][Node_maxNodes];



/* Element */

/* 3D solid, linear elastic */

int Element_nElements[Domain_maxDomains];

/* connectivity */
int Element_nodeIds[Domain_maxDomains][Element_maxElements][Element_nNodes];



/* Edge */

#define Edge_maxEdges 100000

int Edge_nEdges;
double Edge_XVec0[Edge_maxEdges][3];
double Edge_XVec1[Edge_maxEdges][3];



/* FaceGroup */

int Face_nFaces[Domain_maxDomains];
int Face_elementId[Domain_maxDomains][Face_maxFaces];
int Face_faceId[Domain_maxDomains][Face_maxFaces];



void ReadEdge (FILE* fp)
{
  int iEdge;
  
  fscanf (fp, "%d", 
	  &Edge_nEdges);
  assert(Edge_nEdges <= Edge_maxEdges);
  fprintf (stderr, " nEdges %d \n", Edge_nEdges);

  for (iEdge = 0; iEdge < Edge_nEdges; iEdge++) {
    fscanf (fp, "%lf %lf %lf %lf %lf %lf", 
	    &Edge_XVec0[iEdge][0],
	    &Edge_XVec0[iEdge][1], 
	    &Edge_XVec0[iEdge][2],
	    &Edge_XVec1[iEdge][0],
	    &Edge_XVec1[iEdge][1], 
	    &Edge_XVec1[iEdge][2]);
  }
}

void ReadMesh (FILE* fp, int iDomain)
{
  int iElement;
  int iNode_Elem;
  int iNode;
  
  fscanf (fp, "%d", &Element_nElements[iDomain]);
  assert(Element_nElements[iDomain] < Element_maxElements);
  fprintf (stderr, " nElements %d \n", Element_nElements[iDomain]);

  for (iElement = 0; 
       iElement < Element_nElements[iDomain]; 
       iElement++) {

    for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
      fscanf (fp, "%d ", 
	      &Element_nodeIds[iDomain][iElement][iNode_Elem]);
    }
  }

  fscanf (fp, "%d", &Node_nNodes[iDomain]);
  assert(Node_nNodes[iDomain] < Node_maxNodes);
  fprintf (stderr, " nNodes %d \n", Node_nNodes[iDomain]);

  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    fscanf (fp, "%lf %lf %lf", 
	    &Node_XVec[iDomain][iNode][0],
	    &Node_XVec[iDomain][iNode][1], 
	    &Node_XVec[iDomain][iNode][2]);
  }
}

void ReadCondition (FILE* fp, int iDomain)
{
  int iNode;
  int iDof_Node;
  int count;
  int iBc;
  
  for (iNode = 0; iNode < Node_nNodes[iDomain]; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      Node_u_BCVecFlag[iDomain][iNode][iDof_Node] = 0;
      Node_u_BCVec[iDomain][iNode][iDof_Node] = 0.0;
      Node_f_BCVecFlag[iDomain][iNode][iDof_Node] = 0;
      Node_f_BCVec[iDomain][iNode][iDof_Node] = 0.0;
    }
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " displacement BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_u_BCVecFlag[iDomain][iNode][iDof_Node] = 1;
    Node_u_BCVec[iDomain][iNode][iDof_Node] = value;
  }

  fscanf (fp, "%d", &count);
  fprintf (stderr, " load BCs %d \n", count);

  for (iBc = 0; iBc < count; iBc++) {
    int iNode;
    int iDof_Node;
    double value;

    fscanf (fp, "%d %d %lf", 
	    &iNode, &iDof_Node, &value);
    Node_f_BCVecFlag[iDomain][iNode][iDof_Node] = 1;
    Node_f_BCVec[iDomain][iNode][iDof_Node] = value;
  }
}

void ReadResult (FILE* fp, int iDomain)
{
  int iNode;

  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    fscanf (fp, "%lf %lf %lf", 
	    &Node_uVec[iDomain][iNode][0], 
	    &Node_uVec[iDomain][iNode][1],
	    &Node_uVec[iDomain][iNode][2]);
  }

  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    fscanf (fp, "%lf", 
	    &Node_sigma_eq[iDomain][iNode]);
  }
}

void ReadFgr (FILE* fp, int iDomain)
{
  int iFace;
  
  fscanf (fp, "%d", &Face_nFaces[iDomain]);
  assert(Face_nFaces[iDomain] <= Face_maxFaces);

  for (iFace = 0; iFace < Face_nFaces[iDomain]; iFace++) {
    fscanf (fp, "%d %d", 
	    &Face_elementId[iDomain][iFace], 
	    &Face_faceId[iDomain][iFace]);
  }
}

void ReadData (int argc, char* argv[])
{
  FILE *fp;
  int iDomain;

  if (argc < 6) {
    fprintf (stderr, 
	     " post_es_3d_tetra global_file edge_file "
	     " input_file output_file fgr_file \n");
    exit (1);
  }
  
  fp = fopen (argv[1], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open global file  \n");
    exit (1);
  }
  {
    int dummy;
    
    fscanf (fp, "%d %d", 
	    &dummy, &Domain_nDomains);
    assert(Domain_nDomains <= Domain_maxDomains);
  }
  fclose (fp);

  fp = fopen (argv[2], "r");
  if (fp == NULL) {
    fprintf (stderr, " cannot open edge file  \n");
    exit (1);
  }
  ReadEdge (fp);
  fclose (fp);

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    char fileName[256];
    FILE* fp;

    sprintf (fileName, "%s.%d", argv[3], iDomain);
    fp = fopen (fileName, "r");
    if (fp == NULL) {
      fprintf (stderr, " cannot open input file  \n");
      exit (1);
    }
    ReadMesh (fp, iDomain);
    ReadCondition (fp, iDomain);
    fclose (fp);
  }

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    char fileName[256];
    FILE* fp;

    sprintf (fileName, "%s.%d", argv[4], iDomain);
    fp = fopen (fileName, "r");
    if (fp == NULL) {
      fprintf (stderr, " cannot open output file  \n");
      exit (1);
    }
    ReadResult (fp, iDomain);
    fclose (fp);
  }

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    char fileName[256];
    FILE* fp;

    sprintf (fileName, "%s.%d", argv[5], iDomain);
    fp = fopen (fileName, "r");
    if (fp == NULL) {
      fprintf (stderr, " cannot open fgr file  \n");
      exit (1);
    }
    ReadFgr (fp, iDomain);
    fclose (fp);
  }
}



double MinRange = 0.0;
double MaxRange = 1.0;

double Scale = 2.0;

double A = 0.0;
double B = 0.0;
double C = 1.0;
double D = 0.0;
double SectionDelta = 0.001;

double LevelValue = 0.0;
double LevelWidth = 1.0;
double LevelDelta = 0.001;

int EdgeIsShown = 1;
int SurfaceIsShown;
int MeshIsShown;
int DeformationIsShown;
int SurfaceContourIsShown;
int SectionContourIsShown;
int IsosurfaceIsShown;
int DisplacementBcIsShown;
int LoadBcIsShown;

int FaceNodeMap[4][3] = {
  { 1, 2, 3 }, 
  { 0, 3, 2 }, 
  { 0, 1, 3 }, 
  { 0, 2, 1 }
};



static void SetRange ()
{
  int iDomain;
  int iNode;
  double minX, minY, minZ;
  double maxX, maxY, maxZ;
  double size;
  double midX, midY, midZ;
  double maxLength;

  AutoGL_AssignVector3D 
    (&minX, &minY, &minZ,
     Node_XVec[0][0][0], Node_XVec[0][0][1], Node_XVec[0][0][2]);
  AutoGL_AssignVector3D 
    (&maxX, &maxY, &maxZ,
     Node_XVec[0][0][0], Node_XVec[0][0][1], Node_XVec[0][0][2]);
  
  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    for (iNode = 0; iNode < Node_nNodes[iDomain]; iNode++) {
      double x = Node_XVec[iDomain][iNode][0];
      double y = Node_XVec[iDomain][iNode][1];
      double z = Node_XVec[iDomain][iNode][2];
      
      AutoGL_GetMinVector3D (&minX, &minY, &minZ,
			     minX, minY, minZ,
			     x, y, z);
      AutoGL_GetMaxVector3D (&maxX, &maxY, &maxZ,
			     maxX, maxY, maxZ,
			     x, y, z);
    }
  }

  size = AutoGL_GetVectorDistance3D 
    (minX, minY, minZ,
     maxX, maxY, maxZ);
  AutoGL_AssignVector3D 
    (&midX, &midY, &midZ,
     (minX + maxX) / 2, 
     (minY + maxY) / 2,
     (minZ + maxZ) / 2);

  AutoGL_SetViewSize (size / 2); 
  AutoGL_SetViewCenter (midX, midY, midZ); 

  SectionDelta = (maxX - minX + maxY - minY + maxZ - minZ) / 100;

  MinRange = Node_sigma_eq[0][0]; 
  MaxRange = Node_sigma_eq[0][0]; 
  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    for (iNode = 0; iNode < Node_nNodes[iDomain]; iNode++) {
      if (Node_sigma_eq[iDomain][iNode] < MinRange) {
	MinRange = Node_sigma_eq[iDomain][iNode];
      }
      if (MaxRange < Node_sigma_eq[iDomain][iNode]) {
	MaxRange = Node_sigma_eq[iDomain][iNode];
      }
    }
  }

  LevelValue = (MinRange + MaxRange) / 2;
  LevelWidth = (MaxRange - MinRange) / 4;
  LevelDelta = (MaxRange - MinRange) / 100;

  maxLength = 0.0;
  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    for (iNode = 0; iNode < Node_nNodes[iDomain]; iNode++) {
      double ux = Node_uVec[iDomain][iNode][0];
      double uy = Node_uVec[iDomain][iNode][1];
      double uz = Node_uVec[iDomain][iNode][2]; 
      double length = AutoGL_GetVectorLength3D (ux, uy, uz);
      if (maxLength < length) {
	maxLength = length;
      }
    }
  }
    
  Scale = size / maxLength * 0.1;
}

void PlotEdge (void) 
{
  int iEdge;

  AutoGL_SetColor (0, 0, 1);
  for (iEdge = 0; iEdge < Edge_nEdges; iEdge++) {
    AutoGL_DrawLine (Edge_XVec0[iEdge][0],
		     Edge_XVec0[iEdge][1], 
		     Edge_XVec0[iEdge][2],
		     Edge_XVec1[iEdge][0],
		     Edge_XVec1[iEdge][1], 
		     Edge_XVec1[iEdge][2]);
  }
}

void PlotSurface (int iDomain)
{
  int iFace;

  for (iFace = 0; 
       iFace < Face_nFaces[iDomain]; 
       iFace++) {
    int elementId = Face_elementId[iDomain][iFace];
    int faceId = Face_faceId[iDomain][iFace];
    int nodeId0 
      = Element_nodeIds[iDomain][elementId][FaceNodeMap[faceId][0]];
    int nodeId1 
      = Element_nodeIds[iDomain][elementId][FaceNodeMap[faceId][1]];
    int nodeId2 
      = Element_nodeIds[iDomain][elementId][FaceNodeMap[faceId][2]];

    double p0X = Node_XVec[iDomain][nodeId0][0];
    double p0Y = Node_XVec[iDomain][nodeId0][1];
    double p0Z = Node_XVec[iDomain][nodeId0][2];

    double p1X = Node_XVec[iDomain][nodeId1][0];
    double p1Y = Node_XVec[iDomain][nodeId1][1];
    double p1Z = Node_XVec[iDomain][nodeId1][2];

    double p2X = Node_XVec[iDomain][nodeId2][0];
    double p2Y = Node_XVec[iDomain][nodeId2][1];
    double p2Z = Node_XVec[iDomain][nodeId2][2];

    if (DeformationIsShown) {
      p0X += Node_uVec[iDomain][nodeId0][0] * Scale;
      p0Y += Node_uVec[iDomain][nodeId0][1] * Scale;
      p0Z += Node_uVec[iDomain][nodeId0][2] * Scale;
      
      p1X += Node_uVec[iDomain][nodeId1][0] * Scale;
      p1Y += Node_uVec[iDomain][nodeId1][1] * Scale;
      p1Z += Node_uVec[iDomain][nodeId1][2] * Scale;
      
      p2X += Node_uVec[iDomain][nodeId2][0] * Scale;
      p2Y += Node_uVec[iDomain][nodeId2][1] * Scale;
      p2Z += Node_uVec[iDomain][nodeId2][2] * Scale;
    }

    if (MeshIsShown) {
      AutoGL_SetColor (1, 0, 0);
      AutoGL_DrawLine (p0X, p0Y, p0Z,
		       p1X, p1Y, p1Z);
      AutoGL_DrawLine (p1X, p1Y, p1Z,
		       p2X, p2Y, p2Z);
      AutoGL_DrawLine (p2X, p2Y, p2Z,
		       p0X, p0Y, p0Z);
    }

    if (SurfaceContourIsShown) {
      double value0 = (Node_sigma_eq[iDomain][nodeId0] - MinRange)
	/ (MaxRange - MinRange);
      double value1 = (Node_sigma_eq[iDomain][nodeId1] - MinRange)
	/ (MaxRange - MinRange);
      double value2 = (Node_sigma_eq[iDomain][nodeId2] - MinRange)
	/ (MaxRange - MinRange);
      
      AutoGL_DrawContourTriangle (p0X, p0Y, p0Z, value0, 
				  p1X, p1Y, p1Z, value1, 
				  p2X, p2Y, p2Z, value2);
    } else {
      AutoGL_SetColor (0, 1, 1);
      AutoGL_DrawTriangle (p0X, p0Y, p0Z,
			   p1X, p1Y, p1Z,
			   p2X, p2Y, p2Z);
    }
  }
}

void PlotSection (int iDomain) 
{
  int iElement;
  
  for (iElement = 0; 
       iElement < Element_nElements[iDomain]; 
       iElement++) {
    int iVertex;
    double tetraPoints[4][3];
    double tetraValues[4];
    double tetraGrades[4];
    
    for (iVertex = 0; iVertex < 4; iVertex++) {
      int nodeId = Element_nodeIds[iDomain][iElement][iVertex];

      tetraPoints[iVertex][0] = Node_XVec[iDomain][nodeId][0];
      tetraPoints[iVertex][1] = Node_XVec[iDomain][nodeId][1];
      tetraPoints[iVertex][2] = Node_XVec[iDomain][nodeId][2];

      tetraValues[iVertex] = Node_sigma_eq[iDomain][nodeId];
      tetraGrades[iVertex] = (Node_sigma_eq[iDomain][nodeId] - MinRange) 
	/ (MaxRange - MinRange);
    }
    
    AutoGL_DrawTetrahedronSectionContour 
      (A, B, C, D, tetraPoints, tetraGrades);
  }
}

void PlotIsosurfaceAtLevel (double level, int iDomain) 
{
  double grade;
  double red, green, blue;
  int iElement;
  
  grade = (level - MinRange) / (MaxRange - MinRange);
  AutoGL_GetContourColor (&red, &green, &blue,
			  grade);
  AutoGL_SetColor (red, green, blue);
  
  for (iElement = 0; 
       iElement < Element_nElements[iDomain]; 
       iElement++) {
    int iVertex;
    double tetraPoints[4][3];
    double tetraValues[4];
    double tetraGrades[4];
    
    for (iVertex = 0; iVertex < 4; iVertex++) {
      int nodeId = Element_nodeIds[iDomain][iElement][iVertex];

      tetraPoints[iVertex][0] = Node_XVec[iDomain][nodeId][0];
      tetraPoints[iVertex][1] = Node_XVec[iDomain][nodeId][1];
      tetraPoints[iVertex][2] = Node_XVec[iDomain][nodeId][2];

      tetraValues[iVertex] = Node_sigma_eq[iDomain][nodeId];
      tetraGrades[iVertex] = (Node_sigma_eq[iDomain][nodeId] - MinRange) 
	/ (MaxRange - MinRange);
    }
    
    AutoGL_DrawTetrahedronIsosurface 
      (level, tetraPoints, tetraValues);
  }
}

void PlotIsosurface (int iDomain) 
{
  PlotIsosurfaceAtLevel (LevelValue - LevelWidth, iDomain);
  PlotIsosurfaceAtLevel (LevelValue, iDomain);
  PlotIsosurfaceAtLevel (LevelValue + LevelWidth, iDomain);
}

void PlotDisplacementBc (int iDomain) 
{
  int iNode;

  AutoGL_SetColor (1, 0, 0);
  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    double x = Node_XVec[iDomain][iNode][0];
    double y = Node_XVec[iDomain][iNode][1];
    double z = Node_XVec[iDomain][iNode][2];

    if (Node_u_BCVecFlag[iDomain][iNode][0]
	&& Node_u_BCVecFlag[iDomain][iNode][1]
	&& Node_u_BCVecFlag[iDomain][iNode][2]) {
      
      AutoGL_DrawCubeMark3D (x, y, z, 5, 0);
    } else {
      if (Node_u_BCVecFlag[iDomain][iNode][0]) {
	double arrowLength = Node_u_BCVec[iDomain][iNode][0] * Scale;
	
	AutoGL_DrawLine (x, y, z, 
			 x + arrowLength, y, z);
      } 
      if (Node_u_BCVecFlag[iDomain][iNode][1]) {
	double arrowLength = Node_u_BCVec[iDomain][iNode][1] * Scale;
	
	AutoGL_DrawLine (x, y, z, 
			 x, y + arrowLength, z);
      } 
      if (Node_u_BCVecFlag[iDomain][iNode][2]) {
	double arrowLength = Node_u_BCVec[iDomain][iNode][2] * Scale;
	
	AutoGL_DrawLine (x, y, z, 
			 x, y, z + arrowLength);
      }
    }
  }
}

void PlotLoadBc (int iDomain) 
{
  int iNode;

  AutoGL_SetColor (1, 0, 0);
  for (iNode = 0; 
       iNode < Node_nNodes[iDomain]; 
       iNode++) {
    double x = Node_XVec[iDomain][iNode][0];
    double y = Node_XVec[iDomain][iNode][1];
    double z = Node_XVec[iDomain][iNode][2];

    if (Node_f_BCVecFlag[iDomain][iNode][0]) {
      double arrowLength = Node_f_BCVec[iDomain][iNode][0] * Scale;
      
      AutoGL_DrawLine (x, y, z, 
		       x + arrowLength, y, z);
    }
    if (Node_f_BCVecFlag[iDomain][iNode][1]) {
      double arrowLength = Node_f_BCVec[iDomain][iNode][1] * Scale;
      
      AutoGL_DrawLine (x, y, z, 
		       x, y + arrowLength, z);
    }
    if (Node_f_BCVecFlag[iDomain][iNode][2]) {
      double arrowLength = Node_f_BCVec[iDomain][iNode][2] * Scale;
      
      AutoGL_DrawLine (x, y, z, 
		       x, y, z + arrowLength);
    }
  }
}

void RedrawView (void) 
{
  int iDomain;

  if (EdgeIsShown) {
    PlotEdge ();
  }

  for (iDomain = 0; iDomain < Domain_nDomains; iDomain++) {
    if (SurfaceIsShown) {
      PlotSurface (iDomain);
    }
    if (SectionContourIsShown) {
      PlotSection (iDomain);
    }
    if (IsosurfaceIsShown) {
      PlotIsosurface (iDomain);
    }
    if (DisplacementBcIsShown) {
      PlotDisplacementBc (iDomain);
    }
    if (LoadBcIsShown) {
      PlotLoadBc (iDomain);
    }
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();  
}

void UpSection (void) 
{
  D += SectionDelta;
  AutoGL_DrawView ();  
}

void DownSection (void) 
{
  D -= SectionDelta;
  AutoGL_DrawView ();  
}

void UpLevel (void) 
{
  LevelValue += LevelDelta;
  AutoGL_DrawView ();  
}

void DownLevel (void) 
{
  LevelValue -= LevelDelta;
  AutoGL_DrawView ();  
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  ReadData (argc, argv);
  SetRange ();

  AutoGL_SetViewRedrawCallback (RedrawView);   
  AutoGL_SetDefaultCallbackInMode3D (NULL);

  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();  

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");

  AutoGL_AddReal (&MinRange, "MinRange");
  AutoGL_AddReal (&MaxRange, "MaxRange");
  AutoGL_AddReal (&Scale, "Scale");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");

  AutoGL_AddBoolean (&EdgeIsShown, "EdgeIsShown");
  AutoGL_SetLabel ("edge");

  AutoGL_AddBoolean (&SurfaceIsShown, "SurfaceIsShown");
  AutoGL_SetLabel ("surface");
  AutoGL_AddBoolean (&MeshIsShown, "MeshIsShown");
  AutoGL_SetLabel ("mesh");
  AutoGL_AddBoolean (&DeformationIsShown, "DeformationIsShown");
  AutoGL_SetLabel ("deformation");
  AutoGL_AddBoolean (&SurfaceContourIsShown, "SurfaceContourIsShown");
  AutoGL_SetLabel ("surface contour");

  AutoGL_AddBoolean (&DisplacementBcIsShown, "DisplacementBcIsShown");
  AutoGL_SetLabel ("disp. bc");
  AutoGL_AddBoolean (&LoadBcIsShown, "LoadBcIsShown");
  AutoGL_SetLabel ("load bc");

  AutoGL_AddBoolean (&SectionContourIsShown, "SectionContourIsShown");
  AutoGL_SetLabel ("section contour");
  AutoGL_SetLabel ("equation Ax+By+Cz+D=0");
  AutoGL_AddReal (&A, "A");      
  AutoGL_AddReal (&B, "B");      
  AutoGL_AddReal (&C, "C");      
  AutoGL_AddReal (&D, "D");      
  AutoGL_AddReal (&SectionDelta, "SectionDelta");
  AutoGL_AddCallback (UpSection, "UpSection");
  AutoGL_AddCallback (DownSection, "DownSection");

  AutoGL_AddBoolean (&IsosurfaceIsShown, "IsosurfaceIsShown");
  AutoGL_SetLabel ("isosurface");
  AutoGL_AddReal (&LevelValue, "LevelValue");
  AutoGL_AddReal (&LevelWidth, "LevelWidth");
  AutoGL_AddReal (&LevelDelta, "LevelDelta");
  AutoGL_AddCallback (UpLevel, "UpLevel");
  AutoGL_AddCallback (DownLevel, "DownLevel");
}






