/* pre_2d_tri.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../lib/autolinas_predefined.h"



/* Node */

int Node_nNodes;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* global node ID */
int Node_globalId[Node_maxNodes];



/* Element */

int Element_nElements;

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



#define X_SIZE 1.0
#define Y_SIZE 1.0

int XDivision;
int YDivision;
int XDomains;
int YDomains;

void WriteGlobalFile (char* fileName)
{
  FILE* fp;
  int nGlobalNodes 
    = (XDivision * XDomains + 1) * (YDivision * YDomains + 1);
  int nDomains = XDomains * YDomains;
      
  fp = fopen (fileName, "w");
  assert(fp);

  fprintf (fp, "%6d\n", nGlobalNodes);
  fprintf (fp, "%6d\n", nDomains);

  fclose (fp);
}

void WriteDomainFile (char* baseFileName)
{
  int xDom, yDom;
  int iDomain;

  iDomain = 0;
  for (yDom = 0; yDom < YDomains; yDom++) {
    for (xDom = 0; xDom < XDomains; xDom++) {
      char fileName[256];
      FILE* fp;
      int x, y;
      int iNode;
      
      sprintf (fileName, "%s.%d", baseFileName, iDomain);
      fp = fopen (fileName, "w");
      assert(fp);

      iNode = 0;
      for (y = 0; y < YDivision + 1; y++) {
	for (x = 0; x < XDivision + 1; x++) {
	  Node_globalId[iNode] 
	    = x 
	    + y * (XDivision * XDomains + 1)
	    + xDom * XDivision 
	    + yDom * (XDivision * XDomains + 1) * YDivision;
	  iNode++;
	}
      }
      Node_nNodes = iNode;
      
      fprintf (fp, "%6d\n", Node_nNodes);
      fprintf (stderr, " nNodes %d \n", Node_nNodes);
      for (iNode = 0; iNode < Node_nNodes; iNode++) {
	fprintf (fp, "%8d\n", 
		 Node_globalId[iNode]);
      }
      
      fclose (fp);

      iDomain++;
    }
  }
}

void WriteMeshFile (char* baseFileName)
{
  int xDom, yDom;
  int iDomain;

  iDomain = 0;
  for (yDom = 0; yDom < YDomains; yDom++) {
    for (xDom = 0; xDom < XDomains; xDom++) {
      char fileName[256];
      FILE* fp;
      int x, y;
      int iElement;
      int iNode;
      double xBase, yBase;
      
      sprintf (fileName, "%s.%d", baseFileName, iDomain);
      fp = fopen (fileName, "w");
      assert(fp);

      iElement = 0;
      for (y = 0; y < YDivision; y++) {
	for (x = 0; x < XDivision; x++) {
	  int nodeIds[4];

	  nodeIds[0] = (x + 0) + (y + 0) * (XDivision + 1);
	  nodeIds[1] = (x + 1) + (y + 0) * (XDivision + 1);
	  nodeIds[2] = (x + 1) + (y + 1) * (XDivision + 1);
	  nodeIds[3] = (x + 0) + (y + 1) * (XDivision + 1);

	  Element_nodeIds[iElement][0] = nodeIds[0];
	  Element_nodeIds[iElement][1] = nodeIds[2];
	  Element_nodeIds[iElement][2] = nodeIds[3];
	  iElement++;
	  
	  Element_nodeIds[iElement][0] = nodeIds[0];
	  Element_nodeIds[iElement][1] = nodeIds[1];
	  Element_nodeIds[iElement][2] = nodeIds[2];
	  iElement++;
	}
      }
      Element_nElements = iElement;
      
      fprintf (fp, "%6d\n", Element_nElements);
      fprintf (stderr, " nElements %d \n", Element_nElements);
      for (iElement = 0; iElement < Element_nElements; iElement++) {
	fprintf (fp, "%6d %6d %6d\n", 
		 Element_nodeIds[iElement][0],
		 Element_nodeIds[iElement][1],
		 Element_nodeIds[iElement][2]);
      }
      
      xBase = X_SIZE * XDivision * xDom;
      yBase = Y_SIZE * YDivision * yDom;

      iNode = 0;
      for (y = 0; y < YDivision + 1; y++) {
	for (x = 0; x < XDivision + 1; x++) {
	  Node_XVec[iNode][0] = xBase + X_SIZE * x;
	  Node_XVec[iNode][1] = yBase + Y_SIZE * y;
	  iNode++;
	}
      }
      Node_nNodes = iNode;
      
      fprintf (fp, "%6d\n", Node_nNodes);
      fprintf (stderr, " nNodes %d \n", Node_nNodes);
      for (iNode = 0; iNode < Node_nNodes; iNode++) {
	fprintf (fp, "%8.4f %8.4f\n", 
		 Node_XVec[iNode][0],
		 Node_XVec[iNode][1]);
      }
      
      fclose (fp);

      iDomain++;
    }
  }
}

int main (int argc, char* argv[])
{
  if (argc < 8) {
    fprintf (stderr, 
	     " pre_2d_tri xdiv ydiv xdom ydom "
	     " global_file domain_file mesh_file \n");
    exit (1);
  }
  
  XDivision = atoi (argv[1]);
  YDivision = atoi (argv[2]);
  XDomains = atoi (argv[3]);
  YDomains = atoi (argv[4]);

  WriteGlobalFile (argv[5]);
  WriteDomainFile (argv[6]);
  WriteMeshFile (argv[7]);

  return 0;
}





