/* pre_3d_tetra.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../lib/autolinas_predefined.h"



/* Node */

int Node_nNodes;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* global node ID */
int Node_globalId[Node_maxNodes];



/* Element */

int Element_nElements;

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



#define X_SIZE 1.0
#define Y_SIZE 1.0
#define Z_SIZE 1.0

int XDivision;
int YDivision;
int ZDivision;
int XDomains;
int YDomains;
int ZDomains;

void WriteGlobalFile (char* fileName)
{
  FILE* fp;
  int nGlobalNodes 
    = (XDivision * XDomains + 1) 
    * (YDivision * YDomains + 1)
    * (ZDivision * ZDomains + 1);
  int nDomains = XDomains * YDomains * ZDomains;
      
  fp = fopen (fileName, "w");
  assert(fp);

  fprintf (fp, "%6d\n", nGlobalNodes);
  fprintf (fp, "%6d\n", nDomains);

  fclose (fp);
}

void WriteDomainFile (char* baseFileName)
{
  int xDom, yDom, zDom;
  int iDomain;

  iDomain = 0;
  for (zDom = 0; zDom < ZDomains; zDom++) {
    for (yDom = 0; yDom < YDomains; yDom++) {
      for (xDom = 0; xDom < XDomains; xDom++) {
	char fileName[256];
	FILE* fp;
	int x, y, z;
	int iNode;
	
	sprintf (fileName, "%s.%d", baseFileName, iDomain);
	fp = fopen (fileName, "w");
	assert(fp);
	
	iNode = 0;
	for (z = 0; z < ZDivision + 1; z++) {
	  for (y = 0; y < YDivision + 1; y++) {
	    for (x = 0; x < XDivision + 1; x++) {
	      Node_globalId[iNode] 
		= x 
		+ y * (XDivision * XDomains + 1)
		+ z 
		* (XDivision * XDomains + 1)
		* (YDivision * YDomains + 1)
		+ xDom * XDivision 
		+ yDom * (XDivision * XDomains + 1) * YDivision
		+ zDom 
		* (XDivision * XDomains + 1)
		* (YDivision * YDomains + 1) * ZDivision;;
	      iNode++;
	    }
	  }
	}
	Node_nNodes = iNode;
      
	fprintf (fp, "%6d\n", Node_nNodes);
	fprintf (stderr, " nNodes %d \n", Node_nNodes);
	for (iNode = 0; iNode < Node_nNodes; iNode++) {
	  fprintf (fp, "%8d\n", 
		   Node_globalId[iNode]);
	}
	
	fclose (fp);

	iDomain++;
      }
    }
  }
}

int HexToTetraMap[6][4] = {
  { 0, 1, 3, 5 }, 
  { 0, 3, 7, 5 }, 
  { 0, 4, 5, 7 }, 
  { 0, 3, 2, 7 }, 
  { 0, 2, 4, 7 }, 
  { 2, 4, 7, 6 }
};

int Grid (int x, int y, int z)
{
  return x 
    + y * (XDivision + 1) 
    + z * (XDivision + 1) * (YDivision + 1);
}

void WriteMeshFile (char* baseFileName)
{
  int xDom, yDom, zDom;
  int iDomain;

  iDomain = 0;
  for (zDom = 0; zDom < ZDomains; zDom++) {
    for (yDom = 0; yDom < YDomains; yDom++) {
      for (xDom = 0; xDom < XDomains; xDom++) {
	char fileName[256];
	FILE* fp;
	int x, y, z;
	int iElement;
	int iNode;
	double xBase, yBase, zBase;
      
	sprintf (fileName, "%s.%d", baseFileName, iDomain);
	fp = fopen (fileName, "w");
	assert(fp);
	
	iElement = 0;
	for (z = 0; z < ZDivision; z++) {
	  for (y = 0; y < YDivision; y++) {
	    for (x = 0; x < XDivision; x++) {
	      int nodeIds[8];
	      int i, v;

	      nodeIds[0] = Grid (x + 0, y + 0, z + 0);
	      nodeIds[1] = Grid (x + 1, y + 0, z + 0);
	      nodeIds[2] = Grid (x + 0, y + 1, z + 0);
	      nodeIds[3] = Grid (x + 1, y + 1, z + 0);
	      nodeIds[4] = Grid (x + 0, y + 0, z + 1);
	      nodeIds[5] = Grid (x + 1, y + 0, z + 1);
	      nodeIds[6] = Grid (x + 0, y + 1, z + 1);
	      nodeIds[7] = Grid (x + 1, y + 1, z + 1);
	      
	      for (i = 0; i < 6; i++) {
		for (v = 0; v < 4; v++) {
		  Element_nodeIds[iElement][v] 
		    = nodeIds[HexToTetraMap[i][v]];
		}
		iElement++;
	      }
	    }
	  }
	}
	Element_nElements = iElement;
      
	fprintf (fp, "%6d\n", Element_nElements);
	fprintf (stderr, " nElements %d \n", Element_nElements);
	for (iElement = 0; iElement < Element_nElements; iElement++) {
	  fprintf (fp, "%6d %6d %6d %6d\n", 
		   Element_nodeIds[iElement][0],
		   Element_nodeIds[iElement][1],
		   Element_nodeIds[iElement][2],
		   Element_nodeIds[iElement][3]);
	}
	
	xBase = X_SIZE * XDivision * xDom;
	yBase = Y_SIZE * YDivision * yDom;
	zBase = Z_SIZE * ZDivision * zDom;
	
	iNode = 0;
	for (z = 0; z < ZDivision + 1; z++) {
	  for (y = 0; y < YDivision + 1; y++) {
	    for (x = 0; x < XDivision + 1; x++) {
	      Node_XVec[iNode][0] = xBase + X_SIZE * x;
	      Node_XVec[iNode][1] = yBase + Y_SIZE * y;
	      Node_XVec[iNode][2] = zBase + Z_SIZE * z;
	      iNode++;
	    }
	  }
	}
	Node_nNodes = iNode;
	
	fprintf (fp, "%6d\n", Node_nNodes);
	fprintf (stderr, " nNodes %d \n", Node_nNodes);
	for (iNode = 0; iNode < Node_nNodes; iNode++) {
	  fprintf (fp, "%8.4f %8.4f %8.4f\n", 
		   Node_XVec[iNode][0],
		   Node_XVec[iNode][1],
		   Node_XVec[iNode][2]);
	}
	
	fclose (fp);
	
	iDomain++;
      }
    }
  }
}

int main (int argc, char* argv[])
{
  if (argc < 10) {
    fprintf (stderr, 
	     " pre_3d_tetra xdiv ydiv zdiv xdom ydom zdom "
	     " global_file domain_file mesh_file \n");
    exit (1);
  }
  
  XDivision = atoi (argv[1]);
  YDivision = atoi (argv[2]);
  ZDivision = atoi (argv[3]);
  XDomains = atoi (argv[4]);
  YDomains = atoi (argv[5]);
  ZDomains = atoi (argv[6]);

  WriteGlobalFile (argv[7]);
  WriteDomainFile (argv[8]);
  WriteMeshFile (argv[9]);

  return 0;
}





