/* autolinas_dscg_full.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dscg_inside.h"



/* symmetric matrix */
/* iterative method, conjugate gradient */
/* diagonal scaling preconditioner */
/* full matrix */



/* module variables */

static double SM[System_maxDofs][System_maxDofs];



void AutoLinAS_INSIDE_InitializeMatrix (void)
{
}

void AutoLinAS_RegisterNodalIndex (int iNode, int jNode)
{
}

void AutoLinAS_AllocateMatrix (void)
{
}

void AutoLinAS_ClearMatrix (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    for (jDof = 0; jDof < System_nDofs; jDof++) {
      SM[iDof][jDof] = 0.0;
    }
  }
}

void AutoLinAS_AssembleNodalMatrix 
(int iNode, int jNode,
 double matrix[Node_nDofs][Node_nDofs])
{
  int iDof_Node, jDof_Node;
  
  for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
    int iDof = iNode * Node_nDofs + iDof_Node;
    
    for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
      int jDof = jNode * Node_nDofs + jDof_Node;
      
      SM[iDof][jDof] += matrix[iDof_Node][jDof_Node];
    }
  }
}

void AutoLinAS_INSIDE_MultiplyMatrixAndVector
(double sv0[System_maxDofs], 
 double sv1[System_maxDofs])
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    sv1[iDof] = 0.0;

    for (jDof = 0; jDof < System_nDofs; jDof++) {
      sv1[iDof] += SM[iDof][jDof] * sv0[jDof];
    }
  }
}

void AutoLinAS_INSIDE_SetDirichletCondition (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVModified[iDof] = 0.0;
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {
      double value = SVFixed[iDof];

      for (jDof = 0; jDof < System_nDofs; jDof++) {
	SVModified[jDof] -= value * SM[jDof][iDof];
      }
      
      for (jDof = 0; jDof < System_nDofs; jDof++) {
	SM[iDof][jDof] = 0.0;
	SM[jDof][iDof] = 0.0;
      }

      SM[iDof][iDof] = 1.0;
    }
  }
}

void AutoLinAS_INSIDE_MakeScalingVector (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVScale[iDof] = sqrt (SM[iDof][iDof]);
  }
}

void AutoLinAS_INSIDE_ScaleMatrix (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVScale[iDof] = sqrt (SM[iDof][iDof]);
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    for (jDof = 0; jDof < System_nDofs; jDof++) {
      SM[iDof][jDof] /= SVScale[iDof] * SVScale[jDof];
    }
  }
}

