/* autolinas_dscg_nonzero.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_dscg_inside.h"



/* symmetric matrix */
/* iterative method, conjugate gradient */
/* diagonal scaling preconditioner */
/* sparse matrix, non-zero component storage */

/*

  # * - - * - - - - -        #**
  * # * * - - - - - -        *#**
  - * # - * - - - - -        *#*
  - * - # - * * - - -        *#**
  * - * - # * - * * -   --\  **#***
  - - - * * # - - - -   --/  **#
  - - - * - - # - * *        *#**
  - - - - * - - # * *        *#**
  - - - - * - * * # -        ***#
  - - - - - - * * - #        **#

  full matrix (non-zero component storage)

*/



/* module variables */

#if (Space_nDimensions == 2)

#if (Element_nNodes == 3) 
#define System_maxNodesOfOneRow 60
#define System_nAverageNodesOfOneRow 20
#endif

#if (Element_nNodes == 4) 
#define System_maxNodesOfOneRow 60
#define System_nAverageNodesOfOneRow 20
#endif

#endif

#if (Space_nDimensions == 3)

#if (Element_nNodes == 4) 
#define System_maxNodesOfOneRow 120
#define System_nAverageNodesOfOneRow 20
#endif

#if (Element_nNodes == 8) 
#define System_maxNodesOfOneRow 120
#define System_nAverageNodesOfOneRow 20
#endif

#if (Element_nNodes == 10) 
#define System_maxNodesOfOneRow 240
#define System_nAverageNodesOfOneRow 40
#endif

#if (Element_nNodes == 20) 
#define System_maxNodesOfOneRow 240
#define System_nAverageNodesOfOneRow 40
#endif

#endif

#define System_nAverageDofsOfOneRow (System_nAverageNodesOfOneRow * Node_nDofs)

/* node ID */
static int System_nIndexTmps[Node_maxNodes];
static int System_indexTmp[Node_maxNodes][System_maxNodesOfOneRow];
static int System_nMatrixSize;

/* degree of freedom (d.o.f.) ID */
static int System_minIndexs[System_maxDofs];
static int System_maxIndexs[System_maxDofs];
static int System_index[System_maxDofs * System_nAverageDofsOfOneRow];

static double SM[System_maxDofs * System_nAverageDofsOfOneRow];



void AutoLinAS_INSIDE_InitializeMatrix (void)
{
  int iNode;

  for (iNode = 0; iNode < System_nNodes; iNode++) {
    System_nIndexTmps[iNode] = 0; 
  }
}

void AutoLinAS_RegisterNodalIndex (int iNode, int jNode)
{
  int found = 0;
  int k;
  
  for (k = 0; k < System_nIndexTmps[iNode]; k++) {
    if (System_indexTmp[iNode][k] == jNode) { 
      found = 1;  break; 
    }
  }
  if (!found) {
    assert(System_nIndexTmps[iNode] < System_maxNodesOfOneRow);
    
    System_indexTmp[iNode][System_nIndexTmps[iNode]] = jNode;
    System_nIndexTmps[iNode]++;
  }
}

static void CompressIndex ()
{
  int iNode;
  int k;
  int iDof_Node, jDof_Node;
  int count;
  
  count = 0;
  for (iNode = 0; iNode < System_nNodes; iNode++) {

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      System_minIndexs[iNode * Node_nDofs + iDof_Node] = count;

      for (k = 0; k < System_nIndexTmps[iNode]; k++) {
	int jNode = System_indexTmp[iNode][k]; 
	
	for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
	  System_index[count] = jNode * Node_nDofs + jDof_Node;
	  count++;
	}
      }
      
      System_maxIndexs[iNode * Node_nDofs + iDof_Node] = count;
    }
  }
}

void AutoLinAS_AllocateMatrix (void)
{
  int iNode;
  int maxIndex, meanIndex;

  CompressIndex ();

  System_nMatrixSize = 0;
  maxIndex = 0;
  for (iNode = 0; iNode < System_nNodes; iNode++) {
    if (maxIndex < System_nIndexTmps[iNode]) {
      maxIndex = System_nIndexTmps[iNode];
    }
    System_nMatrixSize += System_nIndexTmps[iNode]; 
  }
  meanIndex = System_nMatrixSize / System_nNodes;
  System_nMatrixSize *= Node_nDofs * Node_nDofs;
  assert(System_nMatrixSize <= System_maxDofs * System_nAverageDofsOfOneRow);

#ifdef MEASURE
  fprintf (stderr, " nDofs %d \n", System_nDofs);
  fprintf (stderr, " maxIndex/node %d \n", maxIndex);
  fprintf (stderr, " meanIndex/node %d \n", meanIndex);
  fprintf (stderr, " NonZeroMatrix size %d \n", System_nMatrixSize);
#endif
}

void AutoLinAS_ClearMatrix (void)
{
  int i;

  for (i = 0; i < System_nMatrixSize; i++) {
    SM[i] = 0.0;
  }
}

void AutoLinAS_AssembleNodalMatrix 
(int iNode, int jNode,
 double matrix[Node_nDofs][Node_nDofs])
{
  int iDof_Node, jDof_Node;
  int k;
  
  for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
    int iDof = iNode * Node_nDofs + iDof_Node;
    
    for (jDof_Node = 0; jDof_Node < Node_nDofs; jDof_Node++) {
      int jDof = jNode * Node_nDofs + jDof_Node;
      
      for (k = System_minIndexs[iDof]; 
	   k < System_maxIndexs[iDof]; 
	   k++) {
	if (System_index[k] == jDof) {
	  SM[k] += matrix[iDof_Node][jDof_Node];
	  break;
	}
      }
    }
  }
}

void AutoLinAS_INSIDE_MultiplyMatrixAndVector
(double sv0[System_maxDofs], 
 double sv1[System_maxDofs])
{
  int iDof;
  int k;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    sv1[iDof] = 0.0;
    
    for (k = System_minIndexs[iDof]; 
	 k < System_maxIndexs[iDof]; 
	 k++) {
      int jDof = System_index[k];

      sv1[iDof] += SM[k] * sv0[jDof];
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 3 + System_nDofs * 2);
#endif
}

void AutoLinAS_INSIDE_SetDirichletCondition (void)
{
  int iDof;
  int k;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    SVModified[iDof] = 0.0;
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {

      for (k = System_minIndexs[iDof]; 
	   k < System_maxIndexs[iDof]; 
	   k++) {
	int jDof = System_index[k];

	SVModified[jDof] -= SM[k] * SVFixed[iDof];
      }
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {

    for (k = System_minIndexs[iDof]; 
	 k < System_maxIndexs[iDof]; 
	 k++) {
      int jDof = System_index[k];

      if (SVFixedFlag[jDof]) {
	SM[k] = 0.0;
      }
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {

      for (k = System_minIndexs[iDof]; 
	   k < System_maxIndexs[iDof]; 
	   k++) {
	SM[k] = 0.0;
      }
    }
  }

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    if (SVFixedFlag[iDof]) {

      for (k = System_minIndexs[iDof]; 
	   k < System_maxIndexs[iDof]; 
	   k++) {
	int jDof = System_index[k];
	
	if (jDof == iDof) {
	  SM[k] = 1.0;
	}
      }
    }
  }
}

static double GetValue (int iDof, int jDof)
{
  int k;

  for (k = System_minIndexs[iDof]; 
       k < System_maxIndexs[iDof]; 
       k++) {
    if (System_index[k] == jDof) {
      return SM[k];
    }
  }

  assert(0);
  return 0.0;
}

void AutoLinAS_INSIDE_MakeScalingVector (void)
{
  int iDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    double value = GetValue (iDof, iDof);

    assert(value != 0.0);
    assert(0.0 < value);
    SVScale[iDof] = sqrt (value);
  }
}

void AutoLinAS_INSIDE_ScaleMatrix (void)
{
  int iDof;
  int k;

  for (iDof = 0; iDof < System_nDofs; iDof++) {

    for (k = System_minIndexs[iDof]; 
	 k < System_maxIndexs[iDof]; 
	 k++) {
      int jDof = System_index[k];

      SM[k] /= SVScale[iDof] * SVScale[jDof];
    }
  }
}

