/* autolinas_fortran.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"



#ifdef AUTOLINAS_FORTRAN_BINDING_DOUBLE_UNDERSCORE
#define BIND(name) name##__
#endif

#ifdef AUTOLINAS_FORTRAN_BINDING_UNDERSCORE
#define BIND(name) name##_
#endif



void BIND(autolinas_initialize)
(int *n_nodes)
{
  AutoLinAS_Initialize (*n_nodes);
}

void BIND(autolinas_register_nodal_index)
(int *i_node, int *j_node)
{
  int iNode = *i_node - 1;
  int jNode = *j_node - 1;

  AutoLinAS_RegisterNodalIndex (iNode, jNode);
}

void BIND(autolinas_allocate_matrix) (void)
{
  AutoLinAS_AllocateMatrix ();
}

void BIND(autolinas_clear_constraint) (void)
{
  AutoLinAS_ClearConstraint ();
}

void BIND(autolinas_set_constraint)
(int *node_id, int *dof_id, double *value)
{
  int nodeId = *node_id - 1;
  int dofId = *dof_id - 1;
  
  AutoLinAS_SetConstraint (nodeId, dofId, *value);
}

void BIND(autolinas_clear_right_hand_side) (void)
{
  AutoLinAS_ClearRightHandSide ();
}

void BIND(autolinas_add_right_hand_side)
(int *node_id, int *dof_id, double *value)
{
  int nodeId = *node_id - 1;
  int dofId = *dof_id - 1;
  
  AutoLinAS_AddRightHandSide (nodeId, dofId, *value);
}

void BIND(autolinas_clear_matrix) (void)
{
  AutoLinAS_ClearMatrix ();
}

void BIND(autolinas_assemble_nodal_matrix) 
(int *i_node, int *j_node, double *matrix)
{
  int iNode = *i_node - 1;
  int jNode = *j_node - 1;
  
  AutoLinAS_AssembleNodalMatrix 
    (iNode, jNode, 
     (double (*) [Node_nDofs])matrix);
}

void BIND(autolinas_fix) (void)
{
  AutoLinAS_Fix ();
}

void BIND(autolinas_solve) (void)
{
  AutoLinAS_Solve ();
}

void BIND(autolinas_result)
(int *node_id, int *dof_id, double *value)
{
  int nodeId = *node_id - 1;
  int dofId = *dof_id - 1;
  
  *value = AutoLinAS_Result (nodeId, dofId);
}

