/* autolinas_ldl_skyline_profile.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autolinas_predefined.h"
#include "autolinas.h"

#include "autolinas_measure_inside.h"
#include "autolinas_system_inside.h"
#include "autolinas_ldl_inside.h"
#include "autolinas_ldl_skyline_inside.h"



/* symmetric matrix */
/* direct method, LtDL decomposition */
/* profile form (IJK) */
/* sparse matrix, skyline storage, lower triangle */



void AutoLinAS_INSIDE_LDLDecomposition (void)
{
  int iDof, jDof, kDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {

    for (jDof = System_minIndexs[iDof]; jDof < iDof; jDof++) {
      int iMax;

      if (System_minIndexs[jDof] < System_minIndexs[iDof]) {
	iMax = System_minIndexs[iDof];
      } else {
	iMax = System_minIndexs[jDof];
      }

      for (kDof = iMax; kDof < jDof; kDof++) {
        SM[System_ij[iDof] + jDof] 
	  -= SM[System_ij[iDof] + kDof] 
	  * SM[System_ij[jDof] + kDof]
	  * SM[System_ij[kDof] + kDof]; 
      }

      SM[System_ij[iDof] + jDof] /= SM[System_ij[jDof] + jDof];
#ifdef MEASURE
      if (0 < jDof - iMax) {
	int count = jDof - iMax;

	AutoLinAS_AddOp (M_ADD_OP, count);
	AutoLinAS_AddOp (M_MUL_OP, count * 2);
	AutoLinAS_AddIO (M_MEMORY, count * 3);
      }
#endif
    }

    for (jDof = System_minIndexs[iDof]; jDof < iDof; jDof++) {
      SM[System_ij[iDof] + iDof] 
	-= SM[System_ij[iDof] + jDof] 
	* SM[System_ij[iDof] + jDof]
	* SM[System_ij[jDof] + jDof];  
    }

#ifdef MEASURE
    {
      int count = iDof - System_minIndexs[iDof];

      AutoLinAS_AddOp (M_ADD_OP, count);
      AutoLinAS_AddOp (M_MUL_OP, count * 2);
      AutoLinAS_AddOp (M_OTHER_OP, count);
      AutoLinAS_AddIO (M_MEMORY, count * 6);
    }
#endif
  }

#ifdef MEASURE
  AutoLinAS_AddIO (M_MEMORY, System_nDofs * 2);
#endif
}

void AutoLinAS_INSIDE_ForwardReduction (void)
{
  int iDof, jDof;

  for (iDof = 0; iDof < System_nDofs; iDof++) {
    for (jDof = System_minIndexs[iDof]; jDof < iDof; jDof++) {
      SV1[iDof] -= SM[System_ij[iDof] + jDof] * SV1[jDof];
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 2);
#endif
}

void AutoLinAS_INSIDE_BackwardSubstitution (void)
{
  int iDof, jDof;
  int maxBandWidth;

  maxBandWidth = -System_minIndexs[0];
  for (iDof = 1; iDof < System_nDofs; iDof++) {
    if (maxBandWidth < iDof - System_minIndexs[iDof]) {
      maxBandWidth = iDof - System_minIndexs[iDof];
    }
  }

  SV1[System_nDofs - 1] 
    /= SM[System_ij[System_nDofs - 1] + System_nDofs - 1];

  for (iDof = System_nDofs - 2; 0 <= iDof; iDof--) {
    int jMin;

    jMin = iDof + maxBandWidth;
    if (System_nDofs - 1 < jMin) {
      jMin = System_nDofs - 1;
    } else {
      while (iDof < System_minIndexs[jMin] && jMin != iDof) {
	jMin--;
      }
    }

    SV1[iDof] /= SM[System_ij[iDof] + iDof];

    for (jDof = iDof + 1; jDof <= jMin; jDof++) {
      if (System_minIndexs[jDof] <= iDof) {
        SV1[iDof] -= SM[System_ij[jDof] + iDof] * SV1[jDof];
      }
    }
  }

#ifdef MEASURE
  AutoLinAS_AddOp (M_ADD_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_MUL_OP, System_nMatrixSize);
  AutoLinAS_AddOp (M_OTHER_OP, System_nDofs);
  AutoLinAS_AddIO (M_MEMORY, System_nMatrixSize * 2);
#endif
}

