/* profile.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/* symmetric matrix */
/* direct method, LtDL decomposition */
/* profile form (IJK) */
/* full matrix */

#define MAX_DOFS 10000

int N;
double A[MAX_DOFS][MAX_DOFS];
double B[MAX_DOFS];



void ReadMatrixFile (const char *fileName)
{
  FILE *fp;
  int i, j;

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d",
	  &N);
  assert(0 < N);
  assert(N <= MAX_DOFS);
  
  printf ("N = %d\n",
	  N);
  
  for (i = 0; i < N; i++) {
    for (j = 0; j < N; j++) {
      fscanf (fp, "%lf",
	      &A[i][j]);
    }
  }

  fclose (fp);
}

void ReadRightHandSideVectorFile (const char *fileName)
{
  FILE *fp;
  int nDofs;
  int i;

  assert(0 < N);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d",
	  &nDofs);
  assert(nDofs == N);

  for (i = 0; i < N; i++) {
    int dofId;
    double value;
    
    fscanf (fp, "%d %lf",
	    &dofId, 
	    &value);
    assert(dofId == i);

    B[i] = value;
  }

  fclose (fp);
}

void LDLDecomposition (void)
{
  int i, j, k;

  printf ("LDL decomposition start\n");

  for (i = 0; i < N; i++) {

    for (j = 0; j < i; j++) {

      for (k = 0; k < j; k++) {
        A[i][j] -= A[i][k] * A[j][k] * A[k][k];
      }

      A[i][j] /= A[j][j];
    }

    for (j = 0; j < i; j++) {
      A[i][i] -= A[i][j] * A[i][j] * A[j][j]; 
    }
  }

  printf ("LDL decomposition finish\n");
}

void ForwardReduction (void)
{
  int i, j;

  printf ("forward reduction start\n");

  for (i = 0; i < N; i++) {
    for (j = 0; j < i; j++) {
      B[i] -= A[i][j] * B[j];
    }
  }

  printf ("forward reduction finish\n");
}

void BackwardSubstitution (void)
{
  int i, j;

  printf ("backward substitution start\n");

  B[N - 1] /= A[N - 1][N - 1];

  for (i = N - 2; 0 <= i; i--) {
    B[i] /= A[i][i];

    for (j = i + 1; j < N; j++) {
      B[i] -= A[j][i] * B[j];
    }
  }

  printf ("backward substitution finish\n");
}

void WriteSolutionVectorFile (const char *fileName)
{
  FILE *fp;
  int i;

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   N);

  for (i = 0; i < N; i++) {
    double value = B[i];

    fprintf (fp, "%d %22.15e\n", 
	     i, value);
  }

  fclose (fp);
}

int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " profile  matrix_file  b_file \n"
       "     x_file_OUT \n");
    exit (1);
  }

  ReadMatrixFile (argv[1]);
  ReadRightHandSideVectorFile (argv[2]);

  LDLDecomposition ();
  ForwardReduction ();
  BackwardSubstitution ();

  WriteSolutionVectorFile (argv[3]);

  return 0;
}
