/* es_3d_hexa.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autolinas_predefined.h>

#include "matrix.h"
#include "es_3d_hexa.h"



/* Shape Function */

#define Shape_nDimensions 3
#define Shape_maxOrders 2
#define Shape_maxIPs 27
#define Shape_nNodes Element_nNodes

/* number of integral point */
static int Shape_nIPs[Shape_maxOrders] = {
  8, 27
};

#define S1_3     0.57735027

/* local coordinate at each integral point */
static double Shape_L_IPVec
[Shape_maxOrders][Shape_maxIPs][Shape_nDimensions] = {
  { 
    { -S1_3, -S1_3, -S1_3 },
    {  S1_3, -S1_3, -S1_3 },
    {  S1_3,  S1_3, -S1_3 },
    { -S1_3,  S1_3, -S1_3 },
    { -S1_3, -S1_3,  S1_3 },
    {  S1_3, -S1_3,  S1_3 },
    {  S1_3,  S1_3,  S1_3 },
    { -S1_3,  S1_3,  S1_3 }
  },
  { 
    /* NEED TO REDEFINE */
    { 0, 0, 0 }
  }
};

/* weight value at each integral point for numerical integration */
static double Shape_weight_IP[Shape_maxOrders][Shape_maxIPs] = {
  { 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0 }, 
  { 0.0 }    /* NEED TO REDEFINE */
};

/* shape function vector */
static void Shape_NVec 
(double LVec[Shape_nDimensions], 
 double RETURN_VALUE[Shape_nNodes])
{
  double gzi = LVec[0];
  double eta = LVec[1];
  double zeta = LVec[2];

#if (Element_nNodes == 8)

  RETURN_VALUE[0] = 0.25 * (1.0 - gzi) * (1.0 - eta) * (1.0 - zeta);
  RETURN_VALUE[1] = 0.25 * (1.0 + gzi) * (1.0 - eta) * (1.0 - zeta);
  RETURN_VALUE[2] = 0.25 * (1.0 + gzi) * (1.0 + eta) * (1.0 - zeta);
  RETURN_VALUE[3] = 0.25 * (1.0 - gzi) * (1.0 + eta) * (1.0 - zeta);
  RETURN_VALUE[4] = 0.25 * (1.0 - gzi) * (1.0 - eta) * (1.0 + zeta);
  RETURN_VALUE[5] = 0.25 * (1.0 + gzi) * (1.0 - eta) * (1.0 + zeta);
  RETURN_VALUE[6] = 0.25 * (1.0 + gzi) * (1.0 + eta) * (1.0 + zeta);
  RETURN_VALUE[7] = 0.25 * (1.0 - gzi) * (1.0 + eta) * (1.0 + zeta);

#endif
#if (Element_nNodes == 20)

  NEED TO REDEFINE 
  NEED TO REDEFINE 
  NEED TO REDEFINE 

#endif
}

/* derivative of shape function matrix with local coordinate */
static void Shape_dNVec_dLVec 
(double LVec[Shape_nDimensions],
 double RETURN_VALUE[Shape_nDimensions][Shape_nNodes])
{
  double gzi = LVec[0];
  double eta = LVec[1];
  double zeta = LVec[2];

#if (Element_nNodes == 8)

  RETURN_VALUE[0][0] = -0.25 * (1.0 - eta) * (1.0 - zeta);
  RETURN_VALUE[0][1] =  0.25 * (1.0 - eta) * (1.0 - zeta);
  RETURN_VALUE[0][2] =  0.25 * (1.0 + eta) * (1.0 - zeta);
  RETURN_VALUE[0][3] = -0.25 * (1.0 + eta) * (1.0 - zeta);
  RETURN_VALUE[0][4] = -0.25 * (1.0 - eta) * (1.0 + zeta);
  RETURN_VALUE[0][5] =  0.25 * (1.0 - eta) * (1.0 + zeta);
  RETURN_VALUE[0][6] =  0.25 * (1.0 + eta) * (1.0 + zeta);
  RETURN_VALUE[0][7] = -0.25 * (1.0 + eta) * (1.0 + zeta);

  RETURN_VALUE[1][0] = -0.25 * (1.0 - gzi) * (1.0 - zeta);
  RETURN_VALUE[1][1] = -0.25 * (1.0 + gzi) * (1.0 - zeta);
  RETURN_VALUE[1][2] =  0.25 * (1.0 + gzi) * (1.0 - zeta);
  RETURN_VALUE[1][3] =  0.25 * (1.0 - gzi) * (1.0 - zeta);
  RETURN_VALUE[1][4] = -0.25 * (1.0 - gzi) * (1.0 + zeta);
  RETURN_VALUE[1][5] = -0.25 * (1.0 + gzi) * (1.0 + zeta);
  RETURN_VALUE[1][6] =  0.25 * (1.0 + gzi) * (1.0 + zeta);
  RETURN_VALUE[1][7] =  0.25 * (1.0 - gzi) * (1.0 + zeta);

  RETURN_VALUE[2][0] = -0.25 * (1.0 - gzi) * (1.0 - eta);
  RETURN_VALUE[2][1] = -0.25 * (1.0 + gzi) * (1.0 - eta);
  RETURN_VALUE[2][2] = -0.25 * (1.0 + gzi) * (1.0 + eta);
  RETURN_VALUE[2][3] = -0.25 * (1.0 - gzi) * (1.0 + eta);
  RETURN_VALUE[2][4] =  0.25 * (1.0 - gzi) * (1.0 - eta);
  RETURN_VALUE[2][5] =  0.25 * (1.0 + gzi) * (1.0 - eta);
  RETURN_VALUE[2][6] =  0.25 * (1.0 + gzi) * (1.0 + eta);
  RETURN_VALUE[2][7] =  0.25 * (1.0 - gzi) * (1.0 + eta);

#endif
#if (Element_nNodes == 20)

  NEED TO REDEFINE 
  NEED TO REDEFINE 
  NEED TO REDEFINE 

#endif
}



/* Material */

/* 3D solid */

#define Material_nStressComponents 6  
/* sigma_x, sigma_y, sigma_z, tau_xy, tau_yz, tau_zx */

/* gravity force vector */
double Material_gVec[Node_nDofs];

/* young's modulus */
double Material_E;

/* poisson ratio */
double Material_nyu;

/* mass density */
double Material_rou;

static void Material_DMat
(double RETURN_VALUE
 [Material_nStressComponents][Material_nStressComponents])
{
  double a;
  double b;
  double c;

  Clear_Mat 
    (RETURN_VALUE, Material_nStressComponents, Material_nStressComponents);

  a = Material_E / (1.0 + Material_nyu) 
    / (1.0 - 2.0 * Material_nyu) * (1.0 - Material_nyu);

  b = Material_E / (1.0 + Material_nyu) 
    / (1.0 - 2.0 * Material_nyu) * Material_nyu;

  c = Material_E / (1.0 + Material_nyu) / 2.0;
  
  RETURN_VALUE[0][0] = a;
  RETURN_VALUE[0][1] = b;
  RETURN_VALUE[0][2] = b;

  RETURN_VALUE[1][0] = b;
  RETURN_VALUE[1][1] = a;
  RETURN_VALUE[1][2] = b;

  RETURN_VALUE[2][0] = b;
  RETURN_VALUE[2][1] = b;
  RETURN_VALUE[2][2] = a;

  RETURN_VALUE[3][3] = c;
  RETURN_VALUE[4][4] = c;
  RETURN_VALUE[5][5] = c;
}

static double Material_sigma_eq 
(double sigmaVec[Material_nStressComponents])
{
  double sigma_x = sigmaVec[0];
  double sigma_y = sigmaVec[1];
  double sigma_z = sigmaVec[2];
  double tau_xy = sigmaVec[3];
  double tau_yz = sigmaVec[4];
  double tau_zx = sigmaVec[5];
  
  return 1.0 / 2.0 * 
    (
     (sigma_x - sigma_y) * (sigma_x - sigma_y)
     + (sigma_y - sigma_z) * (sigma_y - sigma_z)
     + (sigma_z - sigma_x) * (sigma_z - sigma_x)
     + 6 * tau_xy * tau_xy
     + 6 * tau_yz * tau_yz
     + 6 * tau_zx * tau_zx
     );
}



/* Node */

/* 3D solid, linear elastic */

int Node_nNodes;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* displacement boundary condition */
int Node_u_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_u_BCVec[Node_maxNodes][Node_nDofs];

/* concentrated load boundary condition */
int Node_f_BCVecFlag[Node_maxNodes][Node_nDofs];
double Node_f_BCVec[Node_maxNodes][Node_nDofs];

/* displacement analysis result */
double Node_uVec[Node_maxNodes][Node_nDofs];

/* nodal equivalent stress analysis result */
double Node_sigma_eq[Node_maxNodes];

/* weight parameter for averaging on node */
double Node_weight[Node_maxNodes];



/* Element */

/* 3D solid, linear elastic */

int Element_nElements;

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];

#if (Element_nNodes == 8)

#define Element_orderOfStiffnessMatrix 1
#define Element_orderOfBodyForceVector 1

#endif
#if (Element_nNodes == 20)

#define Element_orderOfStiffnessMatrix 2
#define Element_orderOfBodyForceVector 2

#endif

static void Move_NodeVec_ElementMat
(int iElement,
 double nv[Node_maxNodes][Node_nDofs],
 double em[Element_nNodes][Node_nDofs])
{
  int iNode_Elem;

  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = Element_nodeIds[iElement][iNode_Elem];
    int iDof_Node;

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      em[iNode_Elem][iDof_Node] = nv[iNode][iDof_Node];
    }
  }
}

static void Move_NodeVec_ElementVec
(int iElement,
 double nv[Node_maxNodes][Node_nDofs],
 double ev[Element_nDofs])
{
  int iNode_Elem;

  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = Element_nodeIds[iElement][iNode_Elem];
    int iDof_Node;

    for (iDof_Node = 0; iDof_Node < Node_nDofs; iDof_Node++) {
      ev[iNode_Elem * Node_nDofs + iDof_Node] = nv[iNode][iDof_Node];
    }
  }
}

/* jacobian matrix */
static void Element_JMat
(int iElement,
 double LVec[Space_nDimensions],
 double RETURN_VALUE[Space_nDimensions][Space_nDimensions])
{
  double XMat[Element_nNodes][Space_nDimensions];
  double tmp0[Shape_nDimensions][Shape_nNodes];
  
  /* Matrix (_nNodes, _space._nDimensions) [X] = _nodeShape._{X}; */
  Move_NodeVec_ElementMat (iElement, Node_XVec, 
			   XMat);
  
  /* _[J] = _normalizedShape._d{N}_d{L} ({L}) * [X]; */
  Shape_dNVec_dLVec (LVec, 
		  tmp0);
  Mul_Mat_Mat_Mat 
    (tmp0, Shape_nDimensions, Shape_nNodes,
     XMat, Element_nNodes, Space_nDimensions, 
     RETURN_VALUE, Space_nDimensions, Space_nDimensions);
}

/* shape function vector */
static void Element_NVec
(int iElement,
 double LVec[Space_nDimensions],
 double RETURN_VALUE[Element_nNodes])
{
  /* _{N} = _normalizedShape._{N} ({L}); */
  Shape_NVec (LVec, 
	      RETURN_VALUE);
}

/* derivative of shape function matrix by global coordinate */
static void Element_dNVec_dXVec 
(int iElement,
 double LVec[Space_nDimensions], 
 double RETURN_VALUE[Space_nDimensions][Element_nNodes])
{
  double tmp0[Space_nDimensions][Space_nDimensions];
  double tmp1[Space_nDimensions][Space_nDimensions];
  double tmp2[Shape_nDimensions][Shape_nNodes];

  /* _d{N}_d{X} = _[J] ({L}).inverse 
   * _normalizedShape._d{N}_d{L} ({L});  */
  Element_JMat (iElement, LVec, 
		tmp0);
  Inverse_Mat_Mat (tmp0, 
		   tmp1);
  Shape_dNVec_dLVec (LVec,
		     tmp2);
  Mul_Mat_Mat_Mat 
    (tmp1, Space_nDimensions, Space_nDimensions, 
     tmp2, Shape_nDimensions, Shape_nNodes,
     RETURN_VALUE, Space_nDimensions, Element_nNodes);
}

/* number of integral point */
static void Element_nIPs 
(int iElement,
 int order,
 int* RETURN_VALUE)
{
  /* _nIPs = _normalizedShape._nIPs (order - 1); */
  *RETURN_VALUE = Shape_nIPs[order - 1];
}

/* local coordinate at each integral point */
static void Element_L_IPVec 
(int iElement,
 int order, 
 int iIP, 
 double RETURN_VALUE[Space_nDimensions])
{
  /* _{L_IP} = _normalizedShape._{L_IP} (order - 1, iIP); */
  Move_Vec_Vec (Shape_L_IPVec[order - 1][iIP], Shape_nDimensions, 
		RETURN_VALUE);
}

/* coefficient value (weight * volume) 
   at each integral point for numerical integration */
static void Element_V_IP
(int iElement,
 int order, 
 int iIP,
 double* RETURN_VALUE)
{
  double LVec[Space_nDimensions];
  double tmp0[Space_nDimensions][Space_nDimensions];
  double tmp1;

  /* Vector (_space._nDimensions) {L} 
     = _normalizedShape._{L_IP} (order - 1, iIP); */
  Move_Vec_Vec (Shape_L_IPVec[order - 1][iIP], Shape_nDimensions, 
		LVec);

  /* _V_IP = _[J] ({L}).determinant 
   * _normalizedShape._weight_IP (order - 1, iIP); */
  Element_JMat (iElement, LVec, 
		tmp0);
  Determinant_Mat_Sca (tmp0, 
		       &tmp1);

  *RETURN_VALUE = tmp1 * Shape_weight_IP[order - 1][iIP];

  assert(0.0 < *RETURN_VALUE);
}

/* shape function matrix */
static void Element_NMat 
(int iElement,
 double LVec[Space_nDimensions],
 double RETURN_VALUE[Node_nDofs][Element_nDofs])
{
  double NVec[Element_nNodes];
  int iNode;
  int base;
  double value;

  /* Vector (_elementShape._nNodes) {N} 
     = _elementShape._{N} ({L}); */
  Element_NVec (iElement, LVec,
		NVec);

  /* _[N].clear; */
  Clear_Mat 
    (RETURN_VALUE, Node_nDofs, Element_nDofs);

  for (iNode = 0; iNode < Element_nNodes; iNode++) {

    /* Integer base = iNode * _node._nDofs; */
    base = iNode * Node_nDofs;

    /* Scalar value = {N}.at (iNode); */
    value = NVec[iNode];

    RETURN_VALUE[0][base + 0] = value;
    RETURN_VALUE[1][base + 1] = value;
    RETURN_VALUE[2][base + 2] = value;
  }
}

/* displacement strain matrix */
static void Element_BMat
(int iElement,
 double LVec[Space_nDimensions],
 double RETURN_VALUE[Material_nStressComponents][Element_nDofs])
{
  double dNVec_dXVec[Space_nDimensions][Element_nNodes];
  int iNode;
  int base;

  /* Matrix (_space._nDimensions, _elementShape._nNodes) d{N}_d{X} 
     = _elementShape._d{N}_d{X} ({L}); */
  Element_dNVec_dXVec (iElement, LVec,
		       dNVec_dXVec);

  /* _[B].clear; */
  Clear_Mat 
    (RETURN_VALUE, Material_nStressComponents, Element_nDofs);

  for (iNode = 0; iNode < Element_nNodes; iNode++) {

    /* Integer base = iNode * _node._nDofs; */
    base = iNode * Node_nDofs;

    RETURN_VALUE[0][base + 0] = dNVec_dXVec[0][iNode];
    RETURN_VALUE[1][base + 1] = dNVec_dXVec[1][iNode];
    RETURN_VALUE[2][base + 2] = dNVec_dXVec[2][iNode];

    RETURN_VALUE[3][base + 0] = dNVec_dXVec[1][iNode];
    RETURN_VALUE[3][base + 1] = dNVec_dXVec[0][iNode];

    RETURN_VALUE[4][base + 1] = dNVec_dXVec[2][iNode];
    RETURN_VALUE[4][base + 2] = dNVec_dXVec[1][iNode];

    RETURN_VALUE[5][base + 2] = dNVec_dXVec[0][iNode];
    RETURN_VALUE[5][base + 0] = dNVec_dXVec[2][iNode];
  }
}

/* stiffness matrix */
void Element_KMat
(int iElement,
 double RETURN_VALUE[Element_nDofs][Element_nDofs])
{
  int order;
  int nIPs;
  int iIP;
  double LVec[Space_nDimensions];
  double V;
  double BMat[Material_nStressComponents][Element_nDofs];
  double tmp0[Material_nStressComponents][Material_nStressComponents];
  double tmp1[Element_nDofs][Material_nStressComponents];
  double tmp2[Element_nDofs][Element_nDofs];

  /* Integer order = _orderOfStiffnessMatrix; */
  order = Element_orderOfStiffnessMatrix;

  /* Integer nIPs = _elementShape._nIPs (order); */
  Element_nIPs (iElement, order,
		&nIPs);

  /* _[K].clear; */
  Clear_Mat 
    (RETURN_VALUE, Element_nDofs, Element_nDofs);

  for (iIP = 0; iIP < nIPs; iIP++) {

    /* Vector (_space._nDimensions) {L} 
       = _elementShape._{L_IP} (order, iIP); */
    Element_L_IPVec (iElement, order, iIP,
		     LVec);

    /* Scalar V = _elementShape._V_IP (order, iIP); */
    Element_V_IP (iElement, order, iIP,
		  &V);

    /* Matrix (_material._nStressComponents, _nDofs) [B] 
       = _[B] ({L}); */
    Element_BMat (iElement, LVec,
		  BMat);

    /* _[K] = _[K] + [B].transpose * _material._[D] * [B] * V; */
    Material_DMat 
      (tmp0);
    Mul_MatTr_Mat_Mat
      (BMat, Material_nStressComponents, Element_nDofs,
       tmp0, Material_nStressComponents, Material_nStressComponents,
       tmp1, Element_nDofs, Material_nStressComponents);
    Mul_Mat_Mat_Mat
      (tmp1, Element_nDofs, Material_nStressComponents,
       BMat, Material_nStressComponents, Element_nDofs,
       tmp2, Element_nDofs, Element_nDofs);
    MulAdd_Mat_Sca_Mat
      (tmp2, Element_nDofs, Element_nDofs,
       V,
       RETURN_VALUE);

  }
}

/* body force vector */
void Element_f_bVec
(int iElement,
 double RETURN_VALUE[Element_nDofs])
{
  double bodyForce[Node_nDofs];
  int order;
  int nIPs;
  int iIP;
  double LVec[Space_nDimensions];
  double V;
  double tmp0[Node_nDofs][Element_nDofs];
  double tmp1[Element_nDofs];

  /* Vector (_node._nDofs) {bodyForce} 
     = _material._{g} * _material._rou; */
  Mul_Vec_Sca_Vec
    (Material_gVec, Space_nDimensions,
     Material_rou,
     bodyForce);

  /* Integer order = _orderOfBodyForceVector; */
  order = Element_orderOfBodyForceVector;

  /* Integer nIPs = _elementShape._nIPs (order); */
  Element_nIPs (iElement, order,
		&nIPs);

  /* _{f_b}.clear; */
  Clear_Vec
    (RETURN_VALUE, Element_nDofs);

  for (iIP = 0; iIP < nIPs; iIP++) {

    /* Vector (_space._nDimensions) {L} 
       = _elementShape._{L_IP} (order, iIP); */
    Element_L_IPVec (iElement, order, iIP,
		     LVec);

    /* Scalar V = _elementShape._V_IP (order, iIP); */
    Element_V_IP (iElement, order, iIP,
		  &V);

    /* _{f_b} = _{f_b} + _[N] ({L}).transpose * {bodyForce} * V; */
    Element_NMat (iElement, LVec,
		  tmp0);
    Mul_MatTr_Vec_Vec
      (tmp0, Node_nDofs, Element_nDofs,
       bodyForce, Node_nDofs,
       tmp1, Element_nDofs);
    MulAdd_Vec_Sca_Vec
      (tmp1, Element_nDofs,
       V,
       RETURN_VALUE);

  }
}

/* calculate nodal equivalent stress */
void Element_calculate_sigma_eq
(int iElement)
{
  int order;
  int nIPs;
  int iIP;
  double LVec[Space_nDimensions];
  double BMat[Material_nStressComponents][Element_nDofs];
  double uVec[Element_nDofs];
  double DMat[Material_nStressComponents][Material_nStressComponents];
  double epsilonVec[Material_nStressComponents];
  double sigmaVec[Material_nStressComponents];
  double sigma_eq;
  double sigma_eq_average;
  int iNode_Elem;

  /* Integer order = _orderOfStiffnessMatrix; */
  order = Element_orderOfStiffnessMatrix;

  /* Integer nIPs = _elementShape._nIPs (order); */
  Element_nIPs (iElement, order,
		&nIPs);

  /* Vector (_nDofs) {u} = _node._{u} */
  Move_NodeVec_ElementVec (iElement, Node_uVec,
			   uVec);

  sigma_eq_average = 0.0;
  for (iIP = 0; iIP < nIPs; iIP++) {

    /* Vector (_space._nDimensions) {L} 
       = _elementShape._{L_IP} (order, iIP); */
    Element_L_IPVec (iElement, order, iIP,
		     LVec);

    /* {epsilon} = [B] * {u}; */
    Element_BMat (iElement, LVec,
		  BMat);
    Mul_Mat_Vec_Vec
      (BMat, Material_nStressComponents, Element_nDofs,
       uVec, Element_nDofs,
       epsilonVec, Material_nStressComponents);
    
    /* {sigma} = _material._[D] * {epsilon}; */
    Material_DMat 
      (DMat);
    Mul_Mat_Vec_Vec
      (DMat, Material_nStressComponents, Material_nStressComponents,
       epsilonVec, Material_nStressComponents,
       sigmaVec, Material_nStressComponents);

    sigma_eq = Material_sigma_eq (sigmaVec);
    sigma_eq_average += sigma_eq;
  }
  sigma_eq_average /=  nIPs;

  for (iNode_Elem = 0; iNode_Elem < Element_nNodes; iNode_Elem++) {
    int iNode = Element_nodeIds[iElement][iNode_Elem];

    Node_sigma_eq[iNode] += sigma_eq_average;
  }
}


