/* check_solution_full.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/* full matrix */

#define MAX_DOFS 10000

int N;
double A[MAX_DOFS][MAX_DOFS];
double B[MAX_DOFS];
double X[MAX_DOFS];

double Eps;



void ReadMatrixFile (const char *fileName)
{
  FILE *fp;
  int i, j;

  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d",
	  &N);
  assert(0 < N);
  assert(N <= MAX_DOFS);
  
  printf ("N = %d\n",
	  N);
  
  for (i = 0; i < N; i++) {
    for (j = 0; j < N; j++) {
      fscanf (fp, "%lf",
	      &A[i][j]);
    }
  }

  fclose (fp);
}

void ReadRightHandSideVectorFile (const char *fileName)
{
  FILE *fp;
  int nDofs;
  int i;

  assert(0 < N);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d",
	  &nDofs);
  assert(nDofs == N);

  for (i = 0; i < N; i++) {
    int dofId;
    double value;
    
    fscanf (fp, "%d %lf",
	    &dofId, 
	    &value);
    assert(dofId == i);

    B[i] = value;
  }

  fclose (fp);
}

void ReadSolutionVectorFile (const char *fileName)
{
  FILE *fp;
  int nDofs;
  int i;

  assert(0 < N);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d",
	  &nDofs);
  assert(nDofs == N);

  for (i = 0; i < N; i++) {
    int dofId;
    double value;
    
    fscanf (fp, "%d %lf",
	    &dofId, 
	    &value);
    assert(dofId == i);

    X[i] = value;
  }

  fclose (fp);
}

void CheckSolution (void)
{
  int i, j;

  for (i = 0; i < N; i++) {
    double b = 0.0;

    for (j = 0; j < N; j++) {
      b += A[i][j] * X[j];
    }

    if (Eps < fabs (B[i] - b)) {
      printf (" error B[%d] AX %22.15e with B %22.15e \n",
	      i,
	      b,
	      B[i]);
    }
  }
}

int main (int argc, char *argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " check_solution_full  matrix_file  b_file  x_file  eps \n");
    exit (1);
  }
  
  ReadMatrixFile (argv[1]);
  ReadRightHandSideVectorFile (argv[2]);
  ReadSolutionVectorFile (argv[3]);
  Eps = atof (argv[4]);

  CheckSolution ();

  return 0;
}
