/* compare_result.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



#define MAX_ITEMS 1000000
#define MAX_COMPONENTS 6

int NItems;
int NComponents;
double A[MAX_ITEMS][MAX_COMPONENTS];
double B[MAX_ITEMS][MAX_COMPONENTS];

double Eps;



void ReadResultAFile (const char *fileName)
{
  FILE *fp;
  int nItems;
  int iItem;
  int iComponent;
  
  assert(0 < NComponents);
  assert(NComponents <= MAX_COMPONENTS);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d",
	  &nItems);
  if (NItems == 0) {
    NItems = nItems;
  }
  assert(0 < NItems);
  assert(NItems <= MAX_ITEMS);

  for (iItem = 0; iItem < NItems; iItem++) {
    int itemId;
    
    fscanf (fp, "%d",
	    &itemId);
    assert(itemId == iItem);

    for (iComponent = 0; iComponent < NComponents; iComponent++) {
      double value;
      
      fscanf (fp, "%lf",
	      &value);
      A[iItem][iComponent] = value;
    }
  }

  fclose (fp);
}

void ReadResultBFile (const char *fileName)
{
  FILE *fp;
  int nItems;
  int iItem;
  int iComponent;
  
  assert(0 < NComponents);
  assert(NComponents <= MAX_COMPONENTS);
  assert(0 < NItems);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d",
	  &nItems);
  assert(nItems == NItems);

  for (iItem = 0; iItem < NItems; iItem++) {
    int itemId;
    
    fscanf (fp, "%d",
	    &itemId);
    assert(itemId == iItem);

    for (iComponent = 0; iComponent < NComponents; iComponent++) {
      double value;
      
      fscanf (fp, "%lf",
	      &value);
      B[iItem][iComponent] = value;
    }
  }

  fclose (fp);
}

void CompareResult (void)
{
  int iItem;
  int iComponent;

  for (iItem = 0; iItem < NItems; iItem++) {
    for (iComponent = 0; iComponent < NComponents; iComponent++) {

      if (Eps < fabs (A[iItem][iComponent] - B[iItem][iComponent])) {
	printf (" error [%d][%d] A %22.15e with B %22.15e \n",
		iItem, iComponent, 
		A[iItem][iComponent],
		B[iItem][iComponent]);
      }
    }
  }
}

int main (int argc, char *argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " compare_result  a_file  b_file  n_components  eps \n");
    exit (1);
  }
  
  NComponents = atoi (argv[3]);
  Eps = atof (argv[4]);

  ReadResultAFile (argv[1]);
  ReadResultBFile (argv[2]);

  CompareResult ();

  return 0;
}
