/* compare_vector.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



#define MAX_DOFS 1000000

int N;
double A[MAX_DOFS];
double B[MAX_DOFS];

double Eps;



void ReadVectorAFile (const char *fileName)
{
  FILE *fp;
  int nDofs;
  int i;

  assert(0 < N);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d",
	  &nDofs);
  if (N == 0) {
    N = nDofs;
  }
  assert(0 < N);
  assert(N <= MAX_DOFS);

  for (i = 0; i < N; i++) {
    int dofId;
    double value;
    
    fscanf (fp, "%d %lf",
	    &dofId, 
	    &value);
    assert(dofId == i);

    A[i] = value;
  }

  fclose (fp);
}

void ReadVectorBFile (const char *fileName)
{
  FILE *fp;
  int nDofs;
  int i;

  assert(0 < N);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);

  fscanf (fp, "%d",
	  &nDofs);
  assert(nDofs == N);

  for (i = 0; i < N; i++) {
    int dofId;
    double value;
    
    fscanf (fp, "%d %lf",
	    &dofId, 
	    &value);
    assert(dofId == i);

    B[i] = value;
  }

  fclose (fp);
}

void CompareVector (void)
{
  int i;

  for (i = 0; i < N; i++) {
    if (Eps < fabs (A[i] - B[i])) {
      printf (" error [%d] A %22.15e with B %22.15e \n",
	      i,
	      A[i],
	      B[i]);
    }
  }
}

int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " compare_vector  a_file  b_file  eps \n");
    exit (1);
  }
  
  ReadVectorAFile (argv[1]);
  ReadVectorBFile (argv[2]);
  Eps = atof (argv[3]);
  
  CompareVector ();

  return 0;
}
