/* pre_2d_quad.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../lib/autolinas_predefined.h"

#define Element_maxElements 1000000



int Node_nNodes;
int Element_nElements;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



#define X_SIZE 1.0
#define Y_SIZE 1.0

#define EPS 0.0001

int XDivision;
int YDivision;

int Grid (int x, int y)
{
  return x 
    + y * (XDivision + 1);
}

void MakeMesh (void)
{
  int iNode;
  int iElement;
  int x, y;
  double xSize, ySize;
  
  assert(Space_nDimensions == 2);
  assert(Element_nNodes == 4);

  xSize = X_SIZE;
  ySize = Y_SIZE;
  
  iNode = 0;
  for (y = 0; y < YDivision + 1; y++) {
    for (x = 0; x < XDivision + 1; x++) {
      Node_XVec[iNode][0] = xSize * x;
      Node_XVec[iNode][1] = ySize * y;
      iNode++;
    }
  }
  Node_nNodes = iNode;
  
  iElement = 0;
  for (y = 0; y < YDivision; y++) {
    for (x = 0; x < XDivision; x++) {
      Element_nodeIds[iElement][0] = Grid (x + 0, y + 0);
      Element_nodeIds[iElement][1] = Grid (x + 1, y + 0);
      Element_nodeIds[iElement][2] = Grid (x + 1, y + 1);
      Element_nodeIds[iElement][3] = Grid (x + 0, y + 1);

      iElement++;
    }
  }
  Element_nElements = iElement;
  
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   Node_nNodes);
  
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e\n", 
	     iNode,
	     Node_XVec[iNode][0],
	     Node_XVec[iNode][1]);
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  
  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   Element_nElements);

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    fprintf (fp, "%d   %d %d %d %d\n", 
	     iElement,
	     Element_nodeIds[iElement][0],
	     Element_nodeIds[iElement][1],
	     Element_nodeIds[iElement][2],
	     Element_nodeIds[iElement][3]);
  }

  fclose (fp);
}

int main (int argc, char* argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " pre_2d_quad  xdiv  ydiv \n"
       "        node_file_OUT  element_file_OUT \n");
    exit (1);
  }
  
  XDivision = atoi (argv[1]);
  YDivision = atoi (argv[2]);

  MakeMesh ();
  
  WriteNodeFile (argv[3]);
  WriteElementFile (argv[4]);

  return 0;
}





