/* pre_3d_hexa.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../lib/autolinas_predefined.h"

#define Element_maxElements 1000000



int Node_nNodes;
int Element_nElements;

/* global coordinate */
double Node_XVec[Node_maxNodes][Space_nDimensions];

/* connectivity */
int Element_nodeIds[Element_maxElements][Element_nNodes];



#define X_SIZE 1.0
#define Y_SIZE 1.0
#define Z_SIZE 1.0

#define EPS 0.0001

int XDivision;
int YDivision;
int ZDivision;

int HexToTetraMap[6][4] = {
  { 0, 1, 3, 5 }, 
  { 0, 3, 7, 5 }, 
  { 0, 4, 5, 7 }, 
  { 0, 3, 2, 7 }, 
  { 0, 2, 4, 7 }, 
  { 2, 4, 7, 6 }
};

int Grid (int x, int y, int z)
{
  return x 
    + y * (XDivision + 1) 
    + z * (XDivision + 1) * (YDivision + 1);
}

void MakeMesh (void)
{
  int iNode;
  int iElement;
  int x, y, z;
  
  assert(Space_nDimensions == 3);
  assert(Element_nNodes == 8);

  iNode = 0;
  for (z = 0; z < ZDivision + 1; z++) {
    for (y = 0; y < YDivision + 1; y++) {
      for (x = 0; x < XDivision + 1; x++) {
	Node_XVec[iNode][0] = X_SIZE * x;
	Node_XVec[iNode][1] = Y_SIZE * y;
	Node_XVec[iNode][2] = Z_SIZE * z;
	iNode++;
      }
    }
  }
  Node_nNodes = iNode;
  
  iElement = 0;
  for (z = 0; z < ZDivision; z++) {
    for (y = 0; y < YDivision; y++) {
      for (x = 0; x < XDivision; x++) {
	Element_nodeIds[iElement][0] = Grid (x + 0, y + 0, z + 0);
	Element_nodeIds[iElement][1] = Grid (x + 1, y + 0, z + 0);
	Element_nodeIds[iElement][2] = Grid (x + 1, y + 1, z + 0);
	Element_nodeIds[iElement][3] = Grid (x + 0, y + 1, z + 0);
	Element_nodeIds[iElement][4] = Grid (x + 0, y + 0, z + 1);
	Element_nodeIds[iElement][5] = Grid (x + 1, y + 0, z + 1);
	Element_nodeIds[iElement][6] = Grid (x + 1, y + 1, z + 1);
	Element_nodeIds[iElement][7] = Grid (x + 0, y + 1, z + 1);

	iElement++;
      }
    }
  }
  Element_nElements = iElement;
}

void WriteNodeFile (const char *fileName)
{
  FILE *fp;
  int iNode;
  
  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   Node_nNodes);
  
  for (iNode = 0; iNode < Node_nNodes; iNode++) {
    fprintf (fp, "%d   %22.15e %22.15e %22.15e\n", 
	     iNode,
	     Node_XVec[iNode][0],
	     Node_XVec[iNode][1],
	     Node_XVec[iNode][2]);
  }

  fclose (fp);
}

void WriteElementFile (const char *fileName)
{
  FILE *fp;
  int iElement;
  
  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   Element_nElements);

  for (iElement = 0; iElement < Element_nElements; iElement++) {
    fprintf (fp, "%d   %d %d %d %d %d %d %d %d\n", 
	     iElement,
	     Element_nodeIds[iElement][0],
	     Element_nodeIds[iElement][1],
	     Element_nodeIds[iElement][2],
	     Element_nodeIds[iElement][3],
	     Element_nodeIds[iElement][4],
	     Element_nodeIds[iElement][5],
	     Element_nodeIds[iElement][6],
	     Element_nodeIds[iElement][7]);
  }

  fclose (fp);
}

int main (int argc, char* argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " pre_3d_hexa  xdiv  ydiv  zdiv \n"
       "        node_file_OUT  element_file_OUT \n");
    exit (1);
  }
  
  XDivision = atoi (argv[1]);
  YDivision = atoi (argv[2]);
  ZDivision = atoi (argv[3]);

  MakeMesh ();

  WriteNodeFile (argv[4]);
  WriteElementFile (argv[5]);

  return 0;
}





